Package: tidywikidatar
Title: Explore 'Wikidata' Through Tidy Data Frames
Version: 0.6.2
Authors@R: 
    c(person(given = "Giorgio",
             family = "Comai",
             role = c("aut", "cre", "cph"),
             email = "giorgio.comai@cci.tn.it",
             comment = c(ORCID = "0000-0002-0515-9542")),
      person(given = "EDJNet",
             role = "fnd"))
Description: Query 'Wikidata' API <https://www.wikidata.org/wiki/Wikidata:Main_Page> with ease, get tidy data frames in response, and cache data in a local database.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: magrittr, dplyr (>= 1.1.0), tidyr, stringr, glue, DBI,
        RSQLite, tibble, purrr, cli, fs, rlang (>= 0.1.2), progress,
        jsonlite, pool, vctrs, httr2
Suggests: spelling, testthat (>= 3.0.0), knitr, rmarkdown, odbc
Config/testthat/edition: 3
Language: en-US
URL: https://edjnet.github.io/tidywikidatar/,
        https://github.com/EDJNet/tidywikidatar
BugReports: https://github.com/EDJNet/tidywikidatar/issues
VignetteBuilder: knitr
Depends: R (>= 3.5)
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-06 12:52:38 UTC; g
Author: Giorgio Comai [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-0515-9542>),
  EDJNet [fnd]
Maintainer: Giorgio Comai <giorgio.comai@cci.tn.it>
Repository: CRAN
Date/Publication: 2026-05-06 14:20:02 UTC
Built: R 4.5.2; ; 2026-05-06 15:08:18 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(":=")
export(.data)
export(as_label)
export(as_name)
export(enquo)
export(enquos)
export(expr)
export(sym)
export(syms)
export(tw_check_cache)
export(tw_check_cache_folder)
export(tw_check_cache_index)
export(tw_check_cached_items)
export(tw_check_pid)
export(tw_check_qid)
export(tw_check_search)
export(tw_connect_to_cache)
export(tw_create_cache_folder)
export(tw_disable_cache)
export(tw_disconnect_from_cache)
export(tw_enable_cache)
export(tw_extract_qualifier)
export(tw_filter)
export(tw_filter_first)
export(tw_filter_people)
export(tw_get)
export(tw_get_all_with_p)
export(tw_get_cache_db)
export(tw_get_cache_file)
export(tw_get_cache_folder)
export(tw_get_cache_table_name)
export(tw_get_cached_item)
export(tw_get_cached_qualifiers)
export(tw_get_cached_search)
export(tw_get_cached_wikipedia_category_members)
export(tw_get_cached_wikipedia_page_links)
export(tw_get_cached_wikipedia_page_qid)
export(tw_get_cached_wikipedia_page_sections)
export(tw_get_description)
export(tw_get_field)
export(tw_get_image)
export(tw_get_image_metadata)
export(tw_get_image_same_length)
export(tw_get_item)
export(tw_get_label)
export(tw_get_language)
export(tw_get_p)
export(tw_get_p1)
export(tw_get_p_wide)
export(tw_get_property)
export(tw_get_property_description)
export(tw_get_property_label)
export(tw_get_property_same_length)
export(tw_get_property_with_details)
export(tw_get_qualifiers)
export(tw_get_user_agent)
export(tw_get_wikipedia)
export(tw_get_wikipedia_base_api_url)
export(tw_get_wikipedia_category_members)
export(tw_get_wikipedia_page_links)
export(tw_get_wikipedia_page_qid)
export(tw_get_wikipedia_page_section_links)
export(tw_get_wikipedia_page_sections)
export(tw_get_wikipedia_section_links_api_url)
export(tw_get_wikipedia_sections_api_url)
export(tw_index_cache_item)
export(tw_index_cache_search)
export(tw_label)
export(tw_query)
export(tw_reset_item_cache)
export(tw_reset_qualifiers_cache)
export(tw_reset_wikipedia_category_members_cache)
export(tw_reset_wikipedia_page_cache)
export(tw_reset_wikipedia_page_links_cache)
export(tw_reset_wikipedia_page_sections_cache)
export(tw_search)
export(tw_search_item)
export(tw_search_property)
export(tw_set_cache_db)
export(tw_set_cache_folder)
export(tw_set_language)
export(tw_set_user_agent)
export(tw_write_item_to_cache)
export(tw_write_qid_of_wikipedia_page_to_cache)
export(tw_write_qualifiers_to_cache)
export(tw_write_search_to_cache)
export(tw_write_wikipedia_category_members_to_cache)
export(tw_write_wikipedia_page_links_to_cache)
export(tw_write_wikipedia_page_sections_to_cache)
import(utils)
importFrom(magrittr,"%>%")
importFrom(rlang,":=")
importFrom(rlang,.data)
importFrom(rlang,as_label)
importFrom(rlang,as_name)
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,expr)
importFrom(rlang,sym)
importFrom(rlang,syms)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tidywikidatar

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/tidywikidatar)](https://cran.r-project.org/package=tidywikidatar)
[![CRAN RStudio mirror
downloads](https://cranlogs.r-pkg.org/badges/last-month/tidywikidatar?color=blue)](https://r-pkg.org/pkg/tidywikidatar)
[![CRAN RStudio mirror
downloads](https://cranlogs.r-pkg.org/badges/grand-total/tidywikidatar?color=blue)](https://r-pkg.org/pkg/tidywikidatar)
[![R-CMD-check](https://github.com/EDJNet/tidywikidatar/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/EDJNet/tidywikidatar/actions/workflows/R-CMD-check.yaml)
[![R-universe
version](https://EDJNet.r-universe.dev/tidywikidatar/badges/version)](https://EDJNet.r-universe.dev/tidywikidatar)
<!-- badges: end -->

The goal of `tidywikidatar` is to facilitate interaction with Wikidata:

- all responses are transformed into data frames or simple character
  vectors
- it is easy to enable efficient caching in a local sqlite database
  (integration with other databases is also available)

If you want to benefit of the wealth of information stored by Wikidata,
but you do not like SPARQL queries and nested lists, then you may find
`tidywikidatar` useful. If you prefer working with SPARQL queries, check
out the [official
documentation](https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service/queries/examples)
and run your queries directly: `tidywikidatar` will not be needed.

## Installation

You can install the released version of `tidywikidatar` from
[CRAN](https://CRAN.R-project.org) with:

``` r
install.packages("tidywikidatar")
```

For the latest fixes and improvements, you can install the development
version from [Github](https://github.com/EDJNet/tidywikidatar) with:

``` r
# install.packages("remotes")
remotes::install_github("EDJNet/tidywikidatar")
```

## Limitations and known issues

`tidywikidatar` strives to strike a balance between ease of use and full
access to information available on Wikidata. This means that, for
examples, dates are returned as simple text strings, without
accompanying details such as calendar (e.g. Julian or Gregorian) and
precision (e.g. precise just to the level of century). Some amounts are
returned as numeric strings, without the accompanying unit of
measurement. The user should be aware of such issues in their own use
cases, and consider using other packages if such matters are determinant
for them. Recent versions of `tidywikidatar` include a dedicated
function to get such details, `tw_get_property_with_details()`, but it
does not currently cache results.

`tidywikidatar` is most useful in particular for the exploratory
analysis of relatively small numbers of wikidata items (dozens or
hundreds), but becomes quickly less efficient when asking for many
properties or thousands of items. Functions will take their time, but
will eventually complete. Some performance improvements may come with
future versions of `tidywikidatar`, but for larger batches of data
(large number of items/many properties), well formed queries will remain
more efficient.

## Use cases and publicly available examples

These articles or repository demonstrate some use cases for
`tidywikidatar`:

- [Finding out more about Members of the European Parliament with
  Wikidata](https://medium.com/european-data-journalism-network/a-new-r-package-for-exploring-the-wealth-of-information-stored-by-wikidata-fe85e82b6440)
- [Retrieve details about Olympics 2020 medalists via Wikipedia and
  Wikidata](https://edjnet.github.io/olympics2020nuts/) / see also [this
  interactive map based on
  Wikidata](https://edjnet.github.io/olympics2020nuts/medalists_map.html)
- [Which among the busiest air routes in Europe could actually be
  travelled by land?](https://edjnet.github.io/european_routes/)
- [Finding gendered street
  names](https://medium.com/european-data-journalism-network/finding-gendered-street-names-a-step-by-step-walkthrough-with-r-7608c2d36a77)

While the code used there may not be fully compatible or be the most
efficient with the latest version of Wikidata, they still provide a
useful term of reference.

See the vignette `vignette("wikipedia_start")` for an example of a
possible workflow.

## Before you start

This package assumes some familiarity with basic Wikidata concepts. For
reference, see [the introduction on the official
website](https://www.wikidata.org/wiki/Wikidata:Introduction).

At the most basic, you should know that every item in Wikidata has an id
(it always starts with a Q, something like `Q123456`). Each item is
described by properties (they always start with a P, something like
`P1234`).

So for example, if I am interested in the anthropologist Margaret Mead,
I will search her name on Wikidata and discover that she is
[`Q180099`](https://www.wikidata.org/wiki/Q180099). She is described by
many properties. For example, she is “an instance of”
([P31](https://www.wikidata.org/wiki/Property:P31))
“[Q5](https://www.wikidata.org/wiki/Q180099)”, which means “human”. Her
“sex or gender” ([P21](https://www.wikidata.org/wiki/Property:P21)) is
“[Q180099](https://www.wikidata.org/wiki/Q6581072)”, which means,
female. By “occupation”
([P106](https://www.wikidata.org/wiki/Property:P106)), she was
“[Q36180](https://www.wikidata.org/wiki/Q36180)”,
“[Q4773904](https://www.wikidata.org/wiki/Q4773904)”, and
“[Q674426](https://www.wikidata.org/wiki/Q674426)”, which means, a
writer, an anthropologist, and a curator. And so forth.

As you’ll see, many queries return just another wikidata id, and if you
want to know what that means, you’ll need to ask for what that id stands
for.

## How to use

`tidywikidatar` makes it easy to cache locally responses (both searches
and details about specific items) in a sqlite database to reduce load on
Wikidata’s servers and increase processing speed. These sqlite databases
are by default stored in the current working directory under a `tw_data`
folder. It may be useful to store them in a folder where they can be
retrieved easily even when working on different projects, but this is
obviously a matter of personal taste. You can enable caching for the
current session with `tw_enable_cache()`, set the cache folder to be
used throughout a session with `tw_set_cache_folder()`, and set the
language used by all functions (if not set, it defaults to English). The
first lines of a script using `tidywikidatar` would often look like
this:

``` r
library("tidywikidatar")
tw_enable_cache()
tw_set_cache_folder(path = fs::path(fs::path_home_r(), "R", "tw_data"))
tw_set_language(language = "en")
tw_create_cache_folder(ask = FALSE)
```

This also means that you can re-run code when offline, as data are
downloaded from Wikidata’s server only at first run (that is, unless you
set `cache = FALSE` or `overwrite_cache = TRUE` when calling the
respective functions, or disable caching for the current session with
`tw_disable_cache()`).

## Finding details about something

Most `tidywikidatar` functions are built around the idea that you know
what you are looking for, and just want to get what Wikidata knows about
it, assuming the preferred choice would be among the top results.

Let’s take this again from the beginning. As I mentioned, I am
interested in Margaret Mead, the famous pioneer anthropologist author of
“Coming of Age in Samoa”. This seems quite straightforward but there are
actually a number of things that are returned by searching for “Margaret
Mead” that are not the woman herself.

``` r
tw_search(search = "Margaret Mead")
#> # A tibble: 10 × 3
#>    id        label                               description                    
#>    <chr>     <chr>                               <chr>                          
#>  1 Q180099   Margaret Mead                       American anthropologist (1901-…
#>  2 Q85724626 Mead & Bateson                      business organisation          
#>  3 Q96077616 Margaret Meadows                    (1718-1781)                    
#>  4 Q75281958 Lady Margaret Meade-Fetherstonhaugh British author (1888–1977)     
#>  5 Q75506638 Margaret Meadows                    Peerage person ID=183057       
#>  6 Q76238541 Margaret Meadowe                    Peerage person ID=628312       
#>  7 Q75812372 Margaret Meade-Waldo                (died 1954)                    
#>  8 Q6759717  Margaret Mead Film Festival         annual film festival held in N…
#>  9 Q96617538 Margaret Meador                     researcher                     
#> 10 Q57231017 Margaret Mead Made Me Gay           2000 hardcover edition
```

If I am running through a list of strings, and, for example, I am
actually interested in the most famous person by that name, I can filter
result by property, using the standard form. If, for example, I want
only the first result that is associated with “an instance of” (P31) -
“human” (Q5), I can run:

``` r
tw_search(search = "Margaret Mead") %>%
  tw_filter_first(p = "P31", q = "Q5")
#> # A tibble: 1 × 3
#>   id      label         description                        
#>   <chr>   <chr>         <chr>                              
#> 1 Q180099 Margaret Mead American anthropologist (1901-1978)
```

and, as expected, I get a single output: my beloved Margaret Mead.

Where was she born? I can ask directly for P19, place of birth:

``` r
tw_get_property(id = "Q180099", p = "P19")
#> # A tibble: 1 × 4
#>   id      property value rank  
#>   <chr>   <chr>    <chr> <chr> 
#> 1 Q180099 P19      Q1345 normal
```

which, as expected, will give me another wikidata id. But what does,
“Q1345” stand for? I should ask for its label.

``` r
tw_get_label(id = "Q1345")
#> [1] "Philadelphia"
```

Alright, I know where Philadelphia is, but if it was a smaller place,
perhaps I’d need to ask in which country it is located. So I would ask
for the correspondent property, P17.

``` r
tw_get_property(id = "Q1345", p = "P17")
#> # A tibble: 1 × 4
#>   id    property value rank  
#>   <chr> <chr>    <chr> <chr> 
#> 1 Q1345 P17      Q30   normal
```

Oh, no, another Wikidata id! That’s the way it works… let’s ask for its
label:

``` r
tw_get_label(id = "Q30")
#> [1] "United States"
```

It takes some time to get used, but I suppose you get the gist of it.

You can also pipe all of the above, like this:

``` r
tw_search(search = "Margaret Mead") %>% # search for Margeret Mead
  tw_filter_first(p = "P31", q = "Q5") %>% # keep only the first result that is of a human
  tw_get_property(p = "P19") %>% # ask for the place of birth
  dplyr::pull(value) %>% # take its result and
  tw_get_property(p = "P17") %>% # ask for the country where that place of birth is located
  dplyr::pull(value) %>% # take its result and
  tw_get_label() # ask what that id stands for
#> [1] "United States"
```

And here we are, we know in which country Margaret Mead was born.

The procedure above may seem a bit convoluted, but it is actually quite
representative of how Wikidata stores information.

As you would expect, such functions can also be combined, for example,
like this:

``` r
get_bio <- function(id, language = "en") {
  tibble::tibble(
    label = tw_get_label(id = id, language = language),
    description = tw_get_description(id = id, language = language),
    year_of_birth = tw_get_property(id = id, p = "P569") %>%
      dplyr::pull(value) %>%
      head(1) %>%
      lubridate::ymd_hms() %>%
      lubridate::year(),
    year_of_death = tw_get_property(id = id, p = "P570") %>%
      dplyr::pull(value) %>%
      head(1) %>%
      lubridate::ymd_hms() %>%
      lubridate::year()
  )
}

tw_search(search = "Margaret Mead") %>%
  tw_filter_first(p = "P31", q = "Q5") %>%
  get_bio()
#> # A tibble: 1 × 4
#>   label         description                         year_of_birth year_of_death
#>   <chr>         <chr>                                       <dbl>         <dbl>
#> 1 Margaret Mead American anthropologist (1901-1978)          1901          1978
```

I can of course get the response in languages other than English, as
long as those are available on Wikidata.

``` r
tw_search(search = "Margaret Mead") %>%
  tw_filter_first(p = "P31", q = "Q5") %>%
  get_bio(language = "it")
#> # A tibble: 1 × 4
#>   label         description              year_of_birth year_of_death
#>   <chr>         <chr>                            <dbl>         <dbl>
#> 1 Margaret Mead antropologa statunitense          1901          1978
```

## Serial operations

More examples regarding serial operations, and streamlined queries over
long lists of ids will be available in a dedicated vignette in a future
version.

In the meantime, let us just say that if we wanted to have a list of all
the “awards received”
([P166](https://www.wikidata.org/wiki/Property:P166)) by Margaret Mead,
and fellow anthropologists and folklorists Ruth Benedict and Zora Neale
Hurston, we can achieve that in a single call:

``` r
tw_get_property(
  id = c("Q180099", "Q228822", "Q220480"),
  p = "P166",
  language = "en"
)
#> # A tibble: 18 × 4
#>    id      property value      rank  
#>    <chr>   <chr>    <chr>      <chr> 
#>  1 Q180099 P166     Q17144     normal
#>  2 Q180099 P166     Q782022    normal
#>  3 Q180099 P166     Q8017107   normal
#>  4 Q180099 P166     Q1967852   normal
#>  5 Q180099 P166     Q52382875  normal
#>  6 Q180099 P166     Q110471679 normal
#>  7 Q180099 P166     Q110939855 normal
#>  8 Q180099 P166     Q23685885  normal
#>  9 Q180099 P166     Q46944187  normal
#> 10 Q228822 P166     Q1967852   normal
#> 11 Q228822 P166     Q52382875  normal
#> 12 Q220480 P166     Q1316544   normal
#> 13 Q220480 P166     Q1967852   normal
#> 14 Q220480 P166     Q5461701   normal
#> 15 Q220480 P166     Q5461189   normal
#> 16 Q220480 P166     Q4765305   normal
#> 17 Q220480 P166     Q1316544   normal
#> 18 Q220480 P166     Q4705361   normal
```

Again, Wikidata ids. We can of course get their relative labels using
the functions outlined above, but `tidywikidatar` has a convenience
function - `tw_label()` that will achieve what you want in most such
cases.

``` r
tw_get_property(
  id = c("Q180099", "Q228822", "Q220480"),
  p = "P166",
  language = "en"
) %>%
  tw_label()
#> # A tibble: 18 × 4
#>    id                 property       value                                 rank 
#>    <chr>              <chr>          <chr>                                 <chr>
#>  1 Margaret Mead      award received Presidential Medal of Freedom         norm…
#>  2 Margaret Mead      award received Kalinga Prize                         norm…
#>  3 Margaret Mead      award received William Procter Prize for Scientific… norm…
#>  4 Margaret Mead      award received National Women's Hall of Fame         norm…
#>  5 Margaret Mead      award received Fellow of the American Academy of Ar… norm…
#>  6 Margaret Mead      award received honorary doctor of the University of… norm…
#>  7 Margaret Mead      award received Gold Medal of the Society of Woman G… norm…
#>  8 Margaret Mead      award received Kurt Lewin Award                      norm…
#>  9 Margaret Mead      award received Elizabeth Blackwell Award             norm…
#> 10 Ruth Benedict      award received National Women's Hall of Fame         norm…
#> 11 Ruth Benedict      award received Fellow of the American Academy of Ar… norm…
#> 12 Zora Neale Hurston award received Guggenheim Fellowship                 norm…
#> 13 Zora Neale Hurston award received National Women's Hall of Fame         norm…
#> 14 Zora Neale Hurston award received Florida Women's Hall of Fame          norm…
#> 15 Zora Neale Hurston award received Florida Artists Hall of Fame          norm…
#> 16 Zora Neale Hurston award received Anisfield-Wolf Book Awards            norm…
#> 17 Zora Neale Hurston award received Guggenheim Fellowship                 norm…
#> 18 Zora Neale Hurston award received Alabama Women's Hall of Fame          norm…
```

## Piped operations

Using the pipe (`%>%`) when working with Wikidata is often not
straightforward, due to the fact that a given property may have an
unspecified number of values. `tidywikidatar` offers dedicated functions
to work with the pipe more consistently, in particular
`tw_get_property_same_length()` (or its shorter alias `tw_get_p()`, or
`tw_get_p1()` to always get a character vector with the first property
in response).

One main distinction to keep in mind in this context is that for some
properties we really just expect to have a single value, and we are
happy to dismiss other values that may be present, while in other cases
we expect and want to retain more values.

For example, some Wikidata items have two reported dates of birth for a
single individual, possibly due to disagreements among historians about
the actual date of birth of the given person. If this is not
specifically the issue we are interested it, we may well be want just to
keep the first reported date of birth and dismiss the others. In other
cases, we probably want to retain all properties, and process them
further in subsequent steps of the pipe.

Let’s look at some of these issues with an example.

The anthropologist Franz Boas
([Q76857](https://www.wikidata.org/wiki/Q76857)) had many influential
doctoral students ([P185](https://www.wikidata.org/wiki/Property:P185)),
including the above-mentioned Margaret Mead. Who where the others? And
when and where were they born? We expect the answer to this latter
questions to be unique, and we may be fine with discarding other values
that may be recorded in Wikidata.

``` r
library("dplyr", warn.conflicts = FALSE)
library("tidyr")
students <-
  tw_get_property(id = "Q76857", p = "P185") %>% # who were Boas' doctoral students?
  transmute(
    student_label = tw_get_label(value), # get their name
    student_id = value
  ) # and keep their id


students %>%
  mutate(date_of_birth = tw_get_p(
    id = student_id,
    p = "P569", # property for date of birth
    only_first = TRUE
  )) %>%
  # we don't care about possible multiple values on when they were born
  mutate(place_of_birth = tw_get_p(
    id = student_id,
    p = "P19", # property for place of birth
    only_first = TRUE
  ) %>%
    tw_get_label())
#> # A tibble: 22 × 4
#>    student_label                 student_id date_of_birth         place_of_birth
#>    <chr>                         <chr>      <chr>                 <chr>         
#>  1 Ruth Benedict                 Q228822    +1887-06-05T00:00:00Z New York City 
#>  2 Edward Sapir                  Q191095    +1884-01-26T00:00:00Z Lębork        
#>  3 Alexander Francis Chamberlain Q32178     +1865-01-01T00:00:00Z Kenninghall   
#>  4 Manuel Gamio                  Q2602445   +1883-01-01T00:00:00Z Mexico City   
#>  5 Alexander Goldenweiser        Q1396805   +1880-01-29T00:00:00Z Kyiv          
#>  6 Irving Goldman                Q6074597   +1911-09-02T00:00:00Z <NA>          
#>  7 Melville J. Herskovits        Q711288    +1895-09-10T00:00:00Z Bellefontaine 
#>  8 George Herzog                 Q15454430  +1901-12-11T00:00:00Z Budapest      
#>  9 E. Adamson Hoebel             Q5321710   +1906-01-01T00:00:00Z Madison       
#> 10 Melville Jacobs               Q6813885   +1902-07-03T00:00:00Z New York City 
#> # ℹ 12 more rows
```

In other cases, however, we do expect multiple valid values. For
example, we expect them to have a single place and date of birth, but
quite possibly to have worked in different locations at different points
in their career.

Here is how we may want to go if we want, for example, to create a map
of all the universities where one of Franz Boas’ doctoral students has
worked. We get the id of all the places where they have worked, check if
they are universities or not, and then get the coordinates for the given
institutions.

``` r
students %>%
  mutate(worked_at_id = tw_get_p(
    id = student_id,
    p = "P108", # property for employer
    only_first = FALSE
  )) %>% # not only the first result
  unnest(worked_at_id) %>%
  filter(is.na(worked_at_id) == FALSE) %>% # remove those for which we have no employer
  mutate(worked_at_label = tw_get_label(worked_at_id)) %>%
  # but keep in mind we are only interested in the employer if they are a university
  # so we ask what `instance of` the employer is.
  mutate(employer_instance_of = tw_get_p(
    id = worked_at_id,
    p = "P31",
    only_first = FALSE
  )) %>%
  unnest(employer_instance_of) %>%
  mutate(employer_instance_of_label = tw_get_label(employer_instance_of)) %>%
  # some institutions may be e.g. "instance of" -> "private university", not of "university"
  # so whe check what "subclass of" that id
  mutate(employer_instance_of2 = tw_get_p(
    id = worked_at_id,
    p = "P31",
    only_first = FALSE
  )) %>%
  unnest(employer_instance_of2) %>%
  mutate(employer_instance_of2_subclass_of = tw_get_p(
    id = employer_instance_of2,
    p = "P279",
    only_first = FALSE
  )) %>%
  unnest(employer_instance_of2_subclass_of) %>%
  # keep only if employer is a university (or something which is a subclass of university)
  filter(employer_instance_of == "Q3918" | employer_instance_of2_subclass_of == "Q3918") %>%
  distinct(student_label, worked_at_id, worked_at_label) %>%
  mutate(worked_at_coordinates = tw_get_p(worked_at_id,
    p = "P625",
    only_first = TRUE
  )) %>%
  select(-worked_at_id) %>%
  separate(worked_at_coordinates, into = c("lat", "lon"), sep = ",")
#> # A tibble: 27 × 4
#>    student_label                 worked_at_label          lat              lon  
#>    <chr>                         <chr>                    <chr>            <chr>
#>  1 Ruth Benedict                 Columbia University      40.8075          -73.…
#>  2 Edward Sapir                  Yale University          41.311111111111  -72.…
#>  3 Edward Sapir                  University of Chicago    41.789722222222  -87.…
#>  4 Alexander Francis Chamberlain Clark University         42.250977        -71.…
#>  5 Alexander Goldenweiser        Columbia University      40.8075          -73.…
#>  6 Alexander Goldenweiser        University of Washington 47.6541666666667 -122…
#>  7 Melville J. Herskovits        Northwestern University  42.0564594       -87.…
#>  8 Melville J. Herskovits        Columbia University      40.8075          -73.…
#>  9 Melville J. Herskovits        Howard University        38.921666666667  -77.…
#> 10 E. Adamson Hoebel             New York University      40.73            -73.…
#> # ℹ 17 more rows
```

Starting with version 0.5, to reduce typing, `tw_get_p()` can be used
instead of the more verbose `tw_get_property_same_length()`. By default,
`tw_get_p()` returns a vector of lists, as it is common for a property
to have multiple values. If one is interested only in the first
preferred value associated with a given property, `tw_get_p1()`, which
consistently returns a character vector of the same length as the input,
can be used instead.

Starting with version 0.5.2, a more efficient `tw_get_p_wide()` has been
introduced to replicate a common use pattern, i.e. getting a number of
property of a given set of identifiers, and retrieving their labels.

``` r
tw_get_p_wide(
  id = c("Q180099", "Q228822", "Q191095"),
  p = c("P27", "P19", "P20"),
  only_first = TRUE,
  preferred = TRUE,
  label = TRUE
)
#> # A tibble: 3 × 5
#>   id      label         P27           P19           P20          
#>   <chr>   <chr>         <chr>         <chr>         <chr>        
#> 1 Q180099 Margaret Mead United States Philadelphia  New York City
#> 2 Q228822 Ruth Benedict United States New York City New York City
#> 3 Q191095 Edward Sapir  United States Lębork        New Haven
```

It is however common for properties to have more than one meaningful
value. By default, `tw_get_p_wide()` would get these as lists columns,
but there is also an additional parameter to facilitate sharing the
result, for example, as a csv file.

``` r
tw_get_p_wide(
  id = c("Q180099", "Q228822", "Q191095"),
  p = c("P108", "P26", "P451"),
  only_first = TRUE,
  preferred = TRUE,
  label = TRUE,
  unlist = TRUE,
  collapse = ";"
)
#> # A tibble: 3 × 5
#>   id      label         P108                P26                       P451      
#>   <chr>   <chr>         <chr>               <chr>                     <chr>     
#> 1 Q180099 Margaret Mead Columbia University Luther Cressman           Ruth Bene…
#> 2 Q228822 Ruth Benedict Columbia University Stanley Rossiter Benedict Margaret …
#> 3 Q191095 Edward Sapir  Yale University     <NA>                      <NA>
```

## Qualifiers

In most cases, things are quite straightforward: each item has one or
more values for a given property.

However, some properties have additional qualifiers.

As an example, let’s look at someone whose life is seemingly less
adventurous than that of Margaret Mead, but whose Wikidata page has
properties with a more interesting combination of qualifiers: the former
president of the European Parliament David Sassoli
([Q2391857](https://www.wikidata.org/wiki/Q2391857)).

(N.B. this example based on David Sassoli was included in this document
before his premature death in early 2022)

If we look at his “positions held”
([P39](https://www.wikidata.org/wiki/Property:P39)), we find the
following:

``` r
purrr::map_chr(
  .x = tw_get_property(id = "Q2391857", p = "P39") %>% dplyr::pull(value),
  .f = tw_get_label
)
#> [1] "President of the European Parliament"
#> [2] "Member of the European Parliament"   
#> [3] "Member of the European Parliament"   
#> [4] "Member of the European Parliament"
```

He has been more than once “member of the European Parliament”, and once
“President of the European Parliament”. But this is not all that
Wikidata knows about it: each of these properties comes with qualifiers.

``` r
qualifiers_df <- tw_get_qualifiers(id = "Q2391857", p = "P39")
qualifiers_df
#> # A tibble: 28 × 8
#>    id       property qualifier_id qualifier_property qualifier_value      
#>    <chr>    <chr>    <chr>        <chr>              <chr>                
#>  1 Q2391857 P39      Q740126      P580               +2019-07-03T00:00:00Z
#>  2 Q2391857 P39      Q740126      P1365              Q440710              
#>  3 Q2391857 P39      Q740126      P582               +2022-01-11T00:00:00Z
#>  4 Q2391857 P39      Q740126      P1534              Q5247364             
#>  5 Q2391857 P39      Q740126      P1366              Q7351526             
#>  6 Q2391857 P39      Q27169       P580               +2019-07-02T00:00:00Z
#>  7 Q2391857 P39      Q27169       P582               +2022-01-11T00:00:00Z
#>  8 Q2391857 P39      Q27169       P1534              Q5247364             
#>  9 Q2391857 P39      Q27169       P1366              Q110513292           
#> 10 Q2391857 P39      Q27169       P2937              Q64038205            
#> # ℹ 18 more rows
#> # ℹ 3 more variables: qualifier_value_type <chr>, rank <chr>, set <dbl>
```

As usual, Wikidata presents everything as combinations of properties and
values. Let’s translate each of these to their respective label, and
separate each set of information we have about the “positions held” by
Mr. Sassoli:

``` r
qualifiers_labelled_df <- qualifiers_df %>%
  dplyr::transmute(
    who = tw_get_label(id = id, language = "en"),
    did = tw_get_property_label(property = property, language = "en"),
    what = tw_get_label(id = qualifier_id, language = "en"),
    how = tw_get_property_label(property = qualifier_property, language = "en"),
    value = purrr::map_chr(
      .x = qualifier_value,
      .f = function(x) {
        if (stringr::str_starts(
          string = x,
          pattern = "Q"
        )) {
          tw_get_label(
            id = x,
            language = "en"
          )
        } else {
          stringr::str_extract(
            string = x,
            pattern = "[[:digit:]]{4}-[[:digit:]]{2}-[[:digit:]]{2}"
          )
        }
      }
    ),
    set = set
  )

qualifiers_labelled_df %>%
  dplyr::group_by(set) %>%
  knitr::kable()
```

| who | did | what | how | value | set |
|:---|:---|:---|:---|:---|---:|
| David Sassoli | position held | President of the European Parliament | start time | 2019-07-03 | 1 |
| David Sassoli | position held | President of the European Parliament | replaces | Antonio Tajani | 1 |
| David Sassoli | position held | President of the European Parliament | end time | 2022-01-11 | 1 |
| David Sassoli | position held | President of the European Parliament | end cause | death in office | 1 |
| David Sassoli | position held | President of the European Parliament | replaced by | Roberta Metsola | 1 |
| David Sassoli | position held | Member of the European Parliament | start time | 2019-07-02 | 2 |
| David Sassoli | position held | Member of the European Parliament | end time | 2022-01-11 | 2 |
| David Sassoli | position held | Member of the European Parliament | end cause | death in office | 2 |
| David Sassoli | position held | Member of the European Parliament | replaced by | Camilla Laureti | 2 |
| David Sassoli | position held | Member of the European Parliament | parliamentary term | Ninth European Parliament | 2 |
| David Sassoli | position held | Member of the European Parliament | parliamentary group | Progressive Alliance of Socialists and Democrats | 2 |
| David Sassoli | position held | Member of the European Parliament | electoral district | Central Italy | 2 |
| David Sassoli | position held | Member of the European Parliament | elected in | 2019 European Parliament election | 2 |
| David Sassoli | position held | Member of the European Parliament | represents | Democratic Party | 2 |
| David Sassoli | position held | Member of the European Parliament | start time | 2014-07-01 | 3 |
| David Sassoli | position held | Member of the European Parliament | end time | 2019-07-01 | 3 |
| David Sassoli | position held | Member of the European Parliament | parliamentary term | Eighth European Parliament | 3 |
| David Sassoli | position held | Member of the European Parliament | parliamentary group | Progressive Alliance of Socialists and Democrats | 3 |
| David Sassoli | position held | Member of the European Parliament | electoral district | Central Italy | 3 |
| David Sassoli | position held | Member of the European Parliament | elected in | 2014 European Parliament election | 3 |
| David Sassoli | position held | Member of the European Parliament | represents | Democratic Party | 3 |
| David Sassoli | position held | Member of the European Parliament | start time | 2009-07-14 | 4 |
| David Sassoli | position held | Member of the European Parliament | end time | 2014-06-30 | 4 |
| David Sassoli | position held | Member of the European Parliament | parliamentary term | Seventh European Parliament | 4 |
| David Sassoli | position held | Member of the European Parliament | parliamentary group | Progressive Alliance of Socialists and Democrats | 4 |
| David Sassoli | position held | Member of the European Parliament | electoral district | Central Italy | 4 |
| David Sassoli | position held | Member of the European Parliament | elected in | 2009 European Parliament election | 4 |
| David Sassoli | position held | Member of the European Parliament | represents | Democratic Party | 4 |

That’s quite a lot of useful detail. The construction of the request can
be quite complicated, but keep in mind that if you do this
programmatically you will likely use this for filtering a specific piece
of information based on a combination of properties, and you will only
less frequently need to extract all available information.

Fundamentally, you won’t be touching anything that is not a vector or a
tidy data frame, which is ultimately a key goal of `tidywikidatar`: make
use of the wealth of information stored by Wikidata from R without
having to deal with either nested lists or SPARQL queries.

## Getting the right property when more than one is available

In Wikidata, the [order in which
statements](https://www.wikidata.org/wiki/Wikidata:Glossary#Order_of_statements)
for a property are shown depends on a number of factors. Consistent with
the API behaviour, `tidywikidatar` returns them in the same order as
they appear on the online on Wikidata dot org. Depending on the use case
and subsequent processing operations this may be either completely
irrelevant or very important, with a big impact even on the most basic
of queries.

For example, let’s compare results when we are trying to find out in
which country ([P17](https://www.wikidata.org/wiki/Property:P17)) London
([Q84](https://www.wikidata.org/wiki/Q84)) and Rome
([Q220](https://www.wikidata.org/wiki/Q220)) are located.

If we ask Wikidata in which country London is located, this is the
response we get:

``` r
tw_get_property(id = "Q84", p = "P17") %>%
  dplyr::mutate(value = tw_get_label(value))
#> # A tibble: 8 × 4
#>   id    property value                                       rank     
#>   <chr> <chr>    <chr>                                       <chr>    
#> 1 Q84   P17      Roman Empire                                normal   
#> 2 Q84   P17      Kingdom of Essex                            normal   
#> 3 Q84   P17      Kingdom of Mercia                           normal   
#> 4 Q84   P17      Kingdom of Wessex                           normal   
#> 5 Q84   P17      Kingdom of England                          normal   
#> 6 Q84   P17      Kingdom of Great Britain                    normal   
#> 7 Q84   P17      United Kingdom of Great Britain and Ireland normal   
#> 8 Q84   P17      United Kingdom                              preferred
```

These statements may all be fairly accurate at different points in time,
as we would see if we looked at the qualifiers of each of these
statements (see above) or check the respective [section on Wikidata’s
website](https://www.wikidata.org/wiki/Q84#P17). The order, however, is
determined by a number of factors and this may lead to inconsistent
results. If we are interested in having just one result, as is often the
case when processing large amounts of items, can we safely pick the
first (or last) and be sure it’s the most recent? As it emerges looking
at the same for property for Rome, this is not the case.

``` r
tw_get_property(id = "Q220", p = "P17") %>%
  dplyr::mutate(value = tw_get_label(value))
#> # A tibble: 11 × 4
#>    id    property value                rank     
#>    <chr> <chr>    <chr>                <chr>    
#>  1 Q220  P17      Italy                preferred
#>  2 Q220  P17      Papal States         normal   
#>  3 Q220  P17      Kingdom of Italy     normal   
#>  4 Q220  P17      Ostrogothic Kingdom  normal   
#>  5 Q220  P17      Byzantine Empire     normal   
#>  6 Q220  P17      Kingdom of Italy     normal   
#>  7 Q220  P17      Roman Kingdom        normal   
#>  8 Q220  P17      Roman Republic       normal   
#>  9 Q220  P17      Roman Empire         normal   
#> 10 Q220  P17      Western Roman Empire normal   
#> 11 Q220  P17      Vatican City         normal
```

So while we may be tempted to just keep the first statement returned by
Wikidata for the given property, this is probably not what we want.

``` r
tibble::tibble(city_qid = c("Q84", "Q220")) %>%
  dplyr::mutate(
    city_label = tw_get_label(city_qid),
    country_qid = tw_get_p(
      id = city_qid,
      p = "P17",
      only_first = TRUE
    )
  ) %>%
  dplyr::mutate(country_label = tw_get_label(country_qid))
#> # A tibble: 2 × 4
#>   city_qid city_label country_qid country_label
#>   <chr>    <chr>      <chr>       <chr>        
#> 1 Q84      London     Q2277       Roman Empire 
#> 2 Q220     Rome       Q38         Italy
```

Besides looking at the qualifiers, the standard way for Wikidata to
choose which is the “preferred” statement is the dedicated ranking
element (in the online interface, a small dot with arrows next to the
label), which can either be “preferred”, “normal”, or “deprecated”. In
piped operations, we get the “preferred” property by setting `preferred`
to `TRUE` in `tw_get_p()`.

``` r
tibble::tibble(city_qid = c("Q84", "Q220")) %>%
  dplyr::mutate(
    city_label = tw_get_label(city_qid),
    country_qid = tw_get_p(
      id = city_qid,
      p = "P17",
      preferred = TRUE,
      only_first = TRUE
    )
  ) %>%
  dplyr::mutate(country_label = tw_get_label(country_qid))
#> # A tibble: 2 × 4
#>   city_qid city_label country_qid country_label 
#>   <chr>    <chr>      <chr>       <chr>         
#> 1 Q84      London     Q145        United Kingdom
#> 2 Q220     Rome       Q38         Italy
```

Keep in mind that there may be more than one “preferred” statement, so
setting `preferred` to `TRUE` is no guarantee of having a single result:
for example, London is both “[capital
of](https://www.wikidata.org/wiki/Q84#P1376)”
([P1376](https://www.wikidata.org/wiki/Property:P1376)) the United
Kingdom and England, and both statements are “preferred”. Rome is
capital of Italy and Lazio (the region where it is located), and both
are “preferred”.

When the “preferred” option does not give the desired result or gives
more than one, in some cases it may be useful to use instead the
parameter `latest_start_time`, to pick the statement that has the most
recent “start time”
([P580](https://www.wikidata.org/wiki/Property:P580)) qualifier (this
can also be used in combination with `preferred`). This option slows a
bit the process as it depends on a call to `tw_get_qualifiers()` to
retrieve and cache relevant details.

``` r
tibble::tibble(city_qid = c("Q84", "Q220")) %>%
  dplyr::mutate(
    city_label = tw_get_label(city_qid),
    country_qid = tw_get_p(
      id = city_qid,
      p = "P17",
      latest_start_time = TRUE,
      only_first = TRUE
    )
  ) %>%
  dplyr::mutate(country_label = tw_get_label(country_qid))
#> # A tibble: 2 × 4
#>   city_qid city_label country_qid country_label 
#>   <chr>    <chr>      <chr>       <chr>         
#> 1 Q84      London     Q145        United Kingdom
#> 2 Q220     Rome       Q38         Italy
```

If none of the above works, then you may still be able to get consistent
results through customs solutions based on `tw_get_qualifiers()`, or by
checking the validity of alternative results based on their properties
(for example, many of the properties of “Roman empire”
([Q2277](https://www.wikidata.org/wiki/Q2277)) could be used to
determine that it is not, in fact, a contemporary country).

## Queries

All of the above works similarly to how we often use websites such as
Wikipedia, or search engines: we search for something specific to find
information about it. Wikidata, however, has powerful tools for complex
queries. Think something like “give me all of these fields for all items
that have this value for this property, but not that other value for
that other property”.

To achieve this, you can run queries, following [instructions on
Wikidata.org](https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service/queries/examples).
This is powerful, but perhaps somewhat intimidating for those who are
less familiar with database queries, SPARQL, and the likes.

`tidiwikidatar` does not currently plan to deal with complex queries.
However, at this stage it has a basic function, `tw_query()`, which
should instantly make sense for R users.

Say, for example, you are interested in all women (P21 == Q6581072) who
are resistance fighters (P106 == Q6581072).

You can then make a data frame with two columns (p and q), and some
requirements, like this:

``` r
query_df <- tibble::tribble(
  ~p, ~q,
  "P106", "Q1397808",
  "P21", "Q6581072"
)

# if you prefer, you can input the same as a list, like this:
# query_l <- list(c(p = "P106", q = "Q1397808"),
#                c(p = "P21", q = "Q6581072"))

query_df
#> # A tibble: 2 × 2
#>   p     q       
#>   <chr> <chr>   
#> 1 P106  Q1397808
#> 2 P21   Q6581072
```

You can then pass it to `tw_query()`, and get a nicely formatted
dataframe with all women who are resistance fighters on Wikidata.

``` r
tw_query(query = query_df)
#> # A tibble: 1,547 × 3
#>    id      label                 description                                    
#>    <chr>   <chr>                 <chr>                                          
#>  1 Q214198 Eva Besnyö            Hungarian photographer (1910-2003)             
#>  2 Q214515 Lisa Fittko           Hungarian resistance fighter                   
#>  3 Q229062 Zoya Kosmodemyanskaya Soviet resistance member of World War II and H…
#>  4 Q237416 Dubravka Ugrešić      Croatian writer (1949–2023)                    
#>  5 Q237429 Laskarina Bouboulina  heroine of the Greek War of Independence       
#>  6 Q241097 Corrie ten Boom       Dutch resistance hero and writer               
#>  7 Q246835 Trinity               fictional character from the films of the Matr…
#>  8 Q255016 Violette Szabo        French-British SOE spy                         
#>  9 Q255271 Änne Meier            German welfare worker and resistance fighter   
#> 10 Q255276 Aenne Saefkow         German politician (1902-1962)                  
#> # ℹ 1,537 more rows
```

Or perhaps, you are interested only in women who are resistance fighters
who have “France” ([Q142](https://www.wikidata.org/wiki/Q142)) as
“country of citizenship”
([P27](https://www.wikidata.org/wiki/Property:P27))? And perhaps you
want the description in Italian, and if not available in French, and
only then look for other fallback options?

``` r
fr_resistance_fighters_df <- tibble::tribble(
  ~p, ~q,
  "P106", "Q1397808", # Occupation: resistance fighter
  "P21", "Q6581072", # Sex or gender: female
  "P27", "Q142"
) %>% # Country of citizenship: France
  tw_query(language = c("it", "fr"))

fr_resistance_fighters_df
#> # A tibble: 238 × 3
#>    id       label                description                                    
#>    <chr>    <chr>                <chr>                                          
#>  1 Q2696536 Yolande Beekman      "espionne et agente secret des Special Operati…
#>  2 Q2837293 Alix d'Unienville    "espionne britannique"                         
#>  3 Q2844547 Amy Collin           "attrice francese"                             
#>  4 Q3009723 Cécile Cerf          "résistante française"                         
#>  5 Q3081207 Francine Fromond     "résistante communiste française"              
#>  6 Q3105910 Gilberte Brossolette "journaliste et politicienne française (1905-2…
#>  7 Q3132483 Henriette Moriamé    ""                                             
#>  8 Q3144891 Hélène de Suzannet   "politica francese"                            
#>  9 Q3175990 Jeanne Bohec         "scrittrice francese"                          
#> 10 Q3176052 Jeanne Gaillard      "partigiana e storica francese"                
#> # ℹ 228 more rows
```

You can also ask other fields, beyond label and description, using the
`field` parameter of `tw_query()`. But for this readme, I’ll keep things
simple. Do you want more information about these results without
learning yet another set of Wikidata terminology? You can still use the
same commands described above, e.g.

``` r
fr_resistance_fighters_df %>%
  dplyr::slice(1) %>%
  get_bio()
#> # A tibble: 1 × 4
#>   label           description                  year_of_birth year_of_death
#>   <chr>           <chr>                                <dbl>         <dbl>
#> 1 Yolande Beekman French SOE agent (1911–1944)          1911          1944
```

Keep in mind that Wikidata queries are not cached locally.

## Getting Wikidata identifiers from a Wikipedia page

Besides querying Wikidata and using the basic `tw_search()` function
described above, `tidywikidatar` includes function that facilitate
retrieving Wikidata identifiers based on Wikipedia pages, as well as the
Wikidata identifiers corresponding to all the Wikipedia pages included
in a given Wikipedia page. This may be useful in particular on Wikipedia
pages that are lists of other pages, or as an alternative approach for
finding relations between various Wikidata items.

In this case, the starting point is usually the full URL or the title of
a Wikipedia page, which give the same result (the user, however, should
be mindful of redirection if using the title).

``` r
tw_get_wikipedia_page_qid(title = "Margaret Mead")
#> # A tibble: 1 × 7
#>   title_url     wikipedia_title wikipedia_id qid     description  disambiguation
#>   <chr>         <chr>                  <dbl> <chr>   <chr>        <lgl>         
#> 1 Margaret Mead Margaret Mead          19617 Q180099 American cu… FALSE         
#> # ℹ 1 more variable: language <chr>
```

``` r
tw_get_wikipedia_page_qid(url = "https://en.wikipedia.org/wiki/Margaret_Mead")
#> # A tibble: 1 × 7
#>   title_url     wikipedia_title wikipedia_id qid     description  disambiguation
#>   <chr>         <chr>                  <dbl> <chr>   <chr>        <lgl>         
#> 1 Margaret_Mead Margaret Mead          19617 Q180099 American cu… FALSE         
#> # ℹ 1 more variable: language <chr>
```

Depending on the workflow, it is also possible to get the full link to
the Wikipedia page starting from a given Wikidata identifier.

``` r
tw_get_wikipedia(id = "Q180099")
#> [1] "https://en.wikipedia.org/wiki/Margaret Mead"
```

Who and what is mentioned in Margaret Mead’s Wikipedia page? As it turns
out, hundreds of pages, including a variety of people, places, concepts,
etc.

``` r
wikipedia_df <- tw_get_wikipedia(id = "Q180099") %>%
  tw_get_wikipedia_page_links()

wikipedia_df
#> # A tibble: 960 × 8
#>    source_title_url source_wikipedia_title source_qid wikipedia_title   
#>    <chr>            <chr>                  <chr>      <chr>             
#>  1 Margaret Mead    Margaret Mead          Q180099    Alex Barker       
#>  2 Margaret Mead    Margaret Mead          Q180099    Alfred S. Hayes   
#>  3 Margaret Mead    Margaret Mead          Q180099    A Rap on Race     
#>  4 Margaret Mead    Margaret Mead          Q180099    Abby Kelley Foster
#>  5 Margaret Mead    Margaret Mead          Q180099    Abigail Adams     
#>  6 Margaret Mead    Margaret Mead          Q180099    Admiralty Islands 
#>  7 Margaret Mead    Margaret Mead          Q180099    Affinity (law)    
#>  8 Margaret Mead    Margaret Mead          Q180099    Aimee Mullins     
#>  9 Margaret Mead    Margaret Mead          Q180099    Akhil Gupta       
#> 10 Margaret Mead    Margaret Mead          Q180099    Alan H. Goodman   
#> # ℹ 950 more rows
#> # ℹ 4 more variables: wikipedia_id <dbl>, qid <chr>, description <chr>,
#> #   language <chr>
```

What if we are potentially interested only in the people mentioned in
this page? We proceed as usual, checking which of these are “instance
of” (“P19”) “human” (“Q5”), and take it from there.

``` r
wikipedia_df %>%
  dplyr::pull(wikidata_id) %>%
  tw_get_property(p = "P31") %>%
  dplyr::filter(value == "Q5")
```

All functions that interact with Wikipedia and the related MediaWiki API
are not cached locally at this stage.

For a more extended example of exploring Wikidata starting from
Wikipedia, consult the dedicated vignette with
`vignette("wikipedia_start")`

## Getting images, including credits

Many Wikidata items have an image that can be used for illustrative
purposes. `tw_get_image()` facilitate getting the link to the WikiMedia
Commons page where more details about the image can be found, as well as
a direct link to the image at the desired resolution for direct embeds.

``` r
tw_get_image(id = "Q180099", format = "commons") %>%
  dplyr::pull(image)
#> [1] "https://commons.wikimedia.org/wiki/File:Margaret Mead (1901-1978).jpg"
```

``` r
tw_get_image(id = "Q180099", format = "embed", width = 300) %>%
  dplyr::pull(image)
#> [1] "https://commons.wikimedia.org/w/index.php?title=Special:Redirect/file/Margaret Mead (1901-1978).jpg&width=300"
```

The user should be mindful that these links depend on the filename of
the image, and (unlike Wikidata Q identifiers) may be changed without
offering a redirect from the previous file to the most recent one. If
there are no relevant copyright limitations, depending on the use case,
it may be wise to store images locally.

Wikidata itself does not include details about copyright of the image,
nor an easy way to get information about the author, or a suggested way
to credit the image. All of these are available through the Wikimedia
commons API. `tidywikidatar` includes a convenience function to get
access to all such details:

``` r
tw_get_image_metadata(id = "Q180099") %>%
  tidyr::pivot_longer(
    cols = dplyr::everything(),
    names_to = "property",
    values_to = "values",
    values_transform = list(values = as.character)
  )
#> # A tibble: 19 × 2
#>    property                   values                                            
#>    <chr>                      <chr>                                             
#>  1 id                         "Q180099"                                         
#>  2 image_filename             "Margaret Mead (1901-1978).jpg"                   
#>  3 object_name                "Margaret Mead (1901-1978)"                       
#>  4 image_description          "<b>Subject</b>: Mead, Margaret\n<p>       Intern…
#>  5 categories                 "Margaret Mead|Pleasure|Moose International|Cente…
#>  6 assessments                ""                                                
#>  7 credit                     "<p><a rel=\"nofollow\" class=\"external text\" h…
#>  8 artist                     "<a rel=\"nofollow\" class=\"external text\" href…
#>  9 permission                 "<a rel=\"nofollow\" class=\"external text\" href…
#> 10 license_short_name         "No restrictions"                                 
#> 11 license_url                "https://www.flickr.com/commons/usage/"           
#> 12 license                     <NA>                                             
#> 13 usage_terms                "No known copyright restrictions"                 
#> 14 attribution_required       "0"                                               
#> 15 copyrighted                "1"                                               
#> 16 restrictions               ""                                                
#> 17 date_time                  "2019-07-02 03:33:00"                             
#> 18 date_time_original         "2011-07-18 16:02"                                
#> 19 commons_metadata_extension "1.2"
```

This function does not currently cache data.

## How caching works

`tidywikidatar` tries to reduce load on Wikidata’s server and speeding
up re-processing of scripts by caching data locally in sqlite databases.
They are stored locally in the folder defined by
`tw_set_cache_folder()` - by default, in the current working directory -
when cache is enabled (typically, with `tw_enable_cache()` at the
beginning of a session).

To reduce the size of local files, if data are requested in a specific
language, then only data in that language are stored locally.

The easiest way to reset the cache is simply to delete the cache folder.

Results are stored in different databases by language, and function
used; `tw_search()`, `tw_get()`, and `tw_get_qualifiers()`, for example,
store data in different files.

`tw_query()` is never cached.

See the the dedicated vignette for more details on caching:
`vignette("caching")`.

## Requirements and installation issues

Fedora users may need to install the package `libjpeg-turbo-devel`,
which is required by one of the packages that `tidywikidatar` relies on,
as well as some of the database drivers that can be used for caching,
such as `unixODBC-devel`, and `mysql-devel`, `mysql-connector-odbc`.

## Copyright and credits

This package has been created by [Giorgio
Comai](https://giorgiocomai.eu), data analyst and researcher at
[OBCT/CCI](https://www.balcanicaucaso.org/), within the scope of
[EDJNet](https://www.europeandatajournalism.eu/), the European Data
Journalism Network.

It is distributed under the MIT license.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(tidywikidatar)

## -----------------------------------------------------------------------------
tw_enable_cache()

## ----eval=FALSE---------------------------------------------------------------
# tw_set_cache_folder(path = fs::path(
#   fs::path_home_r(),
#   "R",
#   "tw_data"
# ))
# tw_create_cache_folder()

## ----eval = FALSE-------------------------------------------------------------
# tw_enable_cache(SQLite = FALSE)
# tw_set_cache_db(
#   driver = "MySQL",
#   host = "localhost",
#   port = 3306,
#   database = "tidywikidatar",
#   user = "secret_username",
#   pwd = "secret_password"
# )
# 
# 
# # for testing, consider running a local database e.g. with:
# # docker run --name tidywikidatar_db -p 3306:3306 -e MYSQL_ROOT_PASSWORD=secret_root_password -e MYSQL_USER=secret_username -e MYSQL_PASSWORD=secret_password -e MYSQL_DATABASE=tidywikidatar mysql:latest

## -----------------------------------------------------------------------------
tw_get_cache_table_name(type = "item", language = "en")

## ----eval=FALSE---------------------------------------------------------------
# db <- tw_connect_to_cache()
# 
# 
# tables_v <- DBI::dbListTables(conn = db)
# 
# # for search cache tables
# purrr::walk(
#   .x = tables_v[stringr::str_starts(string = tables_v, "tw_search_item")],
#   .f = function(x) {
#     tw_index_cache_search(table_name = x)
#   }
# )
# 
# # for item cache tables
# purrr::walk(
#   .x = tables_v[stringr::str_starts(string = tables_v, "tw_item")],
#   .f = function(x) {
#     tw_index_cache_item(table_name = x)
#   }
# )

#' Writes item to cache
#'
#' Writes item to cache. Typically used internally, but exported to enable
#' custom caching solutions.
#'
#' @param item_df A data frame with three columns typically generated with
#'   [tw_get()].
#' @inheritParams tw_get
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#'
#' tw_set_cache_folder(path = fs::path(tempdir(), paste(sample(letters, 24), collapse = "")))
#' tw_create_cache_folder(ask = FALSE)
#' tw_disable_cache()
#'
#' df_from_api <- tw_get(id = "Q180099", language = "en")
#'
#' df_from_cache <- tw_get_cached_item(
#'   id = "Q180099",
#'   language = "en"
#' )
#'
#' is.null(df_from_cache) # expect TRUE, as nothing has yet been stored in cache
#'
#' tw_write_item_to_cache(
#'   item_df = df_from_api,
#'   language = "en",
#'   cache = TRUE
#' )
#'
#' df_from_cache <- tw_get_cached_item(
#'   id = "Q180099",
#'   language = "en",
#'   cache = TRUE
#' )
#'
#' is.null(df_from_cache) # expect a data frame, same as df_from_api
tw_write_item_to_cache <- function(
  item_df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "item",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE id = {id*}",
        id = unique(item_df$id),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(db, name = table_name, value = item_df, append = TRUE)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

#' Set language to be used by all functions
#'
#' Defaults to "en".
#'
#' @param language A character vector of length one, with a string of two letters such as "en". For a full list of available values, see: https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all
#'
#' @return A two letter code for the language, if previously set; the same language as given to the function; or the default, `en` is none is given.
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_set_language(language = "en")
#' }
#' }
#'
tw_set_language <- function(language = NULL) {
  if (is.null(language)) {
    language <- Sys.getenv("tw_language")
  } else {
    Sys.setenv(tw_language = language)
  }
  if (language == "") {
    language <- "en"
  }
  invisible(language)
}

#' @rdname tw_set_language
#' @examples
#' tw_get_language()
#' @export
tw_get_language <- tw_set_language

#' Gets all details of a property
#'
#' Used internally. Users should rely on [tw_get_property_with_details()].
#'
#' @inheritParams tw_get_property_with_details
#'
#' @return A tibble, corresponding to the details for the given property. `NULL`
#'   if no relevant property found.
#'
#' @examples
#' # Get "female form of label", including language
#' \dontrun{
#'   tidywikidatar:::tw_get_property_with_details_single(id = "Q64733534", p = "P2521")
#' }
tw_get_property_with_details_single <- function(
  id,
  p,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  item <- tryCatch(
    tw_get_item(id = id, retry = retry, user_agent = user_agent),
    error = function(e) {
      as.character(e[[1]])
    }
  )

  if (is.character(item)) {
    cli::cli_alert_danger(item)
    return(NULL)
  }

  if (
    is.element(
      el = "redirect",
      set = item %>%
        purrr::pluck(1) %>%
        names()
    )
  ) {
    id <- item %>%
      purrr::pluck(1, "redirect")
    item <- tryCatch(
      tw_get_item(id = id, retry = retry, user_agent = user_agent),
      error = function(e) {
        as.character(e[[1]])
      }
    )
  }

  claims <- item %>% purrr::pluck(1, "claims")

  current_claim_l <- claims[names(claims) == p]

  value_pre <- current_claim_l[[unique(p)]][["mainsnak"]][["datavalue"]][[
    "value"
  ]]

  value_pre %>%
    tibble::as_tibble() %>%
    dplyr::mutate(id = id, p = p, .before = 0)
}


#' Gets all details of a property for one or more Wikidata items.
#'
#' @inheritParams tw_get
#' @inheritParams tw_get_property
#'
#' @return A tibble, corresponding to the details for the given property. `NULL`
#'   if no relevant property found.
#' @export
#'
#' @examples
#' # Get "female form of label", including language
#' tw_get_property_with_details(id = "Q64733534", p = "P2521")
tw_get_property_with_details <- function(
  id,
  p,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  pb <- progress::progress_bar$new(total = length(unique(id)))
  purrr::map2_dfr(
    .x = unique(id),
    .y = p,
    .f = function(id, p) {
      pb$tick()
      Sys.sleep(wait)
      tw_get_property_with_details_single(
        id = id,
        p = p,
        retry = retry,
        user_agent = user_agent
      )
    }
  )
}

#' Retrieve cached item
#'
#' @inheritParams tw_get
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#'
#' tw_set_cache_folder(path = tempdir())
#' tw_enable_cache()
#' tw_create_cache_folder(ask = FALSE)
#'
#' df_from_api <- tw_get(id = "Q180099", language = "en")
#'
#' df_from_cache <- tw_get_cached_item(
#'   id = "Q180099",
#'   language = "en"
#' )
tw_get_cached_item <- function(
  id,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(tidywikidatar::tw_empty_item)
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "item",
    language = language
  )

  if (!pool::dbExistsTable(conn = db, name = table_name)) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )

    return(tidywikidatar::tw_empty_item)
  }

  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(.data$id %in% !!stringr::str_to_upper(id)),
    error = function(e) {
      logical(1L)
    }
  )

  if (isFALSE(db_result)) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
    return(tidywikidatar::tw_empty_item)
  } else {
    if (
      isFALSE(identical(
        colnames(tidywikidatar::tw_empty_item),
        colnames(db_result)
      ))
    ) {
      cli::cli_abort(c(
        x = "The cache has been generated with a previous version of `tidywikidatar` that is not compatible with the current version.",
        i = "You may want to delete the old cache or reset just this table with {.fn tw_reset_item_cache}"
      ))
    }
  }

  cached_items_df <- db_result %>%
    dplyr::collect()

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  cached_items_df
}


#' Gets location of cache file
#'
#' @param extension Defaults to `sqlite`. Extension of the database cache file.
#' @inheritParams tw_get
#'
#' @return A character vector of length one with location of item cache file.
#' @export
#'
#' @examples
#'
#' tw_set_cache_folder(path = tempdir())
#' sqlite_cache_file_location <- tw_get_cache_file() # outputs location of cache file
tw_get_cache_file <- function(
  extension = "sqlite",
  language = tidywikidatar::tw_get_language()
) {
  fs::path(
    tw_get_cache_folder(),
    stringr::str_c(
      "tw_cache_",
      language
    ),
    ext = extension
  )
}

#' Gets name of table inside the database
#'
#' @param type Defaults to "item". Type of cache file to output. Values
#'   typically used by `tidywikidatar` include "item", "search_item",
#'   "search_property", and "qualifier".
#' @inheritParams tw_search
#' @return A character vector of length one with the name of the relevant table
#'   in the cache file.
#' @export
#'
#' @examples
#' # outputs name of table used in the cache database
#' tw_get_cache_table_name(type = "item", language = "en")
tw_get_cache_table_name <- function(
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language()
) {
  if (stringr::str_starts(string = type, pattern = "search")) {
    language <- stringr::str_c(language, "_", response_language)
  }
  stringr::str_c("tw_", type, "_", language)
}

#' Check if given items are present in cache
#'
#' @inheritParams tw_get
#'
#' @return A character vector with IDs of items present in cache. If no item
#'   found in cache, returns `NULL`.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_set_cache_folder(path = tempdir())
#'   tw_enable_cache()
#'   tw_create_cache_folder(ask = FALSE)
#'
#'   # add three items to local cache
#'   invisible(tw_get(id = "Q180099", language = "en"))
#'   invisible(tw_get(id = "Q228822", language = "en"))
#'   invisible(tw_get(id = "Q184992", language = "en"))
#'
#'   # check if these other items are in cache
#'   items_in_cache <- tw_check_cached_items(
#'     id = c(
#'       "Q180099",
#'       "Q228822",
#'       "Q76857"
#'     ),
#'     language = "en"
#'   )
#'   # it should return only the two items from the current list of id
#'   # but not other item already in cache
#'   items_in_cache
#' }
tw_check_cached_items <- function(
  id,
  language = tidywikidatar::tw_get_language(),
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  tw_get_cached_item(
    id = id,
    language = language,
    cache_connection = cache_connection,
    disconnect_db = disconnect_db
  ) %>%
    dplyr::distinct(id) %>%
    dplyr::pull(id)
}

#' Perform simple Wikidata queries
#'
#' This function aims to facilitate only the most basic type of queries: return
#' which items have the following property pairs. For more details on Wikidata
#' queries, the
#' \href{https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service/queries/examples}{examples
#' in the official documentation}.
#'
#' Consider [tw_get_all_with_p()] if you want to get all items with a given
#' property, irrespective of the value.
#'
#' @param query A list of named vectors, or a data frame (see example and
#'   readme).
#' @param fields A character vector of Wikidata fields. Ignored if
#' `return_as_tw_search` is set to `TRUE` (as per default). Defaults to
#' `("item", "itemLabel", "itemDescription")`
#' @param language Defaults to language set with [tw_set_language()]; if not
#'   set, "en". If more than one, can be set in order of preference, e.g.
#'   `c("it", "fr", "en")`. Use "all_available" to keep all languages. For
#'   available language values, see the
#'   \href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{list
#'   of available language codes in the official documentation}.
#' @param return_as_tw_search Logical, defaults to `TRUE`. If `TRUE`, returns a
#'   data frame with three columns (`id`, `label`, and `description`) that can
#'   be piped to other `tw_` functions. If `FALSE`, a data frame with as many
#'   columns as fields.
#' @param user_agent Defaults to a combination of `tidywikidatar` and package
#'   version number. Consider customising it for the current sessions with
#'   [tw_set_user_agent()], in particular if you are making many queries.
#'
#' @return A data frame
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'   query <- list(
#'     c(p = "P106", q = "Q1397808"),
#'     c(p = "P21", q = "Q6581072")
#'   )
#'   tw_query(query)
#' }
tw_query <- function(
  query,
  fields = c("item", "itemLabel", "itemDescription"),
  language = tidywikidatar::tw_get_language(),
  return_as_tw_search = TRUE,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (!is.data.frame(query)) {
    query_df <- dplyr::bind_rows(query)
  } else {
    query_df <- query
  }

  query_t <- query_df %>%
    glue::glue_data("wdt:{p} wd:{q}") %>%
    stringr::str_c(collapse = ";\n")

  sparql_t <- glue::glue(
    "SELECT
             {stringr::str_c(\"?\", stringr::str_c(fields, collapse = \" ?\"))}
             WHERE{{?item {query_t} .
             SERVICE wikibase:label {{ bd:serviceParam wikibase:language '{stringr::str_c(language, collapse = ',')},[AUTO_LANGUAGE]' . }}
             }}"
  )

  req <- httr2::request("https://query.wikidata.org/sparql") %>%
    httr2::req_headers(Accept = "text/csv") %>%
    httr2::req_user_agent(user_agent) %>%
    httr2::req_url_query(query = sparql_t) %>%
    httr2::req_retry(max_tries = 5, backoff = ~2)

  resp <- httr2::req_perform(req)

  response <- read.csv(
    text = httr2::resp_body_string(resp),
    stringsAsFactors = FALSE
  ) %>%
    tibble::as_tibble()

  if (return_as_tw_search) {
    response %>%
      dplyr::transmute(
        id = stringr::str_extract(.data$item, pattern = "Q[[:digit:]]+$"),
        label = .data$itemLabel,
        description = .data$itemDescription
      )
  } else {
    tibble::as_tibble(response)
  }
}

#' Return (most) information from a Wikidata item in a tidy format from a single
#' Wikidata identifier
#'
#' @param read_cache Logical, defaults to TRUE. Mostly used internally to
#'   prevent checking if an item is in cache if it is already known that it is
#'   not in cache.
#'
#' @inheritParams tw_get
#'
#' @return A data.frame (a tibble) with four columns (id, property, value, and
#'   rank). If item not found or trouble connecting with the server, a data
#'   frame with four columns and zero rows is returned, with the warning as an
#'   attribute, which can be retrieved with `attr(output, "warning"))`
#'
#' @examples
#' if (interactive()) {
#'   tidywikidatar:::tw_get_single(
#'     id = "Q180099",
#'     language = "en"
#'   )
#' }
#'
#' ## using `tw_test_items` in examples in order to show output without calling
#' ## on Wikidata servers
#'
#' tidywikidatar:::tw_get_single(
#'   id = "Q180099",
#'   language = "en",
#'   id_l = tw_test_items
#' )
tw_get_single <- function(
  id,
  language = tidywikidatar::tw_get_language(),
  retry = 10,
  cache = NULL,
  overwrite_cache = FALSE,
  read_cache = TRUE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  id_l = NULL,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  id <- tw_check_qid(id)

  if (length(id) > 1) {
    cli::cli_abort(c(
      x = "`id` must have length 1.",
      i = "Consider using {.fn tw_get} for multiple inputs."
    ))
  } else if (length(id) == 0) {
    return(tidywikidatar::tw_empty_item)
  }

  if (tw_check_cache(cache)) {
    db <- tw_connect_to_cache(
      connection = cache_connection,
      language = language,
      cache = cache
    )
  }

  if (
    tw_check_cache(cache) &&
      !overwrite_cache &&
      read_cache
  ) {
    db_result <- tw_get_cached_item(
      id = id,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
      return(
        db_result %>%
          tibble::as_tibble()
      )
    }
  }

  Sys.sleep(time = wait)

  if (!is.null(id_l)) {
    item <- id_l %>%
      purrr::pluck(id) %>%
      list()

    if (length(item) == 0) {
      item <- tryCatch(
        tw_get_item(id = id, retry = retry, user_agent = user_agent),
        error = function(e) {
          as.character(e[[1]])
        }
      )
    } else if (length(item) > 1) {
      item <- item[1]
    }
  } else {
    item <- tryCatch(
      tw_get_item(id = id, retry = retry, user_agent = user_agent),
      error = function(e) {
        as.character(e[[1]])
      }
    )
  }

  if (is.character(item)) {
    cli::cli_alert_danger(item)
    output <- tibble::tibble(
      id = as.character(id),
      property = "error",
      value = stringr::str_remove(
        string = item,
        pattern = "The API returned an error: "
      ),
      rank = NA_character_
    )

    if (tw_check_cache(cache)) {
      tw_write_item_to_cache(
        item_df = output,
        language = language,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = db,
        disconnect_db = disconnect_db
      )
    }

    attr(output, "warning") <- item
    return(output)
  }

  if (
    is.element(
      el = "redirect",
      set = item %>%
        purrr::pluck(1) %>%
        names()
    )
  ) {
    id <- item %>%
      purrr::pluck(1, "redirect")
    return(
      tw_get(
        id = id,
        retry = retry,
        user_agent = user_agent,
        language = language,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        wait = wait
      )
    )
  }

  everything_df <- tw_extract_single(
    w = item,
    language = language
  )

  # keep one row, otherwise nothing remains in cache, and it will query
  # each time the script is re-run
  if (nrow(everything_df) == 0) {
    everything_df <- tibble::tibble(
      id = id,
      property = stringr::str_c("label_", language),
      value = NA_character_,
      rank = NA_character_
    )
  }

  if (tw_check_cache(cache)) {
    tw_write_item_to_cache(
      item_df = everything_df,
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
  }
  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
  everything_df
}

#' Return (most) information from a Wikidata item in a tidy format
#'
#' @param id A character vector, must start with Q, e.g. "Q180099" for the
#'   anthropologist Margaret Mead. Can also be a data frame of one row,
#'   typically generated with [tw_search()] or a combination of [tw_search()]
#'   and [tw_filter_first()].
#' @param language Defaults to language set with [tw_set_language()]; if not
#'   set, "en". Use "all_available" to keep all languages. For available
#'   language values, see
#'   \href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
#'   dedicated Wikimedia page}.
#' @param cache Defaults to `NULL`. If given, it should be given either `TRUE`
#'   or `FALSE`. Typically set with [tw_enable_cache()] or [tw_disable_cache()].
#' @param overwrite_cache Logical, defaults to `FALSE`. If `TRUE`, it overwrites
#'   the table in the local sqlite database. Useful if the original Wikidata
#'   object has been updated.
#' @param cache_connection Defaults to `NULL`. If `NULL`, and caching is
#'   enabled, `tidywikidatar` will use a local sqlite database. A custom
#'   connection to other databases can be given (see vignette `caching` for
#'   details).
#' @param disconnect_db Defaults to `TRUE`. If `FALSE`, leaves the connection to
#'   cache open.
#' @param wait In seconds, defaults to 0. Time to wait between queries to
#'   Wikidata. If data are cached locally, wait time is not applied. If you are
#'   running many queries systematically you may want to add some waiting time
#'   between queries.
#' @param retry Defaults to 10. Maximum number of times to retry if the API
#'   throws an error, such as "too many requests". Each time, it will wait as
#'   much time as requested by the API. Notice that this can be a long time,
#'   e.g. 30 minutes. Set to `FALSE` if you prefer the API to throw an error
#'   immediately. Consider adjusting the `wait` parameter, or customising the
#'   `user_agent` if relevant.
#' @param id_l Defaults to `NULL`. If given, must be an object or list such as
#'   the one generated with [tw_get_item()]. If given, and the requested
#'   id is actually present in `id_l`, then no query to Wikidata servers is
#'   made.
#'
#' @inheritParams tw_set_user_agent
#'
#' @return A data.frame (a tibble) with three columns (`id`, `property`, and
#'   `value`).
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'   tw_get(
#'     id = c("Q180099", "Q228822"),
#'     language = "en"
#'   )
#' }
#'
#' ## using `tw_test_items` in examples in order to show output without calling
#' ## on Wikidata servers
#'
#' tw_get(
#'   id = c("Q180099", "Q228822"),
#'   language = "en",
#'   id_l = tw_test_items
#' )
tw_get <- function(
  id,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  id_l = NULL,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  unique_id <- tw_check_qid(id)

  if (length(unique_id) == 0) {
    return(tidywikidatar::tw_empty_item)
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (length(unique_id) == 1) {
    return(
      dplyr::left_join(
        x = tibble::tibble(id = id),
        y = tw_get_single(
          id = unique_id,
          language = language,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          wait = wait,
          retry = retry,
          id_l = id_l,
          user_agent = user_agent
        ),
        by = "id"
      )
    )
  } else if (length(unique_id) > 1) {
    if (overwrite_cache | !tw_check_cache(cache)) {
      pb <- progress::progress_bar$new(total = length(unique_id))

      item_df <- purrr::map_dfr(
        .x = unique_id,
        .f = function(x) {
          pb$tick()
          tw_get_single(
            id = x,
            language = language,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            wait = wait,
            retry = retry,
            id_l = id_l,
            user_agent = user_agent
          )
        }
      )
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
      return(
        dplyr::left_join(
          x = tibble::tibble(id = id),
          y = item_df,
          by = "id",
          relationship = "many-to-many"
        )
      )
    }

    if (!overwrite_cache & tw_check_cache(cache)) {
      items_from_cache_df <- tw_get_cached_item(
        id = unique_id,
        language = language,
        cache = cache,
        cache_connection = db,
        disconnect_db = FALSE
      )

      id_items_not_in_cache <- unique_id[
        !is.element(unique_id, items_from_cache_df$id)
      ]

      if (length(id_items_not_in_cache) == 0) {
        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language
        )
        return(
          dplyr::left_join(
            x = tibble::tibble(id = id),
            y = items_from_cache_df,
            by = "id",
            relationship = "many-to-many"
          )
        )
      } else if (length(id_items_not_in_cache) > 0) {
        pb <- progress::progress_bar$new(total = length(id_items_not_in_cache))

        items_not_in_cache_df <- purrr::map_dfr(
          .x = id_items_not_in_cache,
          .f = function(x) {
            pb$tick()
            tw_get_single(
              id = x,
              language = language,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              read_cache = FALSE,
              disconnect_db = FALSE,
              wait = wait,
              retry = retry,
              id_l = id_l,
              user_agent = user_agent
            )
          }
        )

        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language
        )

        dplyr::left_join(
          x = tibble::tibble(id = id),
          y = dplyr::bind_rows(
            items_from_cache_df,
            items_not_in_cache_df
          ),
          by = "id",
          relationship = "many-to-many"
        ) %>%
          dplyr::filter(!is.na(.data$id))
      }
    }
  }
}


#' Reset item cache
#'
#' Removes the table where items are cached
#'
#' @param ask Logical, defaults to `TRUE`. If `FALSE`, and cache folder does not
#'   exist, it just removes it without asking (useful for non-interactive
#'   sessions).
#' @inheritParams tw_get
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_reset_item_cache()
#' }
tw_reset_item_cache <- function(
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "item",
    language = language
  )

  if (!pool::dbExistsTable(conn = db, name = table_name)) {
    # do nothing: if table does not exist, nothing to delete
  } else if (isFALSE(ask)) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(c(
      "Item cache reset for language ",
      sQuote(language),
      " completed"
    ))
  } else if (
    utils::menu(
      c("Yes", "No"),
      title = paste0(
        "Are you sure you want to remove from cache the items table for language: ",
        sQuote(language),
        "?"
      )
    ) ==
      1
  ) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(c(
      "Items cache reset for language ",
      sQuote(language),
      " completed"
    ))
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

library("testthat")
library("tidywikidatar")
test_that("check if tw_get returns tibble with four columns and meaningful number of rows", {
  testthat::skip_if_offline()

  expect_true(object = {
    item <- tw_get(
      id = c(
        "Q180099"
      ),
      id_l = tw_test_items
    )

    if (!is.null(attr(item, "warning"))) {
      message("Issues with API in testing")
      test_result <- TRUE
    } else {
      test_result <- Reduce(
        f = `|`,
        x = c(
          ncol(item) == 4,
          nrow(item) > 100
        )
      )
    }
    test_result
  })
})


test_that("check if tw_get works when more than one id as input", {
  testthat::skip_if_offline()

  expect_true(object = {
    item <- tw_get(
      id = c(
        "Q180099",
        "Q228822"
      ),
      language = "en",
      id_l = tw_test_items
    )
    if (is.null(attr(item, "warning")) == FALSE) {
      message("Issues with API in testing")
      test_result <- TRUE
    } else {
      test_result <- Reduce(
        f = `|`,
        x = c(
          ncol(item) == 3,
          nrow(item) > 200
        ),
        length(unique(item$id) == 2)
      )
    }
    test_result
  })
})

#' Set language to be used by all functions
#'
#' Defaults to current package name (`tidywikidatar`) and version.
#'
#' @param user_agent Defaults to `NULL`. If not given, implicitly defaults to
#'   current package name (`tidywikidatar`) and version.
#'
#' @return The user agent set for the session, implicitly.
#' @export
#' @examples
#' # Default user agent
#' default_user_agent <- tw_get_user_agent()
#' default_user_agent
#' # Custom user agent
#' tw_set_user_agent(user_agent = "custom_project_name/email")
#' new_user_agent <- tw_get_user_agent()
#' new_user_agent
#' # Restore
#' tw_set_user_agent(user_agent = default_user_agent)
tw_set_user_agent <- function(user_agent = NULL) {
  if (is.null(user_agent)) {
    user_agent <- Sys.getenv("tw_user_agent")
  } else {
    Sys.setenv(tw_user_agent = user_agent)
  }
  if (user_agent == "") {
    user_agent <- stringr::str_flatten(c(
      "tidywikidatar/",
      as.character(packageVersion("tidywikidatar"))
    ))
  }
  invisible(user_agent)
}

#' @rdname tw_set_user_agent
#' @examples
#' tw_get_user_agent()
#' @export
tw_get_user_agent <- tw_set_user_agent

#' Checks if an input is a search; if not, it tries to return a search
#'
#' Mostly used as a convenience function inside other functions to have
#' consistent inputs.
#'
#' @inheritParams tw_search
#'
#' @return A data frame with three columns, `id`, `label`, and `description`,
#'   filtered by the above criteria. Four columns if `include_search` is set to
#'   `TRUE`.
#' @export
#'
#' @examples
#' # The following two lines should give the same result.
#'
#' \dontrun{
#'   tw_check_search("Sylvia Pankhurst")
#'   tw_check_search(tw_search("Sylvia Pankhurst"))
#' }
tw_check_search <- function(
  search,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  wait = 0,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (is.data.frame(search)) {
    return(search)
  } else if (length(search) > 1) {
    cli::cli_abort(c(
      x = "{.var search} must be a single query or a data frame generate by {.fn tw_search}, not {.obj_type_friendly {search}}."
    ))
  } else if (is.character(search)) {
    search_result <- tw_search(
      search = search,
      type = type,
      language = language,
      response_language = response_language,
      limit = limit,
      include_search = include_search,
      wait = wait,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      disconnect_db = disconnect_db
    )
  }
  search_result
}

#' Get Wikidata property of one or more items as a tidy data frame
#'
#' @param p A character vector, a property. Must always start with the capital
#'   letter "P", e.g. "P31" for "instance of".
#' @param id_df Default to `NULL`. If given, it should be a dataframe typically
#'   generated with [tw_get()], and is used instead of calling Wikidata or
#'   using SQLite cache. Ignored when `id` is of length more than one.
#' @inheritParams tw_get
#' @inheritParams tw_search
#'
#' @return A tibble, corresponding to the value for the given property. A tibble of zero rows if no relevant property found.
#' @export
#'
#' @examples
#' \dontrun{
#' # Who were the doctoral advisors - P184 - of Margaret Mead - Q180099?
#' advisors <- tw_get_property(id = "Q180099", p = "P184")
#' advisors
#'
#' # tw_get_label(advisors)
#'
#' # It is also possible to get one property for many id
#'
#'   tw_get_property(
#'     id = c(
#'       "Q180099",
#'       "Q228822"
#'     ),
#'     p = "P31"
#'   )
#'
#'   # Or many properties for a single id
#'
#'   tw_get_property(
#'     id = "Q180099",
#'     p = c("P21", "P31")
#'   )
#' }
tw_get_property <- function(
  id,
  p,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  unique_id <- tw_check_qid(id)

  if (length(unique_id) == 0) {
    return(NULL)
  }

  if (is.null(id_df)) {
    id_df <- tw_get(
      id = unique_id,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      language = language,
      wait = wait,
      disconnect_db = disconnect_db,
      retry = retry,
      user_agent = user_agent
    )
  } else {
    id_df <- id_df %>%
      dplyr::filter(.data$id %in% unique_id)

    missing_id_v <- unique_id[!unique_id %in% id_df$id]

    if (length(missing_id_v) > 0) {
      id_df <- dplyr::bind_rows(
        id_df,
        tw_get(
          id = missing_id_v,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = cache_connection,
          language = language,
          wait = wait,
          disconnect_db = disconnect_db,
          retry = retry,
          user_agent = user_agent
        )
      )
    }
  }

  property_df <- id_df %>%
    dplyr::filter(.data$property %in% p)
  if (nrow(property_df) == 0) {
    return(tidywikidatar::tw_empty_item)
  } else {
    if (length(p) > 1) {
      property_df <- tibble::tibble(property = p) %>%
        dplyr::left_join(y = property_df, by = "property") %>%
        dplyr::select("id", "property", "value", "rank")
    }
    if (length(id) > 1) {
      property_df <- tibble::tibble(id = id) %>%
        dplyr::left_join(
          y = property_df,
          by = "id"
        )
    }

    property_df
  }
}


#' Get Wikidata property of an item as a vector or list of the same length as
#' input
#'
#' @param only_first Logical, defaults to `FALSE`. If `TRUE`, it just keeps the
#'   first relevant property value for each id (or `NA` if none is available),
#'   and returns a character vector. Warning: this likely discards valid values,
#'   so make sure this is really what you want. If `FALSE`, returns a list of
#'   the same length as input, with all values for each id stored in a list if
#'   more than one is found.
#' @param preferred Logical, defaults to `FALSE`. If `TRUE`, returns properties
#'   that have rank "preferred" if available; if no "preferred" property is
#' found, then it is ignored.
#' @param latest_start_time Logical, defaults to `FALSE`. If `TRUE`, returns the
#'   property that has the most recent start time ("P580") as qualifier if
#'   `only_first` is set to `TRUE`, or returns a list ordered by start time if
#'   `only_first` is set to `FALSE`. If no such qualifier is found, then it is
#'   ignored.
#' @inheritParams tw_get
#' @inheritParams tw_get_property
#'
#' @return A list of the same length as input (or a character vector is
#'   only_first is set to `TRUE`)
#' @export
#'
#' @examples
#'
#'
#' # By default, it returns a list of the same length as input,
#' # no matter how many values for each id/property
#'
#'
#' \dontrun{
#'   tw_get_property_same_length(
#'     id = c(
#'       "Q180099",
#'       "Q228822",
#'       "Q76857"
#'     ),
#'     p = "P26"
#'   )
#'   # Notice that if no relevant match is found, it returns a NA
#'   # This is useful for piped operations
#'
#'   tibble::tibble(id = c(
#'     "Q180099",
#'     "Q228822",
#'     "Q76857"
#'   )) %>%
#'     dplyr::mutate(spouse = tw_get_property_same_length(id, "P26"))
#'
#'   # Consider unnesting for further analysis
#'
#'   tibble::tibble(id = c(
#'     "Q180099",
#'     "Q228822",
#'     "Q76857"
#'   )) %>%
#'     dplyr::mutate(spouse = tw_get_property_same_length(id, "P26")) %>%
#'     tidyr::unnest(cols = spouse)
#'
#'   # If you are sure that you are interested only in the first return value,
#'   # consider setting only_first=TRUE to get a character vector rather than a list
#'   # Be mindful: you may well be discarding valid values.
#'   tibble::tibble(id = c(
#'     "Q180099",
#'     "Q228822",
#'     "Q76857"
#'   )) %>%
#'     dplyr::mutate(spouse = tw_get_property_same_length(id, "P26",
#'       only_first = TRUE
#'     ))
#' }
tw_get_property_same_length <- function(
  id,
  p,
  only_first = FALSE,
  preferred = FALSE,
  latest_start_time = FALSE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  property_df <- tw_get_property(
    id = id,
    p = p,
    language = language,
    id_df = id_df,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    retry = retry,
    user_agent = user_agent
  )

  if (is.null(property_df)) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
    return(rep(as.character(NA), length(id)))
  } else if (nrow(property_df) == 0) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
    if (only_first) {
      return(rep(NA_character_, length(id)))
    } else {
      return(
        rep(NA_character_, length(id)) %>%
          as.list()
      )
    }
  }

  if (!preferred & !latest_start_time) {
    # do nothing, and keep the retrieved property_df
  }
  if (preferred & !latest_start_time) {
    preferred_df <- property_df %>%
      dplyr::mutate(
        rank = factor(
          rank,
          levels = c(
            "preferred",
            "normal",
            "deprecated"
          )
        )
      ) %>%
      dplyr::group_by(id) %>%
      dplyr::arrange(.by_group = TRUE, .data$rank) %>%
      dplyr::ungroup()

    if (nrow(preferred_df) > 0) {
      property_df <- preferred_df
    }
  } else if (latest_start_time) {
    properties_all_df <- property_df

    qualifiers_df <- tw_get_qualifiers(
      id = id,
      p = p,
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait
    )

    properties_without_qualifier_df <- properties_all_df %>%
      dplyr::rename(qualifier_id = "value") %>%
      dplyr::anti_join(
        y = qualifiers_df,
        by = c("id", "property", "qualifier_id")
      ) %>%
      dplyr::mutate(
        qualifier_property = "P580",
        qualifier_value = NA_character_
      )

    qualifiers_with_start_time_df <- dplyr::bind_rows(
      qualifiers_df,
      properties_without_qualifier_df
    ) %>%
      dplyr::filter(.data$qualifier_property == "P580") %>%
      dplyr::distinct(
        .data$id,
        .data$qualifier_id,
        .data$qualifier_value,
        .keep_all = TRUE
      ) %>%
      dplyr::mutate(
        year = stringr::str_extract(
          string = .data$qualifier_value,
          pattern = "[[:print:]][[:digit:]]{4}"
        ) %>%
          as.numeric(),
        month = stringr::str_extract(
          string = .data$qualifier_value,
          pattern = "-[[:digit:]]{2}-"
        ) %>%
          stringr::str_remove_all("-") %>%
          as.numeric(),
        day = stringr::str_extract(
          string = .data$qualifier_value,
          pattern = "-[[:digit:]]{2}T"
        ) %>%
          stringr::str_remove("T") %>%
          stringr::str_remove("-") %>%
          as.numeric()
      )

    if (preferred) {
      qualifiers_latest_start_time_df <- qualifiers_with_start_time_df %>%
        dplyr::mutate(
          rank = factor(
            rank,
            levels = c(
              "preferred",
              "normal",
              "deprecated"
            )
          )
        ) %>%
        dplyr::group_by(.data$id) %>%
        dplyr::arrange(
          .data$id,
          .data$rank,
          dplyr::desc(.data$year),
          dplyr::desc(.data$month),
          dplyr::desc(.data$day),
          dplyr::desc(.data$qualifier_value)
        ) %>%
        dplyr::select(-"year", -"month", -"day") %>%
        dplyr::ungroup() %>%
        dplyr::transmute(
          .data$id,
          .data$property,
          value = .data$qualifier_id,
          .data$rank
        )
    } else {
      qualifiers_latest_start_time_df <- qualifiers_with_start_time_df %>%
        dplyr::group_by(.data$id) %>%
        dplyr::arrange(
          .data$id,
          dplyr::desc(.data$year),
          dplyr::desc(.data$month),
          dplyr::desc(.data$day),
          dplyr::desc(.data$qualifier_value)
        ) %>%
        dplyr::select(-"year", -"month", -"day") %>%
        dplyr::ungroup() %>%
        dplyr::transmute(
          .data$id,
          .data$property,
          value = .data$qualifier_id,
          .data$rank
        )
    }

    property_df <- property_df %>%
      dplyr::distinct(.data$id, .data$property) %>%
      dplyr::right_join(
        y = qualifiers_latest_start_time_df,
        by = c("id", "property")
      )
  }

  if (only_first) {
    property_df_post <- property_df %>%
      dplyr::distinct(.data$id, .keep_all = TRUE)
  } else {
    property_df_post <- property_df %>%
      dplyr::distinct(.data$id, .data$value) %>%
      dplyr::group_by(.data$id) %>%
      dplyr::summarise(value = list(.data$value)) %>%
      dplyr::ungroup()
  }

  property_df_out <- tibble::tibble(id = id) %>%
    dplyr::left_join(
      y = property_df_post,
      by = "id"
    )

  if (is.list(property_df_out$value) == "list") {
    property_df_out$value[purrr::map_lgl(
      .x = property_df_out$value,
      .f = is.null
    )] <- list(NA_character_)
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  property_df_out %>%
    dplyr::pull("value")
}


#' @rdname tw_get_property_same_length
#' @examples tw_get_p(id = "Q180099", "P26")
#' @export
tw_get_p <- tw_get_property_same_length


#' Get Wikidata property of an item as a character vector of the same length as
#' input
#'
#' This function wraps [tw_get_p()], but always sets `only_first` and
#' `preferred` to `TRUE` in order to give back always a character vector.
#'
#' @inheritParams tw_get_property_same_length
#'
#' @return A character vector of the same length as the input.
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_get_p1(id = "Q180099", "P26")
#' }
tw_get_p1 <- function(
  id,
  p,
  latest_start_time = FALSE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  tw_get_property_same_length(
    id = id,
    p = p,
    only_first = TRUE,
    preferred = TRUE,
    latest_start_time = latest_start_time,
    language = language,
    id_df = id_df,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = cache_connection,
    disconnect_db = disconnect_db,
    wait = wait,
    retry = retry,
    user_agent = user_agent
  )
}

#' Get Wikidata label in given language
#'
#' @param id_df Default to `NULL`. If given, it should be a dataframe typically
#'   generated with [tw_get()], and is used instead of calling Wikidata or
#'   using SQLite cache. Ignored when `id` is of length more than one.
#' @inheritParams tw_get_image
#' @inheritParams tw_get
#'
#' @return A character vector of the same length as the vector of id given, with
#'   the Wikidata label in the requested language.
#' @export
#'
#' @examples
#' \dontrun{
#' tw_get_label(
#'   id = c(
#'     "Q180099",
#'     "Q228822"
#'   ),
#'   language = "en"
#' )
#'
#' # If a label is not available, a NA value is returned
#'   tw_get_label(
#'     id = c(
#'       "Q64733534",
#'       "Q4773904",
#'       "Q220480"
#'     ),
#'     language = "sc"
#'   )
#' }
tw_get_label <- function(
  id,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id[["id"]]
  }

  if (!is.character(id)) {
    cli::cli_abort(c(
      x = "{.var id} must be either a character vector of Wikidata identifiers, or a data frame with an {.var id} column with Q identifiers.",
      i = "Wikidata identifiers are always composed by the the letter {.val Q} followed by digits."
    ))
  }

  if (is.null(id_df)) {
    id_df <- tw_get(
      id = id,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      language = language,
      wait = wait,
      retry = retry,
      user_agent = user_agent,
      disconnect_db = disconnect_db
    )
  } else {
    current_id <- id
    id_df <- id_df %>%
      dplyr::filter(.data$id %in% current_id)

    missing_id_v <- current_id[!current_id %in% id_df$id]

    if (length(missing_id_v) > 0) {
      id_df <- dplyr::bind_rows(
        id_df,
        tw_get(
          id = missing_id_v,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = cache_connection,
          language = language,
          wait = wait,
          retry = retry,
          user_agent = user_agent,
          disconnect_db = disconnect_db
        )
      )
    }
  }

  tw_get_field(
    df = id_df,
    field = "label",
    language = language,
    id = id
  )
}


#' Get Wikidata description in given language
#'
#' @inheritParams tw_get_label
#' @inheritParams tw_get_image
#'
#' @return A character vector of the same length as the vector of id given, with
#'   the Wikidata description in the requested language.
#' @export
#'
#' @examples
#' \dontrun{
#'
#' tw_get_description(
#'   id = c(
#'     "Q180099",
#'     "Q228822"
#'   ),
#'   language = "en"
#' )
#'
#' }
tw_get_description <- function(
  id,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  if (is.null(id_df)) {
    id_df <- tw_get(
      id = id,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      language = language,
      wait = wait,
      retry = retry,
      user_agent = user_agent,
      disconnect_db = disconnect_db
    )
  } else {
    current_id <- id
    id_df <- id_df %>%
      dplyr::filter(.data$id %in% current_id)

    missing_id_v <- current_id[!current_id %in% id_df$id]

    if (length(missing_id_v) > 0) {
      id_df <- dplyr::bind_rows(
        id_df,
        tw_get(
          id = missing_id_v,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = cache_connection,
          language = language,
          wait = wait,
          retry = retry,
          user_agent = user_agent,
          disconnect_db = disconnect_db
        )
      )
    }
  }

  tw_get_field(
    df = id_df,
    field = "description",
    language = language,
    id = id
  )
}

#' Get URL to a Wikipedia article corresponding to a Wikidata Q identifier in
#' given language
#'
#' @param full_link Logical, defaults to `TRUE`. If `FALSE`, returns only the
#'   part of the url that corresponds to the title.
#' @inheritParams tw_get_label
#' @inheritParams tw_get
#'
#' @return A character vector of the same length as the vector of id given, with
#'   the Wikipedia link in the requested language.
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_get_wikipedia(id = "Q180099")
#' }
tw_get_wikipedia <- function(
  id,
  full_link = TRUE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  if (sum(is.na(id)) == length(id)) {
    return(rep(as.character(NA), length(id)))
  }

  if (length(tw_check_qid(id = id)) == 0) {
    return(rep(as.character(NA), length(id)))
  }

  if (is.null(id_df)) {
    id_df <- tw_get(
      id = id,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      disconnect_db = disconnect_db,
      language = language,
      wait = wait,
      retry = retry,
      user_agent = user_agent
    )
  } else {
    current_id <- id
    id_df <- id_df %>%
      dplyr::filter(.data$id %in% current_id)

    missing_id_v <- current_id[!current_id %in% id_df$id]

    if (length(missing_id_v) > 0) {
      id_df <- dplyr::bind_rows(
        id_df,
        tw_get(
          id = missing_id_v,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = cache_connection,
          language = language,
          wait = wait,
          retry = retry,
          user_agent = user_agent,
          disconnect_db = disconnect_db
        )
      )
    }
  }

  base_string <- stringr::str_c("sitelink_", language, "wiki")

  field_df <- id_df %>%
    dplyr::filter(.data$property == base_string) %>%
    dplyr::distinct(.data$id, .keep_all = TRUE)

  if (nrow(field_df) == 0) {
    base_link <- rep(as.character(NA), length(id))
  } else if (nrow(field_df) < length(id)) {
    base_link <- tibble::tibble(id = id) %>%
      dplyr::left_join(y = field_df, by = "id") %>%
      dplyr::pull("value")
  } else {
    base_link <- field_df %>%
      dplyr::pull("value")
  }

  if (length(base_link) == 0) {
    rep(as.character(NA), length(id))
  } else {
    if (full_link) {
      stringr::str_c("https://", language, ".wikipedia.org/wiki/", base_link)
    } else {
      base_link
    }
  }
}

library("testthat")

test_that("check if labels are returned correctly when more than one id given, including NA", {
  testthat::skip_if_offline()

  expect_equal(
    object = {
      tw_get_label(
        id = c(
          "Q180099",
          "Q228822",
          "Not an id",
          "Q180099",
          NA
        ),
        language = "en"
      )
    },
    expected = c(
      "Margaret Mead",
      "Ruth Benedict",
      NA,
      "Margaret Mead",
      NA
    )
  )
})

#' Get description of a Wikidata property in a given language
#'
#' @param property A character vector. Each element must start with P, e.g.
#'   "P31".
#' @inheritParams tw_get
#' @inheritParams tw_search
#'
#' @return A character vector, with the Wikidata description in the requested
#'   language.
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_get_property_description(property = "P31")
#' }
tw_get_property_description <- function(
  property,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
) {
  if (is.data.frame(property)) {
    property <- property$id
  }

  if (length(tw_check_pid(property = property)) == 0) {
    return(rep(NA_character_, length(property)))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (length(property) > 1) {
    if (length(unique(property)) < length(property)) {
      pre_processed <- tibble::tibble(property = property)

      unique_processed <- purrr::map_dfr(
        .x = unique(property),
        .f = function(x) {
          tibble::tibble(
            property = x,
            description = tw_get_property_description_single(
              property = x,
              language = language,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE,
              wait = wait
            ) %>%
              as.character()
          )
        }
      )
      description_output <- pre_processed %>%
        dplyr::left_join(
          y = unique_processed,
          by = "property"
        ) %>%
        dplyr::pull("description")
    } else {
      description_output <- purrr::map_chr(
        .x = property,
        .f = function(x) {
          tw_get_property_description_single(
            property = x,
            language = language,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            wait = wait
          )
        }
      )
    }
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
  } else {
    description_output <- tw_get_property_description_single(
      property = property,
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      wait = wait
    )
  }
  description_output
}

#' Get description of a Wikidata property in a given language
#'
#' @param property A character vector of length 1, must start with P, e.g. "P31".
#' @inheritParams tw_get
#' @inheritParams tw_search
#' @inheritParams tw_get_property
#'
#' @return A character vector of length 1, with the Wikidata description in the
#'   requested language.
#'
#' @examples
#' \dontrun{
#' tidywikidatar:::tw_get_property_description_single(property = "P31")
#' }
tw_get_property_description_single <- function(
  property,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
) {
  if (length(tw_check_pid(property = property)) == 0) {
    return(NA_character_)
  }

  description <- tw_search_property(
    search = property,
    cache = cache,
    language = language,
    response_language = language,
    overwrite_cache = overwrite_cache,
    cache_connection = cache_connection,
    disconnect_db = disconnect_db,
    wait = wait
  ) %>%
    dplyr::filter(.data$id == stringr::str_to_upper(property)) %>%
    dplyr::pull("description")

  if (length(description) == 0) {
    NA_character_
  } else {
    description
  }
}

#' Ensures that input appears to be a valid Wikidata property id (i.e. it starts
#' with P and is followed only by digits)
#'
#' Mostly used internally by other functions.
#'
#' @param property A character vector of one or more Wikidata property
#'   identifiers.
#' @param logical_vector Logical, defaults to `FALSE`. If `TRUE`, returns a
#'   logical vector of the same length as input, where `TRUE` corresponds to
#'   seemingly meaningful property identifiers.
#' @param non_pid_as_NA Logical, defaults to `FALSE`. If `TRUE` (and if
#'   `logical_vector` is set to `FALSE`), a vector of the same length is
#'   returned, with `NA` replacing items that are seemingly not meaningful
#'   property identifiers.
#'
#' @return A character vector with only strings appearing to be Wikidata
#'   identifiers; possibly shorter than input.
#' @export
#'
#' @examples
#'
#' tw_check_pid(property = c("P19", "p20", "Not an property id", "20", NA, "Q5", ""))
#'
#' tw_check_pid(
#'   property = c("P19", "p20", "Not an property id", "20", NA, "Q5", ""),
#'   logical_vector = TRUE
#' )
#'
#' tw_check_pid(
#'   property = c("P19", "p20", "Not an property id", "20", NA, "Q5", ""),
#'   non_pid_as_NA = TRUE
#' )
tw_check_pid <- function(
  property,
  logical_vector = FALSE,
  non_pid_as_NA = FALSE
) {
  if (is.null(property)) {
    return(character(0L))
  }

  output_v <- tibble::tibble(property = property) %>%
    dplyr::filter(is.na(property) == FALSE) %>%
    dplyr::distinct(property) %>%
    dplyr::mutate(property = stringr::str_to_upper(.data$property)) %>%
    dplyr::filter(stringr::str_starts(
      string = .data$property,
      pattern = "P[[:digit:]]+$"
    )) %>%
    dplyr::pull("property")

  if (logical_vector | non_pid_as_NA) {
    output_l <- stringr::str_to_upper(property) %in% output_v
    if (non_pid_as_NA) {
      return(dplyr::if_else(
        condition = output_l,
        true = stringr::str_to_upper(property),
        false = NA_character_,
        missing = NA_character_
      ))
    } else {
      return(output_l)
    }
  } else {
    output_v
  }
}

#' Get all Wikidata Q identifiers of all Wikipedia pages (or files, or
#' subcategories) that are members of the given category,
#'
#' @param url Full URL to a Wikipedia category page. If given, title and
#'   language can be left empty.
#' @param category Title of a Wikipedia category page or final parts of its url.
#'   Must include "Category:", or equivalent in other languages. If given, url
#'   can be left empty, but language must be provided.
#' @param type Defaults to "page", defines which kind of members of a category
#'   are returned. Valid values include "page", "file", and "subcat" (for
#'   sub-category). Corresponds to `cmtype`. For details, see
#'   \href{https://www.mediawiki.org/wiki/API:Categorymembers}{the relevant page
#'   of the official documentation}.
#' @param attempts Defaults to 10. Number of times it re-attempts to reach the
#'   API before failing.
#' @inheritParams tw_get
#'
#' @return A data frame (a tibble) with eight columns: `source_title_url`,
#'   `source_wikipedia_title`, `source_qid`, `wikipedia_title`, `wikipedia_id`,
#'   `qid`, `description`, and `language`.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   sub_categories <- tw_get_wikipedia_category_members(
#'     category = "Category:American women anthropologists",
#'     type = "subcat"
#'   )
#'
#'   sub_categories
#'
#'   tw_get_wikipedia_category_members(
#'     category = sub_categories$wikipedia_title,
#'     type = "page"
#'   )
#' }
tw_get_wikipedia_category_members <- function(
  url = NULL,
  category = NULL,
  type = "page",
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
) {
  if (is.null(category) & !is.null(url)) {
    language <- stringr::str_extract(
      string = url,
      pattern = "(?<=https://)[[a-z]][[a-z]](?=.wikipedia.org/)"
    )
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
    title = category,
    language = language,
    url = url,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  source_df <- wikipedia_page_qid_df %>%
    dplyr::transmute(
      source_title_url = .data$title_url,
      source_wikipedia_category = .data$wikipedia_title,
      source_qid = .data$qid,
      language = .data$language
    ) %>%
    dplyr::distinct(.data$source_title_url, .keep_all = TRUE)

  wikipedia_category_members_df <- purrr::map_dfr(
    .x = seq_along(source_df$source_wikipedia_category),
    .f = function(i) {
      current_slice_df <- source_df %>%
        dplyr::slice(i)

      linked_df <- tw_get_wikipedia_category_members_single(
        category = current_slice_df$source_title_url,
        language = current_slice_df$language,
        url = NULL,
        type = type,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = db,
        disconnect_db = FALSE,
        wait = wait,
        attempts = attempts
      )

      linked_df
    }
  )

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  wikipedia_category_members_df
}


#' Get all Wikidata Q identifiers of all Wikipedia pages that appear in a given page
#'
#' @inheritParams tw_get_wikipedia_category_members
#'
#' @return A data frame (a tibble) with four columns: `wikipedia_title`, `wikipedia_id`, `wikidata_id`, `wikidata_description`.
#'
#' @examples
#' if (interactive()) {
#'   tidywikidatar:::tw_get_wikipedia_category_members_single(
#'     category = "Category:American women anthropologists",
#'     type = "subcat"
#'   )
#'
#'   tidywikidatar:::tw_get_wikipedia_category_members_single(
#'     category = "Category:Puerto Rican women anthropologists",
#'     type = "page"
#'   )
#' }
tw_get_wikipedia_category_members_single <- function(
  url = NULL,
  category = NULL,
  type = "page",
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (is.null(category)) {
    category <- stringr::str_extract(
      string = url,
      pattern = "wiki/.*$"
    ) %>%
      stringr::str_remove(pattern = stringr::fixed("wiki/"))
  }

  if (tw_check_cache(cache) & !overwrite_cache) {
    db_result <- tw_get_cached_wikipedia_category_members(
      category = category,
      type = type,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      category_df <- db_result %>%
        dplyr::collect() %>%
        dplyr::distinct()
    } else {
      category_df <- tidywikidatar::tw_empty_wikipedia_category_members
    }
  } else {
    category_df <- tidywikidatar::tw_empty_wikipedia_category_members
  }

  if (nrow(category_df) == 0) {
    json_url <- stringr::str_c(
      tw_get_wikipedia_base_api_url(
        url = url,
        title = category,
        language = language,
        type = "category"
      ),
      "&cmlimit=max&cmtype=",
      type
    )

    api_result <- FALSE

    attempt_n <- 1

    while (isFALSE(api_result) & attempt_n <= attempts) {
      attempt_n <- sum(attempt_n, 1)
      api_result <- tryCatch(
        jsonlite::read_json(path = json_url),
        error = function(e) {
          logical(1L)
        }
      )
      Sys.sleep(time = wait)
    }

    if (isFALSE(api_result)) {
      cli::cli_abort(c(
        "Could not reach the API with {attempts} attempts.",
        i = "Consider increasing the waiting time between calls with the {.arg wait} parameter or check your internet connection."
      ))
    } else if (length(api_result) == 1) {
      cli::cli_abort(
        "Page not found. Make sure that language parameter is consistent with the language of the input title or url."
      )
    } else {
      base_json <- api_result
    }

    continue_check <- base_json %>%
      purrr::pluck(
        "continue",
        "cmcontinue"
      )

    all_jsons <- list()

    page_number <- 1

    all_jsons[[page_number]] <- base_json

    while (is.null(continue_check) == FALSE & page_number < 1000) {
      page_number <- page_number + 1

      json_url <- stringr::str_c(
        json_url,
        "&cmcontinue=",
        continue_check
      )

      api_result <- FALSE

      attempt_n <- 1
      while (isFALSE(api_result) & attempt_n <= attempts) {
        attempt_n <- sum(attempt_n, 1)
        api_result <- tryCatch(
          jsonlite::read_json(path = json_url),
          error = function(e) {
            logical(1L)
          }
        )
        Sys.sleep(time = wait)
      }

      if (isFALSE(api_result)) {
        cli::cli_abort(c(
          "Could not reach the API with {attempts} attempts.",
          i = "Consider increasing the waiting time between calls with the {.arg wait} parameter or check your internet connection."
        ))
      } else {
        base_json <- api_result
      }

      all_jsons[[page_number]] <- base_json

      continue_check <- base_json %>%
        purrr::pluck(
          "continue",
          "cmcontinue"
        )
    }

    all_pages <- purrr::map(
      .x = all_jsons,
      .f = purrr::pluck,
      "query",
      "categorymembers"
    ) %>%
      purrr::flatten()

    page_title_df <- purrr::map_dfr(
      .x = all_pages,
      .f = function(current_page) {
        tibble::tibble(
          wikipedia_title = current_page %>%
            purrr::pluck(
              "title"
            ),
          wikipedia_id = current_page %>%
            purrr::pluck(
              "pageid"
            )
        )
      }
    )

    category_df <- page_title_df %>%
      dplyr::transmute(
        category = category,
        .data$wikipedia_title,
        wikipedia_id = as.numeric(.data$wikipedia_id)
      )

    if (tw_check_cache(cache)) {
      tw_write_wikipedia_category_members_to_cache(
        df = category_df,
        type = type,
        cache_connection = db,
        language = language,
        overwrite_cache = overwrite_cache,
        disconnect_db = disconnect_db
      )
    }
  }

  wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
    title = category_df$wikipedia_title,
    language = language,
    url = NULL,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = cache_connection,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  wikipedia_page_qid_df
}


#' Gets members of Wikipedia categories from local cache
#'
#' Mostly used internally.
#'
#' @inheritParams tw_get_wikipedia_category_members
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_set_cache_folder(path = tempdir())
#'   tw_enable_cache()
#'   tw_create_cache_folder(ask = FALSE)
#'
#'   df_from_api <- tw_get_wikipedia_page_qid(category = "Margaret Mead", language = "en")
#'
#'   df_from_cache <- tw_get_cached_wikipedia_category_members(
#'     category = "Margaret Mead",
#'     language = "en"
#'   )
#'
#'   df_from_cache
#' }
tw_get_cached_wikipedia_category_members <- function(
  category,
  type = "page",
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = stringr::str_c("wikipedia_category_members_", type),
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_category_members)
  }

  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(
        .data$category %in% !!stringr::str_c(category)
      ),
    error = function(e) {
      logical(1L)
    }
  )

  if (isFALSE(db_result)) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_category_members)
  }

  cached_df <- db_result %>%
    dplyr::collect()

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  cached_df
}


#' Write Wikipedia page links to cache
#'
#' Mostly used internally by `tidywikidatar`, use with caution to keep caching consistent.
#'
#' @param df A data frame typically generated with `tw_get_wikipedia_category_members()`.
#' @param language Defaults to language set with `tw_set_language()`; if not set, "en". Use "all_available" to keep all languages. For available language values, see https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all
#' @param type Defaults to "page", defines which kind of members of a category are returned. Valid values include "page", "file", and "subcat" (for sub-category). Corresponds to `cmtype`. For details, see \url{https://www.mediawiki.org/wiki/API:Categorymembers}
#' @param overwrite_cache Logical, defaults to FALSE. If TRUE, it overwrites the table in the local sqlite database. Useful if the original Wikidata object has been updated.
#' @param cache Defaults to NULL. If given, it should be given either TRUE or FALSE. Typically set with `tw_enable_cache()` or `tw_disable_cache()`.
#' @param cache_connection Defaults to NULL. If NULL, and caching is enabled, `tidywikidatar` will use a local sqlite database. A custom connection to other databases can be given (see vignette `caching` for details).
#' @param disconnect_db Defaults to TRUE. If FALSE, leaves the connection to cache open.
#'
#' @return Silently returns the same data frame provided as input. Mostly used internally for its side effects.
#'
#' @export
#'
#' @examples
#' if (interactive()) {
#'   df <- tw_get_wikipedia_category_members(
#'     category = "American women anthropologists",
#'     language = "en",
#'     cache = FALSE
#'   )
#'
#'   tw_write_wikipedia_category_members_to_cache(
#'     df = df,
#'     language = "en"
#'   )
#' }
tw_write_wikipedia_category_members_to_cache <- function(
  df,
  language = tidywikidatar::tw_get_language(),
  type = "page",
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = stringr::str_c("wikipedia_category_members_", type),
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE category = {category*}",
        source_title_url = unique(df$source_title_url),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(db, name = table_name, value = df, append = TRUE)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  invisible(df)
}

#' Reset Wikipedia category members cache
#'
#' Removes from cache the table where data typically gathered with
#' [tw_get_wikipedia_category_members()] are stored.
#'
#' @inheritParams tw_reset_wikipedia_page_links_cache
#' @inheritParams tw_get_wikipedia_category_members
#' @inheritParams tw_reset_item_cache
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_reset_wikipedia_category_members_cache()
#' }
tw_reset_wikipedia_category_members_cache <- function(
  language = tidywikidatar::tw_get_language(),
  type = "page",
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = stringr::str_c("wikipedia_category_members_", type),
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, nothing to delete
  } else if (isFALSE(ask)) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_inform(
      "Wikipedia category members cache reset for language {.val {language}} completed."
    )
  } else if (
    utils::menu(
      c("Yes", "No"),
      title = paste0(
        "Are you sure you want to remove from cache the Wikipedia category members cache for language: ",
        sQuote(language),
        "?"
      )
    ) ==
      1
  ) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_inform(
      "Wikipedia category members cache reset for language {.val {language}} completed."
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

#' Get image from Wikimedia Commons
#'
#' Please consult the
#' \href{https://commons.wikimedia.org/wiki/Commons:Reusing_content_outside_Wikimedia/technical}{relevant
#' documentation for reusing content outside Wikimedia}.
#'
#' @param id A character vector of length 1, must start with Q, e.g. "Q254" for
#'   Wolfgang Amadeus Mozart.
#' @param format A character vector, defaults to `filename`. If set to
#'   `commons`, outputs the link to the Wikimedia Commons page. If set to
#'   `embed`, outputs a link that can be used to embed.
#' @param width A numeric value, defaults to `NULL`, relevant only if format is
#'   set to 'embed'. If not given, defaults to full resolution image.
#' @inheritParams tw_get
#' @inheritParams tw_get_label
#'
#' @return A data frame of two columns, `id` and `image`, corresponding to
#'   reference to the image in the requested format.
#' @export
#'
#' @examples
#' \dontrun{
#' tw_get_image("Q180099",
#'   format = "filename"
#' )
#'
#'   tw_get_image("Q180099",
#'     format = "commons"
#'   )
#'
#'   tw_get_image("Q180099",
#'     format = "embed",
#'     width = 300
#'   )
#' }
tw_get_image <- function(
  id,
  format = "filename",
  width = NULL,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }
  filename_df <- tw_get_property(
    id = stringr::str_to_upper(id),
    p = "P18",
    language = language,
    id_df = id_df,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = cache_connection,
    disconnect_db = disconnect_db,
    wait = wait,
    retry = retry,
    user_agent = user_agent
  )

  image_df <- purrr::map2_dfr(
    .x = filename_df$value,
    .y = stringr::str_to_upper(filename_df$id),
    .f = function(current_filename, current_id) {
      if (is.na(current_filename)) {
        output_filename <- NA_character_
      } else if (format == "filename") {
        output_filename <- current_filename
      } else if (format == "commons") {
        output_filename <- stringr::str_c(
          "https://commons.wikimedia.org/wiki/File:",
          current_filename
        )
      } else if (format == "embed") {
        if (is.null(width)) {
          output_filename <- stringr::str_c(
            "https://commons.wikimedia.org/w/index.php?title=Special:Redirect/file/",
            current_filename
          )
        } else {
          output_filename <- stringr::str_c(
            "https://commons.wikimedia.org/w/index.php?title=Special:Redirect/file/",
            current_filename,
            "&width=",
            width
          )
        }
      } else {
        output_filename <- current_filename
      }
      tibble::tibble(
        id = current_id,
        image = output_filename
      )
    }
  )
  if (nrow(image_df) > 0) {
    image_df
  }
}

#' Get image from Wikimedia Commons
#'
#' Please consult the
#' \href{https://commons.wikimedia.org/wiki/Commons:Reusing_content_outside_Wikimedia/technical}{relevant
#' documentation for reusing content outside Wikimedia}.
#'
#' @param as_tibble Defaults to `FALSE`. If `TRUE`, returns a data frame instead
#'   of a character vector.
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @return A character vector, corresponding to reference to the image in the
#'   requested format.
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_get_image_same_length("Q180099",
#'     format = "filename"
#'   )
#'
#'   tw_get_image_same_length("Q180099",
#'     format = "commons"
#'   )
#'
#'   tw_get_image_same_length("Q180099",
#'     format = "embed",
#'     width = 300
#'   )
#' }
tw_get_image_same_length <- function(
  id,
  format = "filename",
  as_tibble = FALSE,
  only_first = TRUE,
  width = NULL,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  image_df <- tw_get_image(
    id = stringr::str_to_upper(id),
    format = format,
    width = width,
    language = language,
    id_df = id_df,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = cache_connection,
    disconnect_db = disconnect_db,
    wait = wait,
    retry = retry,
    user_agent = user_agent
  )

  if (is.null(image_df)) {
    return(rep(as.character(NA), length(id)))
  }
  if (as_tibble) {
    if (only_first) {
      dplyr::left_join(
        tibble::tibble(id = id),
        image_df %>%
          dplyr::group_by(.data$id) %>%
          dplyr::slice_head(n = 1) %>%
          dplyr::ungroup(),
        by = "id"
      )
    } else {
      dplyr::left_join(
        tibble::tibble(id = id),
        image_df %>%
          dplyr::group_by(.data$id) %>%
          dplyr::summarise(image = list(.data$image)),
        by = "id"
      )
    }
  } else {
    if (only_first) {
      dplyr::left_join(
        tibble::tibble(id = id),
        image_df %>%
          dplyr::group_by(.data$id) %>%
          dplyr::slice_head(n = 1) %>%
          dplyr::ungroup(),
        by = "id"
      ) %>%
        dplyr::pull("image")
    } else {
      dplyr::left_join(
        tibble::tibble(id = id),
        image_df %>%
          dplyr::group_by(.data$id) %>%
          dplyr::summarise(image = list(.data$image)),
        by = "id"
      ) %>%
        dplyr::pull("image")
    }
  }
}

#' Get metadata for images from Wikimedia Commons
#'
#' Please consult the
#' \href{https://commons.wikimedia.org/wiki/Commons:Reusing_content_outside_Wikimedia/technical}{relevant
#' documentation for reusing content outside Wikimedia}.
#'
#' @param id A character vector of length 1, must start with Q, e.g. "Q254" for
#'   Wolfgang Amadeus Mozart.
#' @param image_filename Defaults to `NULL`. If `NULL`, `image_filename` is
#'   obtained from the Wikidata id. If given, must be of the same length as id.
#' @param only_first Defaults to `TRUE`. If `TRUE`, returns only the first image
#'   associated with a given Wikidata id. If `FALSE`, returns all images
#'   available.
#' @param id_df Default to NULL. If given, it should be a dataframe typically
#'   generated with [tw_get()], and is used instead of calling Wikidata or
#'   using SQLite cache. Ignored when `id` is of length more than one.
#' @param attempts Defaults to 10. Number of times it re-attempts to reach the
#'   API before failing.
#' @inheritParams tw_get_image
#'
#' @return A character vector, corresponding to reference to the image in the
#'   requested format.
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_get_image_metadata("Q180099")
#' }
tw_get_image_metadata <- function(
  id,
  image_filename = NULL,
  only_first = TRUE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  attempts = 10,
  wait = 1,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (is.null(image_filename)) {
    image_filename <- tw_get_image_same_length(
      id = id,
      format = "filename",
      as_tibble = FALSE,
      only_first = only_first,
      language = language,
      id_df = id_df,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait,
      retry = retry,
      user_agent = user_agent
    )
  }

  input_df <- tibble::tibble(
    id = id,
    image_filename = image_filename
  ) %>%
    tidyr::unnest(image_filename)

  input_df_distinct <- dplyr::distinct(input_df)

  if (nrow(input_df_distinct) == 1) {
    return(
      dplyr::left_join(
        x = tibble::tibble(id = id),
        y = tw_get_image_metadata_single(
          id = input_df_distinct$id,
          image_filename = input_df_distinct$image_filename,
          only_first = only_first,
          language = language,
          id_df = id_df,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          wait = wait,
          retry = retry,
          attempts = attempts,
          user_agent = user_agent
        ),
        by = "id"
      )
    )
  } else if (nrow(input_df_distinct) > 1) {
    if (overwrite_cache | !tw_check_cache(cache)) {
      pb <- progress::progress_bar$new(total = nrow(input_df_distinct))

      image_metadata <- purrr::map2_dfr(
        .x = input_df_distinct$image_filename,
        .y = input_df_distinct$id,
        .f = function(current_image_filename, current_id) {
          pb$tick()
          tw_get_image_metadata_single(
            current_id,
            image_filename = current_image_filename,
            only_first = only_first,
            language = language,
            id_df = id_df,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            wait = wait,
            attempts = attempts
          )
        }
      )
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
      return(
        dplyr::left_join(
          x = tibble::tibble(id = id),
          y = image_metadata,
          by = "id"
        )
      )
    }

    if (!overwrite_cache & tw_check_cache(cache)) {
      table_name <- tw_get_cache_table_name(
        type = "image_metadata",
        language = language
      )

      if (pool::dbExistsTable(conn = db, name = table_name)) {
        db_result <- tryCatch(
          dplyr::tbl(src = db, table_name) %>%
            dplyr::filter(.data$id %in% !!stringr::str_to_upper(id)),
          error = function(e) {
            logical(1L)
          }
        )
        if (isFALSE(db_result)) {
          image_metadata_from_cache_df <- logical(1L)
        } else {
          image_metadata_from_cache_df <- db_result %>%
            dplyr::collect()
        }
      } else {
        image_metadata_from_cache_df <- logical(1L)
      }
    }

    if (isFALSE(image_metadata_from_cache_df)) {
      image_metadata_not_in_cache <- input_df_distinct
      image_metadata_from_cache_df <- input_df_distinct %>%
        dplyr::slice(0)
    } else {
      image_metadata_not_in_cache <- input_df_distinct %>%
        dplyr::anti_join(image_metadata_from_cache_df, by = "id")
    }

    if (nrow(image_metadata_not_in_cache) == 0) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
      return(
        dplyr::left_join(
          x = tibble::tibble(id = id),
          y = image_metadata_from_cache_df,
          by = "id"
        )
      )
    } else if (nrow(image_metadata_not_in_cache) > 0) {
      pb <- progress::progress_bar$new(
        total = nrow(image_metadata_not_in_cache)
      )

      image_metadata_not_in_cache_df <- purrr::map2_dfr(
        .x = image_metadata_not_in_cache$image_filename,
        .y = image_metadata_not_in_cache$id,
        .f = function(current_image_filename, current_id) {
          pb$tick()
          tw_get_image_metadata_single(
            id = current_id,
            image_filename = current_image_filename,
            only_first = only_first,
            language = language,
            id_df = id_df,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            wait = wait,
            attempts = attempts
          )
        }
      )

      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )

      dplyr::left_join(
        x = tibble::tibble(id = id),
        y = dplyr::bind_rows(
          image_metadata_from_cache_df,
          image_metadata_not_in_cache_df
        ),
        by = "id"
      )
    }
  }
}

#' Get metadata for images from Wikimedia Commons
#'
#' Please consult the
#' \href{https://commons.wikimedia.org/wiki/Commons:Reusing_content_outside_Wikimedia/technical}{relevant
#' documentation for reusing content outside Wikimedia}.
#'
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get
#' @inheritParams tw_get_single
#'
#' @return A character vector, corresponding to reference to the image in the
#'   requested format.
#'
#' @examples
#' if (interactive()) {
#'   tw_get_image_metadata_single("Q180099")
#' }
tw_get_image_metadata_single <- function(
  id,
  image_filename = NULL,
  only_first = TRUE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  read_cache = TRUE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (length(id) > 1) {
    cli::cli_abort(c(
      x = "id` must have length 1.",
      i = "Consider using {.fn tw_get_image_metadata}."
    ))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (is.null(image_filename)) {
    image_filename <- tw_get_image_same_length(
      id = stringr::str_to_upper(id),
      format = "filename",
      only_first = only_first,
      language = language,
      id_df = id_df,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait,
      retry = retry,
      user_agent = user_agent
    )
  }

  if (
    tw_check_cache(cache) &
      overwrite_cache == FALSE &
      read_cache
  ) {
    table_name <- tw_get_cache_table_name(
      type = "image_metadata",
      language = language
    )

    if (pool::dbExistsTable(conn = db, name = table_name)) {
      db_result <- tryCatch(
        dplyr::tbl(src = db, table_name) %>%
          dplyr::filter(.data$id %in% !!stringr::str_to_upper(id)),
        error = function(e) {
          logical(1L)
        }
      )
      if (isFALSE(db_result)) {
        image_metadata_from_cache_df <- logical(1L)
      } else {
        image_metadata_from_cache_df <- db_result %>%
          tibble::as_tibble() %>%
          dplyr::distinct()
        if (nrow(image_metadata_from_cache_df) > 0) {
          tw_disconnect_from_cache(
            cache = cache,
            cache_connection = db,
            disconnect_db = disconnect_db,
            language = language
          )
          return(image_metadata_from_cache_df)
        } else {
          image_metadata_from_cache_df <- logical(1L)
        }
      }
    } else {
      image_metadata_from_cache_df <- logical(1L)
    }
  }

  Sys.sleep(time = wait)

  empty_df <- data.frame(matrix(
    ncol = 19,
    nrow = 1,
    data = as.character(NA),
    dimnames = list(
      NULL,
      c(
        "id",
        "image_filename",
        "object_name",
        "image_description",
        "categories",
        "assessments",
        "credit",
        "artist",
        "permission",
        "license_short_name",
        "license_url",
        "license",
        "usage_terms",
        "attribution_required",
        "copyrighted",
        "restrictions",
        "date_time",
        "date_time_original",
        "commons_metadata_extension"
      )
    )
  )) %>%
    tibble::as_tibble()

  image_metadata <- purrr::map_dfr(
    .x = image_filename,
    .f = function(current_image_filename) {
      if (is.na(current_image_filename)) {
        empty_df$id[1] <- stringr::str_to_upper(id)
        empty_df$copyrighted <- as.logical(NA)
        empty_df$attribution_required <- as.logical(NA)
        return(empty_df)
      }

      api_link <- stringr::str_c(
        "https://commons.wikimedia.org/w/api.php?action=query&titles=File:",
        utils::URLencode(current_image_filename, reserved = TRUE),
        "&prop=imageinfo&iiprop=metadata%7Ccommonmetadata%7Cextmetadata",
        "&format=json"
      )

      api_result <- FALSE

      attempt_n <- 1

      while (isFALSE(api_result) & attempt_n <= attempts) {
        attempt_n <- sum(attempt_n, 1)
        api_result <- tryCatch(
          jsonlite::read_json(api_link),
          error = function(e) {
            logical(1L)
          }
        )
        Sys.sleep(time = wait)
      }

      if (isFALSE(api_result)) {
        cli::cli_abort(c(
          "Could not reach the API with {attempts} attempt{?s}.",
          i = "Consider increasing the waiting time between calls with the {.arg wait} parameter or check your internet connection."
        ))
      } else {
        json_as_list <- api_result
      }

      images <- json_as_list %>%
        purrr::pluck("query", "pages", 1, "imageinfo")

      extmetadata_list <- json_as_list %>%
        purrr::pluck("query", "pages", 1, "imageinfo", 1, "extmetadata")

      tibble::tibble(
        id = stringr::str_to_upper(id),
        image_filename = current_image_filename %>% as.character(),
        object_name = ifelse(
          test = is.null(
            extmetadata_list %>% purrr::pluck("ObjectName", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("ObjectName", "value")
        ) %>%
          as.character(),
        image_description = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("ImageDescription", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("ImageDescription", "value")
        ) %>%
          as.character(),
        categories = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Categories", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Categories", "value")
        ) %>%
          as.character(),
        assessments = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Assessments", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Assessments", "value")
        ) %>%
          as.character(),
        credit = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Credit", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Credit", "value")
        ),
        artist = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Artist", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Artist", "value")
        ) %>%
          as.character(),
        permission = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Permission", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Permission", "value")
        ) %>%
          as.character(),
        license_short_name = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("LicenseShortName", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("LicenseShortName", "value")
        ) %>%
          as.character(),
        license_url = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("LicenseUrl", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("LicenseUrl", "value")
        ) %>%
          as.character(),
        license = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("License", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>% purrr::pluck("License", "value")
        ) %>%
          as.character(),
        usage_terms = ifelse(
          test = is.null(
            extmetadata_list %>% purrr::pluck("UsageTerms", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck(
              "UsageTerms",
              "value"
            )
        ) %>%
          as.character(),
        attribution_required = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("AttributionRequired", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("AttributionRequired", "value")
        ) %>%
          stringr::str_to_upper() %>%
          as.logical(),
        copyrighted = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Copyrighted", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Copyrighted", "value")
        ) %>%
          stringr::str_to_upper() %>%
          as.logical(),
        restrictions = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("Restrictions", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("Restrictions", "value")
        ) %>%
          as.character(),
        date_time = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("DateTime", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("DateTime", "value")
        ) %>%
          as.character(),
        date_time_original = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("DateTimeOriginal", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("DateTimeOriginal", "value")
        ) %>%
          as.character(),
        commons_metadata_extension = ifelse(
          test = is.null(
            extmetadata_list %>%
              purrr::pluck("CommonsMetadataExtension", "value")
          ),
          yes = as.character(NA),
          no = extmetadata_list %>%
            purrr::pluck("CommonsMetadataExtension", "value")
        ) %>%
          as.character()
      )
    }
  )

  if (tw_check_cache(cache)) {
    table_name <- tw_get_cache_table_name(
      type = "image_metadata",
      language = language
    )

    if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
      # do nothing: if table does not exist, previous data cannot be there
    } else {
      if (overwrite_cache) {
        statement <- glue::glue_sql(
          "DELETE FROM {`table_name`} WHERE id = {id*}",
          id = unique(image_metadata$id),
          table_name = table_name,
          .con = db
        )
        result <- pool::dbExecute(
          conn = db,
          statement = statement
        )
      }
    }

    pool::dbWriteTable(
      db,
      name = table_name,
      value = image_metadata,
      append = TRUE
    )
  }
  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db
  )
  image_metadata
}

library("testthat")
library("tidywikidatar")

test_that("check if image returned when valid id given", {
  testthat::skip_if_offline()
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  tw_set_cache_folder(path = tempdir())

  test_id_df <- tw_get(
    id = "Q2",
    id_l = tw_test_items
  )

  expect_true(
    object = {
      tw_get_image(
        id = "Q2",
        cache = FALSE,
        id_df = test_id_df
      ) %>%
        tidyr::drop_na() %>%
        nrow() %>%
        as.logical()
    }
  )

  expect_true(
    object = {
      tw_get_image(
        id = "Q2",
        cache = TRUE,
        id_df = test_id_df
      ) %>%
        tidyr::drop_na() %>%
        nrow() %>%
        as.logical()
    }
  )
})

test_that("check if image returned when invalid id given", {
  testthat::skip_if_offline()
  testthat::skip_on_cran()
  testthat::skip_on_ci()

  test_id_df <- tw_get(
    id = "Q2",
    id_l = tw_test_items
  )

  expect_true(
    object = {
      tw_get_image(id = "non_qid_string") %>%
        is.null()
    }
  )

  expect_true(
    object = {
      tw_get_image_same_length(
        id = "non_qid_string",
        id_df = test_id_df
      ) %>%
        is.na()
    }
  )

  expect_equal(
    object = {
      tw_get_image_same_length(
        id = c("non_qid_string", NA),
        id_df = test_id_df
      ) %>%
        is.na() %>%
        sum()
    },
    expected = 2
  )

  expect_equal(
    object = {
      tw_get_image_same_length(
        id = c("non_qid_string", "Q2", "non_qid_string", NA),
        id_df = test_id_df
      ) %>%
        is.na() %>%
        which()
    },
    expected = c(1, 3, 4)
  )
})


test_that("check if image metadata returned correctly with or without cache", {
  testthat::skip_if_offline()
  testthat::skip_on_cran() # to prevent error due to calls to Wikimedia Commons from CRAN server
  testthat::skip_on_ci()

  test_id_df <- tw_get(
    id = "Q2",
    id_l = tw_test_items
  )

  expect_equal(
    object = {
      df <- tw_get_image_metadata_single(
        id = c("Q2"),
        only_first = TRUE,
        cache = FALSE
      )

      list(
        ncol = ncol(df),
        nrow = nrow(df),
        id = df %>% dplyr::pull(id)
      )
    },
    expected = list(
      ncol = 19,
      nrow = 1,
      id = "Q2"
    )
  )

  expect_equal(
    object = {
      tw_set_cache_folder(path = tempdir())

      df <- tw_get_image_metadata(
        id = c("Q2", NA, "not_an_id", "Q5"),
        only_first = TRUE,
        cache = TRUE,
        id_df = test_id_df,
        wait = 10
      )

      list(
        ncol = ncol(df),
        nrow = nrow(df),
        id = df %>% dplyr::pull(id),
        missing_image = which(is.na(df$image_filename))
      )
    },
    expected = list(
      ncol = 19,
      nrow = 4,
      id = c("Q2", NA, "not_an_id", "Q5"),
      missing_image = c(2, 3)
    )
  )

  expect_equal(
    object = {
      tw_set_cache_folder(path = tempdir())

      df <- tw_get_image_metadata(
        id = c("Q2", NA, "not_an_id", "Q5"),
        only_first = TRUE,
        cache = FALSE,
        id_df = test_id_df,
        wait = 10
      )

      list(
        ncol = ncol(df),
        nrow = nrow(df),
        id = df %>% dplyr::pull(id)
      )
    },
    expected = list(
      ncol = 19,
      nrow = 4,
      id = c("Q2", NA, "not_an_id", "Q5")
    )
  )
})

#' Get all items that have a given property (irrespective of the value)
#'
#' This function does not cache results.
#'
#' @param p A character vector, a property. Must always start with the capital
#'   letter "P", e.g. "P31" for "instance of".
#' @param wait Defaults to 0.1. Used only in method is set to "JSON".
#' @param method Defaults to "SPARQL". The only accepted alternative value is
#'   "JSON", to use instead json-based API.
#' @param limit Defaults to `Inf`. Set to smaller values for testing and cache
#'   locally when possible to reduce load on servers.
#' @inheritParams tw_query
#'
#' @return A data frame with three columns is method is set to "SPARQL", or as
#'   many columns as fields if more are given and `return_as_tw_search` is set
#'   to `FALSE`. A single column with Wikidata identifier if method is set to
#'   "JSON".
#' @export
#'
#' @examples
#' if (interactive()) {
#'   # get all Wikidata items with an ICAO airport code ("P239")
#'   tw_get_all_with_p(p = "P239", limit = 10)
#' }
tw_get_all_with_p <- function(
  p,
  fields = c("item", "itemLabel", "itemDescription"),
  language = tidywikidatar::tw_get_language(),
  method = "SPARQL",
  wait = 0.1,
  limit = Inf,
  return_as_tw_search = TRUE,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  p <- stringr::str_to_upper(string = p)

  if (stringr::str_starts(string = p, pattern = "P", negate = TRUE)) {
    cli::cli_warn("Invalid property. Property should always start with 'P'.")
    if (method == "SPARQL") {
      return(tidywikidatar::tw_empty_search)
    } else if (method == "API") {
      return(
        tibble::tibble(id = NA_character_) %>%
          dplyr::slice(0)
      )
    }
  }

  if (method == "SPARQL") {
    if (is.infinite(limit)) {
      sparql_t <- glue::glue(
        "SELECT
             {stringr::str_c(\"?\", stringr::str_c(fields, collapse = \" ?\"))}
             WHERE{{?item p:{p} ?statement0.
  ?statement0 (ps:{p}) _:anyValue{p}.
             SERVICE wikibase:label {{ bd:serviceParam wikibase:language '{stringr::str_c(language, collapse = ',')},[AUTO_LANGUAGE]' . }}
             }}"
      )
    } else {
      sparql_t <- glue::glue(
        "SELECT
             {stringr::str_c(\"?\", stringr::str_c(fields, collapse = \" ?\"))}
               WHERE{{?item p:{p} ?statement0.
               ?statement0 (ps:{p}) _:anyValue{p}.
               SERVICE wikibase:label {{ bd:serviceParam wikibase:language '{stringr::str_c(language, collapse = ',')},[AUTO_LANGUAGE]' . }}
             }}
      LIMIT {limit}"
      )
    }

    req <- httr2::request("https://query.wikidata.org/sparql") %>%
      httr2::req_headers(Accept = "text/csv") %>%
      httr2::req_user_agent(user_agent) %>%
      httr2::req_url_query(query = sparql_t) %>%
      httr2::req_retry(max_tries = 5, backoff = ~2)

    resp <- httr2::req_perform(req)

    response <- read.csv(
      text = httr2::resp_body_string(resp),
      stringsAsFactors = FALSE
    ) %>%
      tibble::as_tibble()

    if (length(fields) == 3 & return_as_tw_search) {
      all_items_df <- response %>%
        dplyr::transmute(
          id = stringr::str_extract(.data$item, pattern = "Q[[:digit:]]+$"),
          label = .data$itemLabel,
          description = .data$itemDescription
        )
    } else {
      all_items_df <- tibble::as_tibble(response)
    }
  } else if (method == "JSON") {
    api_url <- stringr::str_c(
      "https://www.wikidata.org/w/api.php?action=query&list=backlinks&blnamespace=0&bllimit=500&bltitle=Property:",
      p,
      "&format=json"
    )

    base_json <- jsonlite::read_json(api_url)

    continue_check <- base_json %>%
      purrr::pluck("continue", "blcontinue")

    all_jsons <- list()
    page_number <- 1

    all_jsons[[page_number]] <- base_json

    while (
      is.null(continue_check) == FALSE & page_number < max(2, (limit / 500))
    ) {
      Sys.sleep(wait)
      cli::cli_alert_info("Page {page_number} extracted")
      page_number <- page_number + 1

      base_json <- jsonlite::read_json(stringr::str_c(
        api_url,
        "&blcontinue=",
        continue_check
      ))

      all_jsons[[page_number]] <- base_json

      continue_check <- base_json %>%
        purrr::pluck("continue", "blcontinue")
    }

    all_pages <- purrr::map(
      .x = all_jsons,
      .f = purrr::pluck,
      "query",
      "backlinks"
    ) %>%
      purrr::flatten()

    all_items_df <- purrr::map_dfr(
      .x = all_pages,
      .f = function(current_page) {
        tibble::tibble(
          id = current_page %>%
            purrr::pluck(
              "title"
            )
        )
      }
    )
  }
  all_items_df
}

#' Retrieve item from the Wikidata API and returns it as a list
#'
#' @inheritParams tw_get
#' @inheritParams tw_set_user_agent
#'
#' @returns A list, with as many elements as the unique given id.
#' @export
#'
#' @examples
#' \dontrun{
#'   item_l <- tw_get_item(id = "Q180099")
#'
#'   tidywikidatar:::tw_extract_single(w = item_l)
#' }
tw_get_item <- function(
  id,
  user_agent = tidywikidatar::tw_get_user_agent(),
  retry = 10
) {
  items_l <- tw_check_qid(id = id) %>%
    purrr::set_names() %>%
    purrr::map(.f = function(current_id) {
      api_request <- httr2::request("https://www.wikidata.org/w/api.php") %>%
        httr2::req_url_query(
          action = "wbgetentities",
          ids = current_id,
          format = "json"
        ) %>%
        httr2::req_user_agent(string = "user_agent")

      if (retry) {
        api_request <- api_request %>%
          httr2::req_retry(max_tries = retry)
      }

      resp <- api_request %>%
        httr2::req_perform()

      item_l <- resp %>%
        httr2::resp_body_json() %>%
        purrr::pluck("entities")

      item_l
    })
  items_l
}

#' Facilitates the creation of MediaWiki API base URLs
#'
#' Mostly used internally
#'
#' @param action Defaults to "query". Usually either "query" or "parse". In
#'   principle, any valid action value, see:
#'   \url{https://www.mediawiki.org/w/api.php}
#' @param type Defaults to "page". Either "page" or "category".
#' @inheritParams tw_get_wikipedia_page_qid
#' @inheritParams tw_get
#'
#' @return A character vector of base urls to be used with the MediaWiki API.
#' @export
#'
#' @examples
#' tw_get_wikipedia_base_api_url(title = "Margaret Mead", language = "en")
#' tw_get_wikipedia_base_api_url(
#'   title = "Category:American women anthropologists",
#'   type = "category",
#'   language = "en"
#' )
tw_get_wikipedia_base_api_url <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language(),
  action = "query",
  type = "page"
) {
  if (is.null(url)) {
    if (is.null(title)) {
      cli::cli_abort("Either {.arg url} or {.arg title} must be provided.")
    }
    if (is.null(language)) {
      cli::cli_abort("Either {.arg language} or full url must be provided.")
    }
  } else {
    check_url_lv <- stringr::str_starts(string = url, pattern = "http")
    if (sum(is.na(check_url_lv)) > 0) {
      url <- url[is.na(check_url_lv) == FALSE]
      cli::cli_warn(c(
        "One or more of the given URLs is actually NA.",
        i = "Only valid URLs will be processed."
      ))
    }
    check_url_lv <- stringr::str_starts(string = url, pattern = "http")
    if (sum(check_url_lv) != length(check_url_lv)) {
      cli::cli_abort(c(
        "One or more of the Wikipedia URL provided does not start with `http` as expected for a URL.",
        "If you are actually providing Wikipedia page titles, leave the `url` parameter to NULL, and use the `title` parameter instead."
      ))
    }
    title <- stringr::str_extract(
      string = url,
      pattern = "(?<=https://[[a-z]][[a-z]].wikipedia.org/wiki/).*"
    )
  }

  if (is.null(language)) {
    language <- stringr::str_extract(
      string = url,
      pattern = "(?<=https://)[[a-z]][[a-z]](?=.wikipedia.org/)"
    )
  }

  if (action == "parse") {
    title_reference <- "&page="
  } else {
    title_reference <- "&titles="
  }

  if (type == "page") {
    api_url <- stringr::str_c(
      "https://",
      language,
      ".wikipedia.org/w/api.php?action=",
      action,
      "&redirects=true&format=json",
      title_reference,
      utils::URLencode(URL = title, reserved = TRUE)
    )
  } else if (type == "category") {
    title_reference <- "&cmtitle="

    api_url <- stringr::str_c(
      "https://",
      language,
      ".wikipedia.org/w/api.php?action=",
      action,
      "&redirects=true&format=json",
      title_reference,
      utils::URLencode(URL = title, reserved = TRUE),
      "&list=categorymembers"
    )
  }

  api_url
}


#' Gets the Wikidata Q identifier of one or more Wikipedia pages
#'
#' @param url A character vector with the full URL to one or more Wikipedia
#'   pages. If given, title and language can be left empty.
#' @param title Title of a Wikipedia page or final parts of its url. If given,
#'   url can be left empty, but language must be provided.
#' @param attempts Defaults to 10. Number of times it re-attempts to reach the
#'   API before failing.
#' @inheritParams tw_get
#'
#' @return A a data frame with six columns, including `qid` with Wikidata
#'   identifiers, and a logical `disambiguation` to flag when disambiguation
#'   pages are returned.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_qid(title = "Margaret Mead", language = "en")
#'
#'   # check when Wikipedia returns disambiguation page
#'   tw_get_wikipedia_page_qid(title = c("Rome", "London", "New York", "Vienna"))
#' }
tw_get_wikipedia_page_qid <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
) {
  if (is.null(url) == FALSE) {
    if (is.null(title)) {
      title <- dplyr::case_when(
        stringr::str_starts(string = url, pattern = "http") ~
          stringr::str_extract(
            string = url,
            pattern = "(?<=https://[[a-z]][[a-z]].wikipedia.org/wiki/).*"
          ),
        stringr::str_starts(string = url, pattern = stringr::fixed("/wiki/")) ~
          stringr::str_remove(
            string = url,
            pattern = stringr::fixed("/wiki/")
          ),
        .default = ""
      )
    } else {
      cli::cli_abort("Either url or title must be provided, not both.")
    }

    if (
      sum(stringr::str_starts(string = url, pattern = "http")) == length(url)
    ) {
      language <- stringr::str_extract(
        string = url,
        pattern = "(?<=https://)[[a-z]][[a-z]](?=.wikipedia.org/)"
      )
    }
  }

  unique_language <- unique(language)

  if (length(unique_language) == 0) {
    return(tidywikidatar::tw_empty_wikipedia_page)
  } else if (length(unique_language) > 1) {
    cli::cli_abort(
      "{.fn tw_get_wikipedia_page_qid} currently accepts only inputs with one language at a time."
    )
  }

  unique_title <- unique(title)

  if (length(unique_title) == 0) {
    return(tidywikidatar::tw_empty_wikipedia_page)
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = unique_language,
    cache = cache
  )

  if (length(unique_title) == 1) {
    return(
      dplyr::left_join(
        x = tibble::tibble(title_url = title),
        y = tw_get_wikipedia_page_qid_single(
          url = NULL,
          title = unique_title,
          language = unique_language,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          wait = wait,
          attempts = attempts
        ),
        by = "title_url"
      )
    )
  } else if (length(unique_title) > 1) {
    if (overwrite_cache | tw_check_cache(cache) == FALSE) {
      pb <- progress::progress_bar$new(total = length(unique_title))

      df <- purrr::map_dfr(
        .x = unique_title,
        .f = function(x) {
          pb$tick()
          tw_get_wikipedia_page_qid_single(
            url = NULL,
            title = x,
            language = unique_language,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            wait = wait,
            attempts = attempts
          )
        }
      )
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = unique_language
      )
      return(
        dplyr::left_join(
          x = tibble::tibble(title_url = title),
          y = df %>%
            dplyr::distinct(.data$title_url, .keep_all = TRUE),
          by = "title_url"
        )
      )
    }

    if (overwrite_cache == FALSE & tw_check_cache(cache)) {
      titles_in_cache_df <- tw_get_cached_wikipedia_page_qid(
        title = unique_title,
        language = unique_language,
        cache_connection = db,
        disconnect_db = FALSE
      )

      titles_not_in_cache <- unique_title[
        !is.element(unique_title, titles_in_cache_df$title_url)
      ]

      if (length(titles_not_in_cache) == 0) {
        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = unique_language
        )
        return(
          dplyr::left_join(
            x = tibble::tibble(title_url = title),
            y = titles_in_cache_df %>%
              dplyr::distinct(.data$title_url, .keep_all = TRUE),
            by = "title_url"
          )
        )
      } else if (length(titles_not_in_cache) > 0) {
        pb <- progress::progress_bar$new(total = length(titles_not_in_cache))

        titles_not_in_cache_df <- purrr::map_dfr(
          .x = titles_not_in_cache,
          .f = function(x) {
            pb$tick()
            tw_get_wikipedia_page_qid_single(
              url = NULL,
              title = x,
              language = unique_language,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE,
              wait = wait,
              attempts = attempts
            )
          }
        )

        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = unique_language
        )
        dplyr::left_join(
          x = tibble::tibble(title_url = title),
          y = dplyr::bind_rows(
            titles_in_cache_df,
            titles_not_in_cache_df
          ) %>%
            dplyr::distinct(.data$title_url, .keep_all = TRUE),
          by = "title_url"
        )
      }
    }
  }
}

#' Gets the Wikidata id of a Wikipedia page
#'
#' @inheritParams tw_get
#' @inheritParams tw_get_wikipedia_page_qid
#'
#' @return A data frame (a tibble) with eight columns: `title`,
#'   `wikipedia_title`, `wikipedia_id`, `qid`, `description`, `disambiguation`,
#'   and `language`.
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_qid_single(title = "Margaret Mead", language = "en")
#' }
tw_get_wikipedia_page_qid_single <- function(
  title = NULL,
  url = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
) {
  if (!is.null(url) && !is.function(url)) {
    if (is.null(title) & is.function(title) == FALSE) {
      if (
        stringr::str_starts(string = url, pattern = stringr::fixed("/wiki/"))
      ) {
        title <- stringr::str_remove(
          string = url,
          pattern = stringr::fixed("/wiki/")
        )
      } else {
        title <- stringr::str_extract(
          string = url,
          pattern = "(?<=https://[[a-z]][[a-z]].wikipedia.org/wiki/).*"
        )

        language <- stringr::str_extract(
          string = url,
          pattern = "(?<=https://)[[a-z]][[a-z]](?=.wikipedia.org/)"
        )
      }
      url <- NULL
    } else {
      cli::cli_abort("Either url or title must be provided, not both.")
    }
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (tw_check_cache(cache) & overwrite_cache == FALSE) {
    db_result <- tw_get_cached_wikipedia_page_qid(
      title = title,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      return(
        db_result %>%
          tibble::as_tibble()
      )
    }
  }

  json_url <- stringr::str_c(
    tw_get_wikipedia_base_api_url(
      url = url,
      title = title,
      language = language
    ),
    "&prop=pageprops"
  )

  api_result <- FALSE

  attempt_n <- 1
  while (isFALSE(api_result) & attempt_n <= attempts) {
    attempt_n <- sum(attempt_n, 1)
    api_result <- tryCatch(
      jsonlite::read_json(path = json_url),
      error = function(e) {
        logical(1L)
      }
    )
    Sys.sleep(time = wait)
  }

  if (isFALSE(api_result)) {
    cli::cli_abort(c(
      x = "Could not reach the API with {attempts} attempts.",
      i = "Consider increasing the waiting time between calls with the {.arg wait} parameter or check your internet connection."
    ))
  } else {
    wikidata_id_l <- api_result
  }

  wikipedia_id <- wikidata_id_l %>%
    purrr::pluck(
      "query",
      "pages",
      1,
      "pageid"
    )

  if (is.null(wikipedia_id)) {
    wikipedia_id <- NA_real_
  } else {
    wikipedia_id <- as.numeric(wikipedia_id)
  }

  wikidata_id <- wikidata_id_l %>%
    purrr::pluck(
      "query",
      "pages",
      1,
      "pageprops",
      "wikibase_item"
    )

  if (is.null(wikidata_id)) {
    wikidata_id <- NA_character_
  }

  description <- wikidata_id_l %>%
    purrr::pluck(
      "query",
      "pages",
      1,
      "pageprops",
      "wikibase-shortdesc"
    )

  if (is.null(description)) {
    description <- NA_character_
  }

  disambiguation <- is.element(
    el = "disambiguation",
    set = wikidata_id_l %>%
      purrr::pluck(
        "query",
        "pages",
        1,
        "pageprops"
      ) %>%
      names()
  )

  normalised <- wikidata_id_l %>%
    purrr::pluck(
      "query",
      "normalized",
      1,
      "to"
    )

  if (is.null(normalised)) {
    normalised <- NA_character_
  }

  redirected <- wikidata_id_l %>%
    purrr::pluck(
      "query",
      "redirects",
      1,
      "to"
    )

  if (is.null(redirected)) {
    redirected <- NA_character_
  }

  wikipedia_title <- dplyr::case_when(
    is.na(redirected) == FALSE ~ redirected,
    is.na(normalised) == FALSE ~ normalised,
    is.na(wikipedia_id) == TRUE ~ NA_character_,
    TRUE ~ title
  )

  df <- tibble::tibble(
    title_url = as.character(title),
    wikipedia_title = as.character(wikipedia_title),
    wikipedia_id = as.numeric(wikipedia_id),
    qid = as.character(wikidata_id),
    description = as.character(description),
    disambiguation = as.logical(disambiguation),
    language = as.character(language)
  )

  if (tw_check_cache(cache)) {
    tw_write_qid_of_wikipedia_page_to_cache(
      df = df,
      cache = cache,
      cache_connection = db,
      language = language,
      overwrite_cache = overwrite_cache,
      disconnect_db = disconnect_db
    )
  }

  df
}


#' Gets id of Wikipedia pages from local cache
#'
#' Mostly used internally.
#'
#' @param title Title of a Wikipedia page or final parts of its url. If given,
#'   url can be left empty, but language must be provided.
#' @inheritParams tw_get
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_set_cache_folder(path = tempdir())
#'   tw_enable_cache()
#'   tw_create_cache_folder(ask = FALSE)
#'
#'   df_from_api <- tw_get_wikipedia_page_qid(title = "Margaret Mead", language = "en")
#'
#'   df_from_cache <- tw_get_cached_wikipedia_page_qid(
#'     title = "Margaret Mead",
#'     language = "en"
#'   )
#'
#'   df_from_cache
#' }
tw_get_cached_wikipedia_page_qid <- function(
  title,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  title_url <- title

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_page)
  }

  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(
        .data$title_url %in% !!stringr::str_c(title_url)
      ),
    error = function(e) {
      logical(1L)
    }
  )
  if (isFALSE(db_result)) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_page)
  }

  cached_df <- db_result %>%
    dplyr::collect() %>%
    dplyr::mutate(disambiguation = as.logical(.data$disambiguation))

  if (disconnect_db) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
  }
  cached_df
}


#' Write Wikidata identifier (qid) of Wikipedia page to cache
#'
#' Mostly used internally by `tidywikidatar`, use with caution to keep caching
#' consistent.
#'
#' @param df A data frame typically generated with
#'   [tw_get_wikipedia_page_qid()].
#'
#' @inheritParams tw_get
#'
#' @return Silently returns the same data frame provided as input. Mostly used
#'   internally for its side effects.
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'   df <- tw_get_wikipedia_page_qid(
#'     title = "Margaret Mead",
#'     language = "en",
#'     cache = FALSE
#'   )
#'
#'   tw_write_qid_of_wikipedia_page_to_cache(
#'     df = df,
#'     language = "en"
#'   )
#' }
tw_write_qid_of_wikipedia_page_to_cache <- function(
  df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE title_url = {title_url*}",
        title_url = unique(df$title_url),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(db, name = table_name, value = df, append = TRUE)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  invisible(df)
}

#' Reset Wikipedia page cache
#'
#' Removes the table where data typically gathered with
#' [tw_get_wikipedia_page_qid()] from cache.
#'
#' @inheritParams tw_get
#' @inheritParams tw_reset_item_cache
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_reset_wikipedia_page_cache()
#' }
tw_reset_wikipedia_page_cache <- function(
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page",
    language = language
  )

  if (!pool::dbExistsTable(conn = db, name = table_name)) {
    # do nothing: if table does not exist, nothing to delete
  } else if (isFALSE(ask)) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Wikipedia page cache reset for language {.val {language}} completed."
    )
  } else if (
    utils::menu(
      c("Yes", "No"),
      title = paste0(
        "Are you sure you want to remove from cache the qualifiers table for language: ",
        sQuote(language),
        "?"
      )
    ) ==
      1
  ) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Wikipedia page cache reset for language {.val {language}} completed."
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

#' Gets labels for all columns with names such as "id" and "property".
#'
#' @param df A data frame, typically generated with other `tidywikidatar`
#'   functions such as [tw_get_property()].
#' @param value Logical, defaults to `TRUE`. If `TRUE`, it tries to get labels
#'   for all supposed id in the column called value. May break if the columns
#'   include some value which starts with Q and some digits, but is not a
#'   Wikidata id.
#' @inheritParams tw_get
#'
#' @return A data frame, with the same shape as the input data frame, but with
#'   labels instead of identifiers.
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_get_qualifiers(id = "Q180099", p = "P26", language = "en") %>%
#'     head(2) %>%
#'     tw_label()
#' }
#' }
tw_label <- function(
  df,
  value = TRUE,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (is.element("id", colnames(df))) {
    df[["id"]] <- tw_get_label(
      id = df[["id"]],
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait
    )
  }

  if (is.element("qualifier_id", colnames(df))) {
    df[["qualifier_id"]] <- tw_get_label(
      id = df[["qualifier_id"]],
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait
    )
  }

  if (is.element("property", colnames(df))) {
    df[["property"]] <- tw_get_property_label(
      property = df[["property"]],
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait
    )
  }

  if (is.element("qualifier_property", colnames(df))) {
    df[["qualifier_property"]] <- tw_get_property_label(
      property = df[["qualifier_property"]],
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait
    )
  }

  if (isTRUE(value)) {
    if (is.element("value", colnames(df))) {
      df[["value"]] <- purrr::map_chr(
        .x = df[["value"]],
        .f = function(x) {
          if (is.na(x)) {
            output <- x
          } else if (
            stringr::str_starts(
              string = x,
              pattern = "Q[[:digit:]]+"
            )
          ) {
            output <- tw_get_label(
              id = x,
              language = language,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE,
              wait = wait
            )
          } else {
            output <- x
          }
          output
        }
      )
    }

    if (is.element("qualifier_value", colnames(df))) {
      df[["qualifier_value"]] <- purrr::map_chr(
        .x = df[["qualifier_value"]],
        .f = function(x) {
          if (is.na(x)) {
            output <- x
          } else if (
            stringr::str_starts(
              string = x,
              pattern = "Q[[:digit:]]+"
            )
          ) {
            output <- tw_get_label(
              id = x,
              language = language,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE,
              wait = wait
            )
          } else {
            output <- x
          }
          output
        }
      )
    }
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  df
}

#' Writes search to cache
#'
#' Writes search to cache. Typically used internally, but exported to enable
#' custom caching solutions.
#'
#' @param search_df A data frame with four columns typically generated with
#'   `tw_search(include_search = TRUE)`.
#' @inheritParams tw_search
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' \dontrun{
#' tw_set_cache_folder(path = fs::path(tempdir(), paste(sample(letters, 24), collapse = "")))
#' tw_create_cache_folder(ask = FALSE)
#' tw_disable_cache()
#'
#' search_from_api <- tw_search(search = "Sylvia Pankhurst", include_search = TRUE)
#'
#' search_from_cache <- tw_get_cached_search("Sylvia Pankhurst")
#'
#' nrow(search_from_cache) == 0 # expect TRUE, as nothing has yet been stored in cache
#'
#' tw_write_search_to_cache(search_df = search_from_api)
#'
#' search_from_cache <- tw_get_cached_search("Sylvia Pankhurst")
#'
#' search_from_cache
#' }
tw_write_search_to_cache <- function(
  search_df,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  mandatory_cols <- c("search", "id", "label", "description")

  if (!identical(colnames(search_df), mandatory_cols)) {
    cli::cli_abort(c(
      "`search_df` must have exactly four columns. {.val {mandatory_cols}}."
    ))
  }

  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  language_combo <- stringr::str_c(language, "_", response_language)

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language_combo,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = stringr::str_c("search_", type),
    language = language,
    response_language = response_language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE search = {search*}",
        search = unique(search_df$search),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(db, name = table_name, value = search_df, append = TRUE)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language_combo
  )
}

#' Get Wikidata qualifiers for a given property of a given item
#'
#' N.B. In order to provide for consistently structured output, this function
#' outputs either id or value for each  qualifier. The user should keep in mind
#' that some of these come with additional detail (e.g. the unit, precision, or
#' reference calendar).
#'
#' @inheritParams tw_get
#' @inheritParams tw_get_property
#'
#' @return A data frame (a tibble) with eight columns: `id` for the input id,
#'   `property`,  `qualifier_id`, `qualifier_property`, `qualifier_value`,
#'   `rank`, `qualifier_value_type`, and `set` (to distinguish sets of data when
#'   a property is present more than once)
#'
#' @examples
#' if (interactive()) {
#'   tidywikidatar:::tw_get_qualifiers_single(id = "Q180099", p = "P26", language = "en")
#' }
#'
#' #' ## using `tw_test_items` in examples in order to show output without calling
#' ## on Wikidata servers
#'
#' tidywikidatar:::tw_get_qualifiers_single(
#'   id = "Q180099",
#'   p = "P26",
#'   language = "en",
#'   id_l = tw_test_items
#' )
tw_get_qualifiers_single <- function(
  id,
  p,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent(),
  id_l = NULL
) {
  if (tw_check_cache(cache) & !overwrite_cache) {
    db_result <- tw_get_cached_qualifiers(
      id = id,
      p = p,
      language = language,
      cache = cache,
      cache_connection = cache_connection,
      disconnect_db = FALSE
    )

    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = cache_connection,
        disconnect_db = disconnect_db,
        language = language
      )

      return(
        db_result %>%
          tibble::as_tibble()
      )
    }
  }

  Sys.sleep(time = wait)

  if (!is.null(id_l)) {
    w <- id_l %>%
      purrr::pluck(id) %>%
      list()

    if (length(w) == 0) {
      w <- tryCatch(
        tw_get_item(id = id, retry = retry, user_agent = user_agent),
        error = function(e) {
          as.character(e[[1]])
        }
      )
    } else if (length(w) > 1) {
      w <- w[1]
    }
  } else {
    w <- tryCatch(
      tw_get_item(id = id, retry = retry, user_agent = user_agent),
      error = function(e) {
        as.character(e[[1]])
      }
    )
  }

  if (is.character(w)) {
    cli::cli_alert_danger(w)
    return(tidywikidatar::tw_empty_qualifiers)
  } else {
    qualifiers_df <- tw_extract_qualifier(id = id, p = p, w = w)
  }

  if (nrow(qualifiers_df) == 0) {
    # keep one row, otherwise nothing remains in cache, and it will query
    # each time the script is re-run
    qualifiers_df <- tibble::tibble(
      id = as.character(id),
      property = as.character(p),
      qualifier_id = NA_character_,
      qualifier_property = NA_character_,
      qualifier_value = NA_character_,
      qualifier_value_type = NA_character_,
      rank = NA_character_,
      set = NA_real_
    )
  }

  if (tw_check_cache(cache)) {
    tw_write_qualifiers_to_cache(
      qualifiers_df = qualifiers_df,
      language = language,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      disconnect_db = disconnect_db
    )
  }

  qualifiers_df
}


#' Get Wikidata qualifiers for a given property of a given item
#'
#' N.B. In order to provide for consistently structured output, this function
#' outputs either id or value for each  qualifier. The user should keep in mind
#' that some of these come with additional detail (e.g. the unit, precision, or
#' reference calendar).
#'
#' @inheritParams tw_get
#' @inheritParams tw_get_property
#'
#' @return A data frame (a tibble) with eight columns: `id` for the input id,
#'   `property`,  `qualifier_id`, `qualifier_property`, `qualifier_value`,
#'   `rank`, `qualifier_value_type`, and `set` (to distinguish sets of data when
#'   a property is present more than once)
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tidywikidatar::tw_get_qualifiers(id = "Q180099", p = "P26", language = "en")
#' }
#'
#' #' ## using `tw_test_items` in examples in order to show output without calling
#' ## on Wikidata servers
#'
#' tidywikidatar::tw_get_qualifiers(
#'   id = "Q180099",
#'   p = "P26",
#'   language = "en",
#'   id_l = tw_test_items
#' )
tw_get_qualifiers <- function(
  id,
  p,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent(),
  id_l = NULL
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (length(id) == 0 || length(p) == 0) {
    cli::cli_abort("`id` and `p` must have length 1 or more.")
  } else if (length(id) == 1 & length(p) == 1) {
    return(tw_get_qualifiers_single(
      id = id,
      p = p,
      language = language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      wait = wait,
      id_l = id_l,
      retry = retry,
      user_agent = user_agent
    ))
  } else if (length(id) > 1 | length(p) > 1) {
    if (overwrite_cache | !tw_check_cache(cache)) {
      pb <- progress::progress_bar$new(total = length(id) * length(p))

      qualifiers_df <- purrr::map2_dfr(
        .x = id,
        .y = p,
        .f = function(x, y) {
          pb$tick()
          tw_get_qualifiers_single(
            id = x,
            p = y,
            language = language,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            wait = wait,
            id_l = id_l,
            retry = retry,
            user_agent = user_agent
          )
        }
      )
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
      return(qualifiers_df)
    }

    if (!overwrite_cache & tw_check_cache(cache)) {
      qualifiers_from_cache_df <- tw_get_cached_qualifiers(
        id = id,
        p = p,
        language = language,
        cache_connection = db,
        disconnect_db = FALSE
      )

      not_in_cache_df <- tibble::tibble(id = id, property = p) %>%
        dplyr::anti_join(qualifiers_from_cache_df, by = c("id", "property"))

      if (nrow(not_in_cache_df) == 0) {
        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language
        )
        return(
          dplyr::left_join(
            x = tibble::tibble(id = id),
            y = qualifiers_from_cache_df,
            by = "id"
          )
        )
      } else if (nrow(not_in_cache_df) > 0) {
        pb <- progress::progress_bar$new(total = nrow(not_in_cache_df))

        qualifiers_not_in_cache_df <- purrr::map2_dfr(
          .x = unique(not_in_cache_df$id),
          .y = unique(not_in_cache_df$property),
          .f = function(x, y) {
            pb$tick()
            tw_get_qualifiers_single(
              id = x,
              p = y,
              language = language,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE,
              wait = wait,
              id_l = id_l,
              retry = retry,
              user_agent = user_agent
            )
          }
        )

        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language
        )

        dplyr::left_join(
          x = tibble::tibble(id = id),
          y = dplyr::bind_rows(
            qualifiers_from_cache_df,
            qualifiers_not_in_cache_df
          ),
          by = "id"
        )
      }
    }
  }
}


#' Retrieve cached qualifier
#'
#' @inheritParams tw_get
#' @inheritParams tw_get_property
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#'
#' tw_set_cache_folder(path = tempdir())
#' tw_enable_cache()
#' tw_create_cache_folder(ask = FALSE)
#'
#' df_from_api <- tw_get_qualifiers(id = "Q180099", p = "P26", language = "en")
#'
#' df_from_cache <- tw_get_cached_qualifiers(
#'   id = "Q180099",
#'   p = "P26",
#'   language = "en"
#' )
#'
#' df_from_cache
tw_get_cached_qualifiers <- function(
  id,
  p,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(tidywikidatar::tw_empty_qualifiers))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "qualifiers",
    language = language
  )

  if (!pool::dbExistsTable(conn = db, name = table_name)) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
    return(tidywikidatar::tw_empty_qualifiers)
  }

  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(
        .data$id %in% !!stringr::str_to_upper(id),
        .data$property %in% !!stringr::str_to_upper(p)
      ),
    error = function(e) {
      logical(1L)
    }
  )
  if (isFALSE(db_result)) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
    return(tidywikidatar::tw_empty_qualifiers)
  } else if (
    isFALSE(identical(
      colnames(tidywikidatar::tw_empty_qualifiers),
      colnames(db_result)
    ))
  ) {
    cli::cli_abort(c(
      x = "The cache has been generated with a previous version of `tidywikidatar` that is not compatible with the current version.",
      i = "You may want to delete the old cache or reset just this table with {.fn tw_reset_qualifiers_cache}."
    ))
  }

  cached_qualifiers_df <- tibble::as_tibble(db_result)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  cached_qualifiers_df
}

#' Write qualifiers to cache
#'
#' Mostly to be used internally by `tidywikidatar`, use with caution to keep
#' caching consistent.
#'
#' @param qualifiers_df A data frame typically generated with
#'   [tw_get_qualifiers()].
#'
#' @inheritParams tw_get
#' @inheritParams tw_get_property
#'
#' @return Silently returns the same data frame provided as input. Mostly used
#'   internally for its side effects.
#'
#' @export
#'
#' @examples
#' if (interactive()) {
#' q_df <- tw_get_qualifiers(
#'   id = "Q180099",
#'   p = "P26",
#'   language = "en",
#'   cache = FALSE
#' )
#'
#' tw_write_qualifiers_to_cache(
#'   qualifiers_df = q_df,
#'   language = "en",
#'   cache = TRUE
#' )
#' }
tw_write_qualifiers_to_cache <- function(
  qualifiers_df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(qualifiers_df))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "qualifiers",
    language = language
  )

  if (!pool::dbExistsTable(conn = db, name = table_name)) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE id = {id*} AND property = {p*}",
        id = unique(qualifiers_df$id),
        p = unique(qualifiers_df$property),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(
    db,
    name = table_name,
    value = qualifiers_df,
    append = TRUE
  )

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  invisible(qualifiers_df)
}

#' Reset qualifiers cache
#'
#' Removes the table where qualifiers are cached
#'
#' @inheritParams tw_get
#' @inheritParams tw_reset_item_cache
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_reset_qualifiers_cache()
#' }
tw_reset_qualifiers_cache <- function(
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "qualifiers",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, nothing to delete
  } else if (isFALSE(ask)) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Qualifiers cache reset for language {.val {language}} completed."
    )
  } else if (
    utils::menu(
      c("Yes", "No"),
      title = paste0(
        "Are you sure you want to remove from cache the qualifiers table for language: ",
        sQuote(language),
        "?"
      )
    ) ==
      1
  ) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Qualifiers cache reset for language {.val {language}} completed."
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

#' Extract item data from a list object retrieved from the Wikidata API
#'
#' This function is mostly used internally and for testing.
#'
#' @param w A list, typically retrieved with [tw_get_item()]
#' @inheritParams tw_get
#'
#' @return A data frame (a tibble) with four columns, such as the one created by
#'   [tw_get()].
#'
#' @examples
#' #' Retrieving from tests, but normally:
#' # w <- tw_get_item(id = "Q180099")
#'
#' tidywikidatar:::tw_extract_single(w = list(tw_test_items[["Q180099"]]))
tw_extract_single <- function(w, language = tidywikidatar::tw_get_language()) {
  id <- w %>%
    purrr::pluck(1, 1, "id")

  labels <- w %>%
    purrr::pluck(1, 1, "labels")

  if (length(labels) == 0) {
    labels_df <- tibble::tibble(
      property = NA_character_,
      value = NA_character_,
      rank = NA_character_
    ) %>%
      dplyr::slice(0)
  } else {
    labels_df <- purrr::map_dfr(
      .x = labels,
      function(current_label_l) {
        tibble::tibble(
          property = paste0("label_", current_label_l$language),
          value = current_label_l$value,
          rank = as.character(NA)
        )
      }
    )
  }

  if (language == "all_available") {
    # do nothing
  } else {
    labels_df <- labels_df %>%
      dplyr::filter(.data$property == stringr::str_c("label_", language))
  }

  aliases <- w %>% purrr::pluck(1, 1, "aliases")

  if (length(aliases) == 0) {
    aliases_df <- tibble::tibble(
      property = NA_character_,
      values = NA_character_,
      rank = NA_character_
    ) %>%
      dplyr::slice(0)
  } else {
    aliases_df <- purrr::map_dfr(
      .x = aliases,
      function(current_alias_l) {
        tibble::tibble(
          property = paste0("alias_", current_alias_l$language),
          value = current_alias_l$value,
          rank = NA_character_
        )
      }
    )

    if (language == "all_available") {
      # do nothing
    } else {
      aliases_df <- aliases_df %>%
        dplyr::filter(.data$property == stringr::str_c("alias_", language))
    }
  }

  descriptions <- w %>%
    purrr::pluck(1, 1, "descriptions")

  if (length(descriptions) == 0) {
    descriptions_df <- tibble::tibble(
      property = NA_character_,
      values = NA_character_,
      rank = NA_character_
    ) %>%
      dplyr::slice(0)
  } else {
    descriptions_df <- purrr::map_dfr(
      .x = descriptions,
      function(current_description_l) {
        tibble::tibble(
          property = paste0("description_", current_description_l$language),
          value = current_description_l$value,
          rank = as.character(NA)
        )
      }
    )

    if (language == "all_available") {
      # do nothing
    } else {
      descriptions_df <- descriptions_df %>%
        dplyr::filter(
          .data$property == stringr::str_c("description_", language)
        )
    }
  }

  claims <- w %>%
    purrr::pluck(1, 1, "claims")

  claims_df <- purrr::map_dfr(
    .x = claims,
    .f = function(current_claim_l) {
      property <- current_claim_l %>%
        purrr::pluck(1, "mainsnak", "property")

      rank <- current_claim_l %>%
        purrr::pluck(1, "rank")

      value_pre <- current_claim_l %>%
        purrr::pluck(1, "mainsnak", "datavalue", "value")

      if (is.null(value_pre)) {
        value <- as.character("NA")
      } else if (is.list(value_pre)) {
        if (is.element("time", names(value_pre))) {
          value <- value_pre$time
        } else if (is.element("text", names(value_pre))) {
          value <- value_pre$text
        } else if (is.element("amount", names(value_pre))) {
          value <- value_pre$amount
        } else if (is.element("latitude", names(value_pre))) {
          value <- stringr::str_c(
            value_pre$latitude,
            value_pre$longitude,
            sep = ","
          )
        } else if (is.element("id", names(value_pre))) {
          value <- value_pre$id
        } else if (!is.na(value_pre[[1]])) {
          value <- value_pre[[1]]
        } else {
          value <- as.character("NA")
        }
      } else if (is.character(value_pre)) {
        value <- value_pre
      } else {
        value <- as.character("NA")
      }

      tibble::tibble(
        property = property,
        value = value,
        rank = rank
      )
    }
  )

  sitelinks <- w %>%
    purrr::pluck(1, 1, "sitelinks")

  sitelinks_df <- purrr::map_dfr(
    .x = sitelinks,
    function(current_sitelink_l) {
      tibble::tibble(
        property = paste0("sitelink_", current_sitelink_l$site),
        value = current_sitelink_l$title,
        rank = NA_character_
      )
    }
  )

  if (language == "all_available" | nrow(sitelinks_df) == 0) {
    # do nothing
  } else {
    sitelinks_df <- sitelinks_df %>%
      dplyr::filter(
        (.data$property ==
          stringr::str_c(
            "sitelink_",
            language,
            "wiki"
          )) |
          (.data$property ==
            stringr::str_c(
              "sitelink_",
              language,
              "wikiquote"
            )) |
          (.data$property ==
            stringr::str_c(
              "sitelink_",
              language,
              "wikisource"
            )) |
          (.data$property == "sitelink_commonswiki")
      )
  }

  everything_df <- dplyr::bind_rows(
    labels_df,
    aliases_df,
    claims_df,
    descriptions_df,
    sitelinks_df
  ) %>%
    tibble::as_tibble() %>%
    dplyr::transmute(
      id = id,
      .data$property,
      .data$value,
      .data$rank
    )
  everything_df
}


#' Extract qualifiers from a list created with [tw_get_item()]
#'
#' This function is mostly used internally and for testing.
#'
#' @param w A list, typically created with [tw_get_item()]
#' @inheritParams tw_get_property
#'
#' @return A data frame (a tibble) with eight columns: `id` for the input id,
#'   `property`,  `qualifier_id`, `qualifier_property`, `qualifier_value`,
#'   `rank`, `qualifier_value_type`, and `set` (to distinguish sets of data when
#'   a property is present more than once)
#' @export
#'
#' @examples
#' # w <- tw_get_item(id = "Q180099")
#'
#' tw_extract_qualifier(id = "Q180099", p = "P26", w = list(tw_test_items[["Q180099"]]))
tw_extract_qualifier <- function(
  id,
  p,
  w = NULL,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
) {
  if (is.null(w)) {
    w <- tryCatch(
      tw_get_item(id = id, retry = retry, user_agent = user_agent),
      error = function(e) {
        as.character(e[[1]])
      }
    )

    if (is.character(w)) {
      cli::cli_alert_danger(w)
      return(tidywikidatar::tw_empty_qualifiers)
    }
  }

  qualifiers <- w %>%
    purrr::pluck(
      1,
      1,
      "claims",
      p
    )

  qualifiers_df <- purrr::imap(
    .x = qualifiers,
    function(current_qualifier, i) {
      if (!is.element("qualifiers", names(current_qualifier))) {
        return(tidywikidatar::tw_empty_qualifiers)
      }

      current_rank <- current_qualifier %>%
        purrr::pluck("rank")

      qualifier_parent_pre <- current_qualifier %>%
        purrr::pluck("mainsnak", "datavalue", "value")

      if (is.character(qualifier_parent_pre)) {
        qualifier_parent <- qualifier_parent_pre[[1]]
      } else if (is.element("id", names(qualifier_parent_pre))) {
        qualifier_parent <- qualifier_parent_pre %>%
          purrr::pluck("id")
      } else {
        qualifier_parent <- qualifier_parent_pre %>%
          purrr::pluck(1)
      }

      qualifiers_set <- current_qualifier %>%
        purrr::pluck("qualifiers")

      purrr::map_dfr(
        .x = qualifiers_set,
        .f = function(x) {
          current_qualifier_set <- x %>%
            purrr::pluck(1) %>%
            tibble::as_tibble()

          value_df <- current_qualifier_set[["datavalue"]][["value"]] %>%
            tibble::as_tibble()
          if (is.element("id", names(value_df))) {
            value <- value_df %>%
              dplyr::pull("id")
          } else if (is.element("value", names(value_df))) {
            value <- value_df %>%
              dplyr::pull("value")
          } else if (is.element("amount", names(value_df))) {
            value <- value_df %>%
              dplyr::pull("amount")
          } else if (is.element("time", names(value_df))) {
            value <- value_df %>%
              dplyr::pull("time")
          } else {
            return(
              tidywikidatar::tw_empty_qualifiers %>%
                dplyr::select(
                  -"id",
                  -"property"
                )
            )
          }

          tibble::tibble(
            qualifier_id = qualifier_parent,
            qualifier_property = current_qualifier_set[["property"]][[1]],
            qualifier_value = value,
            qualifier_value_type = current_qualifier_set[["datavalue"]][[
              "type"
            ]],
            rank = current_rank,
            set = i
          )
        }
      )
    }
  ) %>%
    purrr::list_rbind() %>%
    dplyr::mutate(
      id = id,
      property = p,
      .before = 0
    )
  qualifiers_df
}

#' Efficiently get a wide table with various properties of a given set of
#' Wikidata identifiers
#'
#' @param label Logical, defaults to `FALSE`. If `TRUE` labels of Wikidata Q
#'   identifiers are reported instead of the identifiers themselves (or labels
#'   are presented along of them, if `both_id_and_label` is set to `TRUE`)
#' @param property_label_as_column_name Logical, defaults to `FALSE`. If `FALSE`,
#'   names of columns with properties are the "P" identifiers of the property.
#'   If `TRUE`, the label of the correspondent property is assigned as column
#'   name.
#' @param both_id_and_label Logical, defaults to `FALSE`. Relevant only if `label`
#'   is set to `TRUE`, otherwise ignored. If `TRUE`, the label is added as a
#'   separate column along the original one. Column name is the same as the
#'   property column, followed by "_label".
#' @param id_df_label Defaults to NULL. If given, it should be a dataframe
#'   typically generated with [tw_get()] with *all* items for which labels will
#'   be requested. It is used instead of calling Wikidata or relying on cache.
#' @param unlist Logical, defaults to `FALSE`. Typically used sharing or exporting
#'   data as csv files. Collapses all properties in a single string. The
#'   separator is defined by the `collapse` parameter. Relevant only when
#'   `only_first` is set to `FALSE`.
#' @param collapse Defaults to ";". Character used to separate results when
#'   `unlist` is set to `TRUE`.
#'
#' @inheritParams tw_get_property_same_length
#'
#' @return A data frame, with a column for each given property.
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'   tw_get_p_wide(
#'     id = c("Q180099", "Q228822", "Q191095"),
#'     p = c("P27", "P19", "P20"),
#'     label = TRUE,
#'     only_first = TRUE
#'   )
#' }
tw_get_p_wide <- function(
  id,
  p,
  label = FALSE,
  property_label_as_column_name = FALSE,
  both_id_and_label = FALSE,
  only_first = FALSE,
  preferred = FALSE,
  unlist = FALSE,
  collapse = ";",
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  id_df_label = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
) {
  if (is.data.frame(id)) {
    id <- id$id
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  unique_p <- unique(p)
  property <- NA_character_
  value <- NA_character_

  property_df <- tw_get_property(
    id = id,
    p = unique_p,
    language = language,
    id_df = id_df,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait
  )

  if (preferred) {
    preferred_df <- property_df %>%
      dplyr::mutate(
        rank = factor(
          .data$rank,
          levels = c(
            "preferred",
            "normal",
            "deprecated"
          )
        )
      ) %>%
      dplyr::group_by(id) %>%
      dplyr::arrange(
        .by_group = TRUE,
        .data$rank
      ) %>%
      dplyr::ungroup()

    if (nrow(preferred_df) > 0) {
      property_df <- preferred_df
    }
  }

  if (label) {
    property_df <- property_df %>%
      dplyr::mutate(
        label = dplyr::if_else(
          condition = tw_check_qid(
            id = .data$value,
            logical_vector = TRUE
          ),
          true = tw_get_label(
            id = .data$value,
            language = language,
            cache = cache,
            overwrite_cache = overwrite_cache,
            cache_connection = db,
            disconnect_db = FALSE,
            id_df = id_df_label,
            wait = wait
          ),
          false = .data$value,
          missing = NA_character_
        )
      )
  }

  if (only_first) {
    property_df <- property_df %>%
      dplyr::distinct(.data$id, .data$property, .keep_all = TRUE)
  } else if (!only_first) {
    property_df <- property_df %>%
      dplyr::distinct(.data$id, .data$property, .data$value, .keep_all = TRUE)

    if (label) {
      property_df <- property_df %>%
        dplyr::group_by(.data$id, .data$property) %>%
        dplyr::summarise(
          value = list(.data$value),
          label = list(.data$label),
          .groups = "drop"
        ) %>%
        dplyr::ungroup()
    } else {
      property_df <- property_df %>%
        dplyr::group_by(.data$id, .data$property) %>%
        dplyr::summarise(
          value = list(.data$value),
          .groups = "drop"
        ) %>%
        dplyr::ungroup()
    }
  }

  if (label) {
    if (both_id_and_label) {
      if (only_first) {
        property_df_wide <- property_df %>%
          tidyr::pivot_wider(
            id_cols = "id",
            names_from = "property",
            values_from = c("value", "label"),
            values_fill = NA_character_,
            names_glue = "{property}_{.value}"
          )
      } else if (!only_first) {
        property_df_wide <- property_df %>%
          tidyr::pivot_wider(
            id_cols = "id",
            names_from = "property",
            values_from = c("value", "label"),
            values_fill = list(NA_character_),
            names_glue = "{property}_{.value}"
          )
      }
    } else if (!both_id_and_label) {
      if (only_first) {
        property_df_wide <- property_df %>%
          tidyr::pivot_wider(
            id_cols = "id",
            names_from = "property",
            values_from = "label",
            values_fill = NA_character_
          )
      } else if (!only_first) {
        property_df_wide <- property_df %>%
          tidyr::pivot_wider(
            id_cols = "id",
            names_from = "property",
            values_from = "label",
            values_fill = list(NA_character_)
          )
      }
    } else {
      cli::cli_abort(
        message = c(
          x = "The parameter `both_id_and_label` must be either {.val {TRUE}} or {.val {FALSE}}."
        )
      )
    }

    if (both_id_and_label) {
      new_order <- c(
        "id",
        t(matrix(
          c(
            stringr::str_c(unique_p, "_value"),
            stringr::str_c(unique_p, "_label")
          ),
          ncol = 2
        )) %>%
          as.character()
      )

      property_df_wide_ordered <- property_df_wide[, new_order]

      names(property_df_wide_ordered) <- stringr::str_remove_all(
        string = new_order,
        pattern = "_value$"
      )
    } else {
      new_order <- c(
        "id",
        unique_p
      )

      property_df_wide_ordered <- property_df_wide[, new_order]
    }
  } else if (label == FALSE) {
    if (only_first) {
      property_df_wide <- property_df %>%
        tidyr::pivot_wider(
          id_cols = "id",
          names_from = "property",
          values_from = "value",
          values_fill = NA_character_
        )
    } else if (!only_first) {
      property_df_wide <- property_df %>%
        tidyr::pivot_wider(
          id_cols = "id",
          names_from = "property",
          values_from = "value",
          values_fill = list(NA_character_)
        )
    }

    new_order <- c(
      "id",
      unique_p
    )

    property_df_wide_ordered <- property_df_wide[, new_order]
  }

  if (property_label_as_column_name) {
    p_labels <- names(property_df_wide_ordered)[-1] %>%
      stringr::str_extract(pattern = "P[[:digit:]]+") %>%
      tw_get_property_label(
        language = language,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = db,
        disconnect_db = FALSE,
        wait = wait
      )

    if (both_id_and_label) {
      p_labels[!seq_along(p_labels) %% 2] <- stringr::str_c(
        p_labels[!seq_along(p_labels) %% 2],
        "_label"
      )
    }

    new_col_names <- p_labels %>%
      vctrs::vec_as_names(repair = "universal", quiet = TRUE) %>%
      stringr::str_replace_all(
        pattern = stringr::fixed("."),
        replacement = "_"
      )

    names(property_df_wide_ordered) <- c("id", new_col_names)
  }

  if (label) {
    output_df <- tibble::tibble(id = id) %>%
      dplyr::mutate(
        label = tw_get_label(
          id = .data$id,
          language = language,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = db,
          disconnect_db = FALSE,
          id_df = id_df_label,
          wait = wait
        )
      ) %>%
      dplyr::left_join(
        y = property_df_wide_ordered,
        by = "id"
      )
  } else {
    output_df <- tibble::tibble(id = id) %>%
      dplyr::left_join(
        y = property_df_wide_ordered,
        by = "id"
      )
  }
  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  if (unlist & !only_first) {
    output_df %>%
      dplyr::group_by(.data$id) %>%
      dplyr::mutate(
        dplyr::across(
          where(is.list),
          function(x) {
            stringr::str_c(unique(unlist(x)), collapse = collapse)
          }
        )
      ) %>%
      dplyr::ungroup()
  } else {
    output_df
  }
}

#' Get all Wikidata Q identifiers of all Wikipedia pages that appear in one or
#' more pages
#'
#' @param url Full url to a Wikipedia page. If given, title and language can be
#'   left empty.
#' @param title Title of a Wikipedia page or final parts of its url. If given,
#'   url can be left empty, but language must be provided.
#' @param language Two-letter language code used to define the Wikipedia version
#'   to use. Defaults to language set with [tw_set_language()]; if not set,
#'   "en". If url given, this can be left empty.
#'
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#'
#' @return A data frame (a tibble) with eight columns: `source_title_url`,
#'   `source_wikipedia_title`, `source_qid`, `wikipedia_title`, `wikipedia_id`,
#'   `qid`, `description`, and `language`.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_links(title = "Margaret Mead", language = "en")
#' }
tw_get_wikipedia_page_links <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
    title = title,
    language = language,
    url = url,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  source_df <- wikipedia_page_qid_df %>%
    dplyr::transmute(
      source_title_url = .data$title_url,
      source_wikipedia_title = .data$wikipedia_title,
      source_qid = .data$qid,
      language = .data$language
    ) %>%
    dplyr::distinct(.data$source_title_url, .keep_all = TRUE)

  if (tw_check_cache(cache) & !overwrite_cache) {
    db_result <- tw_get_cached_wikipedia_page_links(
      title = title,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      previously_cached_df <- db_result %>%
        dplyr::collect()
      unique_title <- unique(title)
      titles_not_in_cache <- unique_title[
        !is.element(unique_title, previously_cached_df$source_title_url)
      ]

      if (length(titles_not_in_cache) == 0) {
        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language
        )
        return(previously_cached_df)
      }
      source_df <- source_df %>%
        dplyr::filter(.data$source_title_url %in% titles_not_in_cache)
    } else {
      previously_cached_df <- tidywikidatar::tw_empty_wikipedia_page_links
    }
  } else {
    previously_cached_df <- tidywikidatar::tw_empty_wikipedia_page_links
  }

  pb <- progress::progress_bar$new(total = nrow(source_df))
  wikipedia_page_links_new_df <- purrr::map_dfr(
    .x = seq_along(source_df$source_title_url),
    .f = function(i) {
      pb$tick()

      current_slice_df <- source_df %>%
        dplyr::slice(i)

      linked_df <- tw_get_wikipedia_page_links_single(
        title = current_slice_df$source_title_url,
        language = current_slice_df$language,
        url = NULL,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = db,
        disconnect_db = FALSE,
        wait = wait,
        attempts = attempts,
        wikipedia_page_qid_df = current_slice_df %>%
          dplyr::transmute(title_url = .data$source_title_url) %>%
          dplyr::left_join(
            y = wikipedia_page_qid_df,
            by = "title_url"
          )
      )

      linked_df
    }
  )

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  wikipedia_page_qid_df %>%
    dplyr::distinct(.data$title_url) %>%
    dplyr::rename(source_title_url = .data$title_url) %>%
    dplyr::left_join(
      dplyr::bind_rows(
        previously_cached_df,
        wikipedia_page_links_new_df
      ),
      by = "source_title_url"
    )
}


#' Get all Wikidata Q identifiers of all Wikipedia pages that appear in a given
#' page
#'
#' @param wikipedia_page_qid_df Defaults to `NULL`. If given, used to reduce
#'   calls to cache. Must be a data frame.
#'
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#'
#' @return A data frame (a tibble) with four columns: `wikipedia_title`,
#'   `wikipedia_id`, `wikidata_id`, `wikidata_description`.
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_links_single(title = "Margaret Mead", language = "en")
#' }
tw_get_wikipedia_page_links_single <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10,
  wikipedia_page_qid_df = NULL
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (tw_check_cache(cache) & overwrite_cache == FALSE) {
    db_result <- tw_get_cached_wikipedia_page_links(
      title = title,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )

      return(
        db_result %>%
          dplyr::collect()
      )
    }
  }

  json_url <- stringr::str_c(
    tw_get_wikipedia_base_api_url(
      url = url,
      title = title,
      language = language
    ),
    "&prop=pageprops&generator=links&gpllimit=500"
  )

  api_result <- FALSE

  attempt_n <- 1

  while (isFALSE(api_result) & attempt_n <= attempts) {
    attempt_n <- sum(attempt_n, 1)
    api_result <- tryCatch(
      jsonlite::read_json(path = json_url),
      error = function(e) {
        logical(1L)
      }
    )
    Sys.sleep(time = wait)
  }

  if (isFALSE(api_result)) {
    cli::cli_abort(c(
      x = "Could not reach the API with {attempts} attempts.",
      i = "Consider increasing the waiting time between calls with the {.arg wait} argument or check your internet connection."
    ))
  } else if (length(api_result) == 1) {
    if (is.null(wikipedia_page_qid_df)) {
      wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
        title = title,
        language = language,
        url = url,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = cache_connection,
        disconnect_db = FALSE,
        wait = wait,
        attempts = attempts
      )
    }

    output_linked_df <- wikipedia_page_qid_df %>%
      dplyr::transmute(
        source_title_url = .data$title_url,
        source_wikipedia_title = .data$wikipedia_title,
        source_qid = .data$qid,
        wikipedia_title = NA_character_,
        wikipedia_id = NA_real_,
        qid = NA_character_,
        description = NA_character_,
        language = as.character(wikipedia_page_qid_df$language),
      )

    if (tw_check_cache(cache)) {
      tw_write_wikipedia_page_links_to_cache(
        df = output_linked_df,
        cache_connection = db,
        language = language,
        overwrite_cache = overwrite_cache,
        disconnect_db = disconnect_db
      )
    }

    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )

    return(output_linked_df)
  } else {
    base_json <- api_result
  }

  continue_check <- base_json %>%
    purrr::pluck(
      "continue",
      "gplcontinue"
    )

  all_jsons <- list()

  page_number <- 1

  all_jsons[[page_number]] <- base_json

  while (is.null(continue_check) == FALSE & page_number < 200) {
    page_number <- page_number + 1

    json_url <- stringr::str_c(
      json_url,
      "&gplcontinue=",
      utils::URLencode(URL = continue_check, reserved = TRUE)
    )

    api_result <- FALSE

    attempt_n <- 1
    while (isFALSE(api_result) & attempt_n <= attempts) {
      attempt_n <- sum(attempt_n, 1)
      api_result <- tryCatch(
        jsonlite::read_json(path = json_url),
        error = function(e) {
          logical(1L)
        }
      )
      Sys.sleep(time = wait)
    }

    if (isFALSE(api_result)) {
      cli::cli_abort(c(
        x = "Could not reach the API with {attempts} attempts.",
        i = "Consider increasing the waiting time between calls with the {.arg wait} argument or check your internet connection."
      ))
    } else {
      base_json <- api_result
    }

    all_jsons[[page_number]] <- base_json

    continue_check <- base_json %>%
      purrr::pluck(
        "continue",
        "gplcontinue"
      )
  }

  all_pages <- purrr::map(
    .x = all_jsons,
    .f = purrr::pluck,
    "query",
    "pages"
  ) %>%
    purrr::flatten()

  linked_df <- purrr::map_dfr(
    .x = all_pages,
    .f = function(current_page) {
      description <- current_page %>%
        purrr::pluck(
          "pageprops",
          "wikibase-shortdesc"
        )

      if (is.null(description)) {
        description <- NA_character_
      }

      tibble::tibble(
        qid = current_page %>%
          purrr::pluck(
            "pageprops",
            "wikibase_item"
          ),
        description = description,
        wikipedia_id = current_page %>%
          purrr::pluck(
            "pageid"
          ),
        wikipedia_title = current_page %>%
          purrr::pluck(
            "title"
          )
      )
    }
  ) %>%
    dplyr::select(
      "wikipedia_title",
      "wikipedia_id",
      "qid",
      "description"
    ) %>%
    dplyr::mutate(language = language)

  if (is.null(wikipedia_page_qid_df)) {
    wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
      title = title,
      language = language,
      url = url,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      disconnect_db = FALSE,
      wait = wait,
      attempts = attempts
    )
  }

  output_linked_df <- dplyr::bind_cols(
    wikipedia_page_qid_df %>%
      dplyr::transmute(
        source_title_url = .data$title_url,
        source_wikipedia_title = .data$wikipedia_title,
        source_qid = .data$qid
      ) %>%
      dplyr::distinct(.data$source_title_url, .keep_all = TRUE),
    linked_df
  )

  if (tw_check_cache(cache)) {
    tw_write_wikipedia_page_links_to_cache(
      df = output_linked_df,
      cache_connection = db,
      language = language,
      overwrite_cache = overwrite_cache,
      disconnect_db = disconnect_db
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  output_linked_df
}


#' Gets links of Wikipedia pages from local cache
#'
#' Mostly used internally.
#'
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_set_cache_folder(path = tempdir())
#'   tw_enable_cache()
#'   tw_create_cache_folder(ask = FALSE)
#'
#'   df_from_api <- tw_get_wikipedia_page_qid(title = "Margaret Mead", language = "en")
#'
#'   df_from_cache <- tw_get_cached_wikipedia_page_links(
#'     title = "Margaret Mead",
#'     language = "en"
#'   )
#'
#'   df_from_cache
#' }
tw_get_cached_wikipedia_page_links <- function(
  title,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page_links",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_page_links)
  }

  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(
        .data$source_title_url %in% !!stringr::str_c(title)
      ),
    error = function(e) {
      logical(1L)
    }
  )

  if (isFALSE(db_result)) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_page_links)
  }

  cached_df <- db_result %>%
    dplyr::collect()

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  cached_df
}


#' Write Wikipedia page links to cache
#'
#' Mostly used internally by `tidywikidatar`, use with caution to keep caching
#' consistent.
#'
#' @param df A data frame typically generated with
#'   [tw_get_wikipedia_page_links()].
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get
#'
#' @return Silently returns the same data frame provided as input. Mostly used
#'   internally for its side effects.
#'
#' @export
#'
#' @examples
#' if (interactive()) {
#'   df <- tw_get_wikipedia_page_links(
#'     title = "Margaret Mead",
#'     language = "en",
#'     cache = FALSE
#'   )
#'
#'   tw_write_wikipedia_page_links_to_cache(
#'     df = df,
#'     language = "en"
#'   )
#' }
tw_write_wikipedia_page_links_to_cache <- function(
  df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page_links",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE source_title_url = {source_title_url*}",
        source_title_url = unique(df$source_title_url),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(db, name = table_name, value = df, append = TRUE)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  invisible(df)
}

#' Reset Wikipedia page link cache
#'
#' Removes from cache the table where data typically gathered with
#' [tw_get_wikipedia_page_links()] are stored.
#'
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get
#' @inheritParams tw_reset_item_cache
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_reset_wikipedia_page_links_cache()
#' }
tw_reset_wikipedia_page_links_cache <- function(
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page_links",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, nothing to delete
  } else if (isFALSE(ask)) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Wikipedia page links cache reset for language {.val {language}} completed."
    )
  } else if (
    utils::menu(
      c("Yes", "No"),
      title = stringr::str_flatten(
        string = c(
          "Are you sure you want to remove from cache the Wikipedia page links cache for language: ",
          sQuote(language),
          "?"
        )
      )
    ) ==
      1
  ) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      c(
        i = "Wikipedia page links cache reset for language {.val {language}} completed."
      )
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

#' Gets a field such a label or description from a data frame typically generated
#' with [tw_get()]
#'
#' @param df A data frame typically generated with [tw_get()]. It should include
#'   data for the `id` included in the dedicated parameter.
#' @param field A character vector of length one. Typically, either "label" or
#'   "description".
#' @param id A character vector, typically of Wikidata identifiers. The output
#'   will be of the same length and in the same order as the identifiers
#'   provided with this parameter.
#' @inheritParams tw_get
#'
#' @return A character vector of the same length, and with data in the same
#'   order, as `id`.
#' @export
#'
#' @examples
#' \dontrun{
#' tw_get("Q180099") %>%
#'   tw_get_field(field = "label", id = "Q180099")
#' }
#'
#' # using test item for the sake of this example
#' tw_get("Q180099", id_l = tw_test_items) %>%
#'  tw_get_field(field = "label", id = "Q180099")
tw_get_field <- function(
  df,
  field,
  id,
  language = tidywikidatar::tw_get_language()
) {
  if (sum(is.na(id)) == length(id)) {
    return(rep(NA_character_, length(id)))
  }

  if (length(tw_check_qid(id = id)) == 0) {
    return(rep(NA_character_, length(id)))
  }

  field_df <- df %>%
    dplyr::filter(
      stringr::str_starts(
        string = .data$property,
        pattern = stringr::str_c(field, "_")
      ),
      stringr::str_ends(
        string = .data$property,
        pattern = stringr::str_c(language, collapse = "|")
      )
    ) %>%
    dplyr::distinct(.data$id, .keep_all = TRUE)

  if (nrow(field_df) == 0) {
    rep(NA_character_, length(id))
  } else if (nrow(field_df) < length(id)) {
    tibble::tibble(id = id) %>%
      dplyr::left_join(y = field_df, by = "id") %>%
      dplyr::pull("value")
  } else {
    field_df %>%
      dplyr::pull("value")
  }
}

#' Search for Wikidata items or properties and return Wikidata id, label, and
#' description.
#'
#' By defaults, this search returns items. Set `type` to property or use
#' [tw_search_property()] for properties.
#'
#' @param search A string to be searched in Wikidata
#' @param type Defaults to "item". Either "item" or "property".
#' @param language Language to be used for the search. Can be set once per
#'   session with [tw_set_language()]. If not set, defaults to "en". For a full
#'   list, see
#'   \href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
#'   dedicated Wikimedia page}.
#' @param response_language Language to be used for the returned labels and
#'   descriptions. Corresponds to the `uselang` parameter of the MediaWiki API,
#'   as described
#'   \href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
#'   the official documentation}. Can be set once per session with
#'   [tw_set_language()]. If not set, defaults to "en". For a full list, see
#'   \href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
#'   available language codes}.
#' @param limit Maximum numbers of responses to be given.
#' @param retry Defaults to 10. Maximum number of times to retry if the API
#'   throws an error, such as "too many requests". Each time, it will wait as
#'   much time as requested by the API. Notice that this can be a long time,
#'   e.g. 30 minutes. Set to `FALSE` if you prefer the API to throw an error
#'   immediately. Consider adjusting the `wait` parameter, or customising the
#'   `user_agent` if relevant.
#' @param include_search Logical, defaults to `FALSE`. If `TRUE`, the search is
#'   returned as an additional column.
#' @param wait In seconds, defaults to 0. Time to wait between queries to
#'   Wikidata. If data are cached locally, wait time is not applied. If you are
#'   running many queries systematically you may want to add some waiting time
#'   between queries.
#' @param cache Defaults to `NULL`. If given, it should be given either `TRUE`
#'   or `FALSE.` Typically set with [tw_enable_cache()] or [tw_disable_cache()].
#' @param overwrite_cache Defaults to `FALSE`. If `TRUE`, overwrites cache.
#' @param cache_connection Defaults to `NULL`. If `NULL`, and caching is
#'   enabled, `tidywikidatar` will use a local sqlite database. A custom
#'   connection to other databases can be given (see vignette `caching` for
#'   details).
#' @param disconnect_db Defaults to `TRUE`. If `FALSE`, leaves the connection to
#'   cache open.
#'
#' @inheritParams tw_set_user_agent
#'
#' @return A data frame (a tibble) with three columns (`id`, `label`, and
#'   `description`), and as many rows as there are results (by default, limited
#'   to 10). Four columns when `include_search` is set to `TRUE`.
#' @export
#'
#' @examples
#' tw_search(search = c("Margaret Mead", "Ruth Benedict"))
tw_search <- function(
  search,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  retry = 10,
  include_search = FALSE,
  wait = 0,
  user_agent = tidywikidatar::tw_get_user_agent(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  rlang::check_required(search)

  if (length(search) == 0) {
    cli::cli_abort("`search` must have length 1 or more.")
  }

  if (is.null(language)) {
    cli::cli_abort("A search language must be given.")
  }

  language_combo <- stringr::str_c(language, "_", response_language)

  unique_search <- unique(search)

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language_combo,
    cache = cache
  )

  if (length(unique_search) == 1) {
    search_df <-
      dplyr::left_join(
        x = tibble::tibble(search = search),
        y = tw_search_single(
          search = search,
          type = type,
          language = language,
          response_language = response_language,
          limit = limit,
          retry = retry,
          include_search = TRUE,
          wait = wait,
          user_agent = user_agent,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = db,
          disconnect_db = disconnect_db
        ),
        by = "search"
      )

    if (include_search) {
      return(search_df)
    } else {
      return(
        search_df %>%
          dplyr::select(-"search")
      )
    }
  } else if (length(unique_search) > 1) {
    if (overwrite_cache | !tw_check_cache(cache)) {
      pb <- progress::progress_bar$new(total = length(unique_search))
      search_df <- dplyr::left_join(
        x = tibble::tibble(search = search),
        y = purrr::map_dfr(
          .x = unique_search,
          .f = function(x) {
            pb$tick()
            tw_search_single(
              search = x,
              type = type,
              language = language,
              response_language = response_language,
              limit = limit,
              retry = retry,
              include_search = TRUE,
              wait = wait,
              user_agent = user_agent,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE
            )
          }
        ),
        by = "search"
      )

      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        language = language_combo,
        disconnect_db = disconnect_db
      )

      if (include_search) {
        return(search_df)
      } else {
        return(
          search_df %>%
            dplyr::select(-"search")
        )
      }
    }

    if (!overwrite_cache & tw_check_cache(cache)) {
      search_from_cache_df <- tw_get_cached_search(
        search = unique_search,
        type = type,
        language = language,
        response_language = response_language,
        cache = cache,
        include_search = TRUE,
        cache_connection = db,
        disconnect_db = FALSE
      )

      search_not_in_cache_v <- unique_search[
        !is.element(unique_search, search_from_cache_df$search)
      ]

      if (length(search_not_in_cache_v) == 0) {
        search_df <- dplyr::left_join(
          x = tibble::tibble(search = search),
          y = search_from_cache_df,
          by = "search"
        )
        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language_combo
        )
        if (include_search) {
          return(search_df)
        } else {
          return(
            search_df %>%
              dplyr::select(-"search")
          )
        }
      } else if (length(search_not_in_cache_v) > 0) {
        pb <- progress::progress_bar$new(total = length(search_not_in_cache_v))

        items_not_in_cache_df <- purrr::map_dfr(
          .x = search_not_in_cache_v,
          .f = function(x) {
            pb$tick()
            tw_search_single(
              search = x,
              type = type,
              language = language,
              response_language = response_language,
              limit = limit,
              retry = retry,
              include_search = TRUE,
              wait = wait,
              user_agent = user_agent,
              cache = cache,
              overwrite_cache = overwrite_cache,
              cache_connection = db,
              disconnect_db = FALSE
            )
          }
        )

        tw_disconnect_from_cache(
          cache = cache,
          cache_connection = db,
          disconnect_db = disconnect_db,
          language = language_combo
        )

        search_merged_df <- dplyr::left_join(
          x = tibble::tibble(search = search),
          y = dplyr::bind_rows(
            search_from_cache_df,
            items_not_in_cache_df
          ),
          by = "search"
        )

        if (include_search) {
          search_merged_df
        } else {
          search_merged_df %>%
            dplyr::select(-"search")
        }
      }
    }
  }
}


#' Search for Wikidata items or properties and return Wikidata id, label, and
#' description.
#'
#' This search returns only items, use [tw_search_property()] for properties.
#'
#' @inheritParams tw_search
#' @inheritParams tw_set_user_agent
#'
#' @return A data frame (a tibble) with three columns (`id`, `label`, and
#'   `description`), and as many rows as there are results (by default, limited
#'   to 10). Four columns when `include_search` is set to `TRUE`.
#'
#' @examples
#' \dontrun{
#'   tidywikidatar:::tw_search_single(search = "Sylvia Pankhurst")
#' }
tw_search_single <- function(
  search,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  retry = 10,
  include_search = FALSE,
  user_agent = tidywikidatar::tw_get_user_agent(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
) {
  rlang::check_required(search)
  if (is.null(search)) {
    cli::cli_abort("A search string must be given.")
  }

  if (is.null(language)) {
    cli::cli_abort("A search language must be given.")
  }

  if (length(search) > 1) {
    cli::cli_abort(c(
      x = "`search` must have length 1.",
      i = "Consider using {.fn tw_search}."
    ))
  }

  if (is.na(search)) {
    if (include_search) {
      return(tidywikidatar::tw_empty_search)
    } else {
      return(
        tidywikidatar::tw_empty_search %>%
          dplyr::select(-"search")
      )
    }
  }

  if (search == "") {
    if (include_search) {
      tidywikidatar::tw_empty_search %>%
        dplyr::add_row(search = "")
    } else {
      return(
        tidywikidatar::tw_empty_search %>%
          dplyr::select(-"search")
      )
    }
  }

  language_combo <- stringr::str_flatten(
    string = c(language, response_language),
    collapse = "_"
  )

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language_combo,
    cache = cache
  )

  if (tw_check_cache(cache) & !overwrite_cache) {
    db_result <- tw_get_cached_search(
      search = search,
      type = type,
      include_search = include_search,
      language = language,
      response_language = response_language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language_combo
      )

      return(
        db_result %>%
          tibble::as_tibble()
      )
    }
  }

  Sys.sleep(time = wait)

  base_url <- "https://www.wikidata.org/w/api.php"

  api_request <- httr2::request(base_url = base_url) %>%
    httr2::req_headers(`Accept-Encoding` = "gzip") %>%
    httr2::req_user_agent(user_agent) %>%
    httr2::req_url_query(
      action = "wbsearchentities",
      type = type,
      language = language,
      limit = limit,
      search = search,
      uselang = response_language,
      format = "json"
    ) %>%
    httr2::req_error(is_error = function(resp) FALSE)

  if (retry) {
    api_request <- api_request %>%
      httr2::req_retry(max_tries = retry)
  }

  response <- httr2::req_perform(api_request)

  if (httr2::resp_status(response) == 429) {
    wait_time <- httr2::resp_header(response, "Retry-After")

    error_message <- c(
      "x" = "API rate limit exceeded (HTTP 429), too many requests.",
      "i" = "Consider increasing your {.arg wait} time, or customising your {.arg user_agent} if relevant."
    )

    if (!is.null(wait_time)) {
      error_message <- c(
        error_message,
        "*" = "Please wait {wait_time} seconds before trying again."
      )
    }

    cli::cli_abort(error_message)
  } else if (httr2::resp_status(response)) {
    httr2::resp_check_status(response)
  }

  response_json <- response %>%
    httr2::resp_body_json()

  if (is.null(response_json[["error"]][["info"]]) == FALSE) {
    cli::cli_inform(c(
      `!` = "{.code {response_json[['error']][['code']]}} error when searching for: {.str {search}}",
      i = "{response_json[['error']][['info']]}"
    ))
    search_response <- tibble::tibble(
      id = NA_character_,
      label = NA_character_,
      description = NA_character_
    )
  } else {
    search_response <- response_json %>%
      purrr::pluck("search")
  }

  if (length(search_response) == 0) {
    search_response_df <- tibble::tibble(
      id = NA_character_,
      label = NA_character_,
      description = NA_character_
    )
  } else if (tibble::is_tibble(search_response)) {
    search_response_df <- search_response
  } else {
    search_response_df <- purrr::map_dfr(
      .x = search_response,
      .f = function(x) {
        extracted_label <- x %>% purrr::pluck("label", .default = NA_character_)
        extracted_description <- x %>%
          purrr::pluck("description", .default = NA_character_)

        tibble::tibble(
          id = x %>% purrr::pluck("id"),
          label = dplyr::if_else(
            condition = is.na(extracted_label),
            true = NA_character_,
            false = extracted_label
          ),
          description = dplyr::if_else(
            condition = is.na(extracted_description),
            true = NA_character_,
            false = extracted_description
          )
        )
      }
    )
  }
  search_response_df <- search_response_df %>%
    dplyr::mutate(search = search) %>%
    dplyr::select(
      "search",
      "id",
      "label",
      "description"
    )

  if (tw_check_cache(cache)) {
    tw_write_search_to_cache(
      search_df = search_response_df,
      type = type,
      language = language,
      response_language = response_language,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language_combo
  )

  if (include_search) {
    search_response_df %>%
      dplyr::filter(!is.na(.data$id)) %>%
      tibble::as_tibble()
  } else {
    search_response_df %>%
      dplyr::filter(!is.na(.data$id)) %>%
      dplyr::select(-"search") %>%
      tibble::as_tibble()
  }
}


#' Search for Wikidata properties in Wikidata and return Wikidata id, label, and
#' description.
#'
#' This search returns only items, use [tw_search_property()] for properties.
#'
#' @inheritParams tw_search
#'
#' @return A data frame (a tibble) with three columns (`id`, `label`, and
#'   `description`), and as many rows as there are results (by default, limited
#'   to 10).
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_search_item(search = "Sylvia Pankhurst")
#' }
tw_search_item <- function(
  search,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  wait = 0,
  user_agent = tidywikidatar::tw_get_user_agent(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  tw_search(
    search = search,
    type = "item",
    language = language,
    response_language = response_language,
    limit = limit,
    include_search = include_search,
    wait = wait,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = cache_connection,
    disconnect_db = disconnect_db
  )
}


#' Search for Wikidata properties in Wikidata and return Wikidata id, label, and
#' description.
#'
#' This search returns only properties, use [tw_search_item()] for properties.
#'
#' @inheritParams tw_search
#' @inheritParams tw_set_user_agent
#'
#' @return A data frame (a tibble) with three columns (`id`, `label`, and
#'   `description`), and as many rows as there are results (by default, limited
#'   to 10).
#' @export
#'
#' @examples
#' tw_search_property(search = "gender")
tw_search_property <- function(
  search,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  wait = 0,
  user_agent = tidywikidatar::tw_get_user_agent(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  tw_search(
    search = search,
    type = "property",
    language = language,
    response_language = response_language,
    limit = limit,
    include_search = include_search,
    wait = wait,
    cache = cache,
    cache_connection = cache_connection,
    overwrite_cache = overwrite_cache,
    disconnect_db = disconnect_db
  )
}

#' The Wikidata Q identifier of European airports found in Eurostat's `avia_par_` dataset
#'
#'
#' @format A data frame with 429 rows and 1 column:
#' \describe{
#'   \item{id}{Q identifiers}
#' }
#' @source \url{https://www.wikidata.org/wiki/Wikidata:Main_Page}
"tw_qid_airports"


#' The Wikidata Q identifier of all members of the European Parliament since its establishment
#'
#' A dataset with all the Wikidata items that have "Q27169" (member of the European Parliament) for the property "P39" (position held).
#'
#' @format A data frame with 4581 rows and 1 column:
#' \describe{
#'   \item{id}{Q identifiers}
#' }
#' @source \url{https://www.wikidata.org/wiki/Wikidata:Main_Page}
"tw_qid_meps"

#' A zero-rows tibble used internally when `tw_get()` would not return any value.
#'
#' @format A data frame with 0 rows and 3 columns
"tw_empty_item"

#' A zero-rows tibble used internally when `tw_get_qualifiers()` would not return any value.
#'
#' @format A data frame with 0 rows and 8 columns
"tw_empty_qualifiers"

#' A zero-rows tibble used internally when `tw_search()` would not return any value.
#'
#' @format A data frame with 0 rows and 4 columns
"tw_empty_search"

#' A zero-rows tibble used internally when `tw_get_wikipedia_page_qid()` would not return any value.
#'
#' @format A data frame with 0 rows and 6 columns
"tw_empty_wikipedia_page"


#' A zero-rows tibble used internally when `tw_get_wikipedia_page_links()` would not return any value.
#'
#' @format A data frame with 0 rows and 8 columns
"tw_empty_wikipedia_page_links"

#' A zero-rows tibble used internally when `tw_get_wikipedia_page_sections()` would not return any value.
#'
#' @format A data frame with 0 rows and 8 columns
"tw_empty_wikipedia_page_sections"


#' A zero-rows tibble used internally when `tw_get_image_metadata()` would not return any value.
#'
#' @format A data frame with 0 rows and 19 columns
"tw_empty_image_metadata"

#' A zero-rows tibble used internally when `tw_empty_wikipedia_category_members()` would not return any value.
#'
#' @format A data frame with 0 rows and 3 columns
"tw_empty_wikipedia_category_members"

#' A list mostly used for testing with some Wikidata items in the format resulting from `tw_get_item()`
#'
#' @format A list, an object such as the one resulting from `tw_get_item()`
"tw_test_items"

#' Filter search result and keep only items with matching property and Q
#' identifier
#'
#' @param search A data frame generated by `tw_search()`, or a search query. If
#'   a data frame is given, language and limits are ignore.
#' @param p A character vector of length 1, a property. Must always start with
#'   the capital letter "P", e.g. "P31" for "instance of".
#' @param q A character vector of length 1, a wikidata id. Must always start
#'   with the capital letter "Q", e.g. "Q5" for "human being".
#' @inheritParams tw_search
#'
#' @return A data frame with three columns, `id`, `label`, and `description`,
#'   filtered by the above criteria.
#' @export
#'
#' @examples
#' \dontrun{
#' tw_search(search = "Margaret Mead", limit = 3) %>%
#'   tw_filter(p = "P31", q = "Q5")
#' }
tw_filter <- function(
  search,
  p,
  q,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  wait = 0,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  search_result <- tw_check_search(
    search = search,
    language = language,
    response_language = response_language,
    limit = limit,
    include_search = include_search,
    wait = wait,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE
  )

  if (nrow(search_result) == 0) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )

    return(search_result)
  }

  p_df <- tw_get_property(
    id = search_result$id,
    p = p,
    language = language,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait
  ) %>%
    dplyr::filter(.data$value %in% q)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  search_result %>%
    dplyr::semi_join(
      y = p_df,
      by = "id"
    )
}

#' Filter search result and keep only and keep only the first match
#'
#' Same as [tw_filter()], but consistently returns data frames with a single
#' row.
#'
#' @inheritParams tw_filter
#'
#' @return A data frame with one row and three columns, `id`, `label`, and
#'   `description`, filtered by the above criteria.
#' @export
#'
#' @examples
#' \dontrun{
#' tw_search("Margaret Mead") %>%
#'   tw_filter_first(p = "P31", q = "Q5")
#' }
tw_filter_first <- function(
  search,
  p,
  q,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  wait = 0,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (tw_check_cache(cache)) {
    db <- tw_connect_to_cache(
      connection = cache_connection,
      language = language,
      cache = cache
    )
  }

  search_result <- tw_check_search(
    search = search,
    language = language,
    response_language = response_language,
    limit = limit,
    include_search = include_search,
    wait = wait,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE
  )

  if (nrow(search_result) == 0) {
    return(search_result)
  }

  first_match_id <- purrr::detect(
    .x = seq_along(search_result$id),
    .f = function(current_row_number) {
      check_v <- search_result %>%
        dplyr::slice(current_row_number) %>%
        tw_filter(
          p = p,
          q = q,
          language = language,
          response_language = response_language,
          limit = limit,
          include_search = include_search,
          wait = wait,
          cache = cache,
          overwrite_cache = overwrite_cache,
          cache_connection = cache_connection,
          disconnect_db = FALSE
        ) %>%
        nrow()

      check_v > 0
    }
  )

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  if (is.null(first_match_id)) {
    search_result %>%
      dplyr::slice(0)
  } else {
    search_result %>%
      dplyr::slice(first_match_id)
  }
}

#' Filter search result and keep only people
#'
#' A wrapper of [tw_filter()] that defaults to keep only "instance of" (P31)
#' "human being" (Q5).
#'
#' @param stop_at_first Logical, defaults to TRUE. If TRUE, returns only the
#'   first match from the search that satisfies the criteria.
#' @inheritParams tw_filter
#' @return A data frame with three columns, `id`, `label`, and `description`;
#'   all rows refer to a human being.
#' @export
#'
#' @examples
#' \dontrun{
#'   tw_search("Ruth Benedict")
#'
#'   tw_search("Ruth Benedict") %>%
#'     tw_filter_people()
#' }
tw_filter_people <- function(
  search,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  stop_at_first = TRUE,
  wait = 0,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (stop_at_first) {
    tw_filter_first(
      search = search,
      p = "P31",
      q = "Q5",
      language = language,
      response_language = response_language,
      limit = limit,
      include_search = include_search,
      wait = wait,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      disconnect_db = disconnect_db
    )
  } else {
    tw_filter(
      search = search,
      p = "P31",
      q = "Q5",
      language = language,
      response_language = response_language,
      limit = limit,
      include_search = include_search,
      wait = wait,
      overwrite_cache = overwrite_cache,
      cache_connection = cache_connection,
      disconnect_db = disconnect_db
    )
  }
}

#' Tidy eval helpers
#'
#' @description
#'
#' * \code{\link[rlang]{sym}()} creates a symbol from a string and
#'   \code{\link[rlang:sym]{syms}()} creates a list of symbols from a
#'   character vector.
#'
#' * \code{\link[rlang:nse-defuse]{enquo}()} and
#'   \code{\link[rlang:nse-defuse]{enquos}()} delay the execution of one or
#'   several function arguments. \code{enquo()} returns a single quoted
#'   expression, which is like a blueprint for the delayed computation.
#'   \code{enquos()} returns a list of such quoted expressions.
#'
#' * \code{\link[rlang:nse-defuse]{expr}()} quotes a new expression _locally_. It
#'   is mostly useful to build new expressions around arguments
#'   captured with [enquo()] or [enquos()]:
#'   \code{expr(mean(!!enquo(arg), na.rm = TRUE))}.
#'
#' * \code{\link[rlang]{as_name}()} transforms a quoted variable name
#'   into a string. Supplying something else than a quoted variable
#'   name is an error.
#'
#'   That's unlike \code{\link[rlang]{as_label}()} which also returns
#'   a single string but supports any kind of R object as input,
#'   including quoted function calls and vectors. Its purpose is to
#'   summarise that object into a single label. That label is often
#'   suitable as a default name.
#'
#'   If you don't know what a quoted expression contains (for instance
#'   expressions captured with \code{enquo()} could be a variable
#'   name, a call to a function, or an unquoted constant), then use
#'   \code{as_label()}. If you know you have quoted a simple variable
#'   name, or would like to enforce this, use \code{as_name()}.
#'
#'
#' @md
#' @name tidyeval
#' @keywords internal
#' @importFrom rlang expr enquo enquos sym syms .data := as_name as_label
#' @aliases expr enquo enquos sym syms .data := as_name as_label
#' @export expr enquo enquos sym syms .data := as_name as_label
NULL


#' @import utils
utils::globalVariables("where")

#' Ensures that input appears to be a valid Wikidata id
#'
#' Mostly used internally by other functions.
#'
#' @param id A character vector of one or more Wikidata id.
#' @param logical_vector Logical, defaults to `FALSE`. If `TRUE`, returns a
#'   logical vector of the same length as input, where `TRUE` corresponds to
#'   seemingly meaningful Q identifiers.
#' @param non_id_as_NA Logical, defaults to `FALSE`. If `TRUE` (and if
#'   `logical_vector` is set to `FALSE`), a vector of the same length is
#'   returned, with `NA` replacing items that are seemingly not meaningful Q
#'   identifiers.
#'
#' @return A character vector with only strings appearing to be Wikidata
#'   identifiers; possibly shorter than input.
#' @export
#'
#' @examples
#'
#' tw_check_qid(id = c("Q180099", "q228822", "Not an id", "00180099", NA, "Q5"))
#'
#' tw_check_qid(
#'   id = c("Q180099", "q228822", "Not an id", "00180099", NA, "Q5"),
#'   logical_vector = TRUE
#' )
#'
#' tw_check_qid(
#'   id = c("Q180099", "q228822", "Not an id", "00180099", NA, "Q5"),
#'   non_id_as_NA = TRUE
#' )
tw_check_qid <- function(id, logical_vector = FALSE, non_id_as_NA = FALSE) {
  if (is.null(id)) {
    return(character(0L))
  }

  output_v <- tibble::tibble(id = id) %>%
    dplyr::filter(!is.na(id)) %>%
    dplyr::distinct(id) %>%
    dplyr::mutate(id = stringr::str_to_upper(.data$id)) %>%
    dplyr::filter(stringr::str_starts(
      string = .data$id,
      pattern = "Q[[:digit:]]+$"
    )) %>%
    dplyr::pull("id")

  if (logical_vector | non_id_as_NA) {
    output_l <- stringr::str_to_upper(id) %in% output_v
    if (non_id_as_NA) {
      return(dplyr::if_else(
        condition = output_l,
        true = stringr::str_to_upper(id),
        false = NA_character_,
        missing = NA_character_
      ))
    } else {
      return(output_l)
    }
  } else {
    output_v
  }
}

#' Check if cache table is indexed
#'
#' Tested only with SQLite and MySql. May work with other drivers. Used to check
#' if given cache table is indexed (if created with any version of
#' `tidywikidatar` before 0.6, they are probably not indexed and less
#' efficient).
#'
#' @param table_name Name of the table in the database. If given, it takes
#'   precedence over other parameters.
#' @param show_details Logical, defaults to `FALSE`. If `FALSE`, return a
#'   logical vector of length one (`TRUE` if the table was indexed, `FALSE` if
#'   it was not). If `TRUE`, returns a data frame with more details about the
#'   index.
#'
#' @inheritParams tw_get_cache_table_name
#' @inheritParams tw_get
#'
#' @return If `show_details` is set to `FALSE`, return a logical vector of
#'   length one (`TRUE` if the table was indexed, `FALSE` if it was not). If
#'   `show_details` is set to `TRUE`, returns a data frame with more details
#'   about the index.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_enable_cache()
#'   tw_set_cache_folder(path = fs::path(
#'     fs::path_home_r(),
#'     "R",
#'     "tw_data"
#'   ))
#'
#'   tw_set_language(language = "en")
#'
#'   tw_check_cache_index()
#' }
tw_check_cache_index <- function(
  table_name = NULL,
  type = "item",
  show_details = FALSE,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (is.null(table_name)) {
    table_name <- tw_get_cache_table_name(
      type = type,
      language = language,
      response_language = response_language
    )
  }

  if (
    stringr::str_starts(string = type, pattern = "search") &
      stringr::str_detect(string = language, pattern = "_", negate = TRUE)
  ) {
    language <- stringr::str_c(language, "_", response_language)
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (is.null(db)) {
    cli::cli_abort(c(
      "No valid connection found.",
      i = "Enable caching with `tw_enable_cache()` or through the relevant parameters."
    ))
  }

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    cli::cli_warn("Table {.code {table_name}} does not exist.")
    return(invisible())
  }

  driver <- db$fetch()

  if (inherits(x = driver, "SQLiteConnection")) {
    db_no_pool <- pool::poolCheckout(db)

    statement <- glue::glue_sql(
      "PRAGMA INDEX_LIST({`table_name`});",
      table_name = table_name,
      .con = db_no_pool
    )

    index_odbc_result <- DBI::dbSendQuery(
      conn = db_no_pool,
      statement = statement
    )

    index_odbc_result_fetched <- DBI::dbFetch(index_odbc_result)

    pool::poolReturn(db_no_pool)

    if (show_details) {
      output <- tibble::as_tibble(index_odbc_result_fetched)
    } else {
      output <- nrow(tibble::as_tibble(index_odbc_result_fetched)) > 0
    }
  } else {
    if (show_details) {
      db_no_pool <- pool::poolCheckout(db)

      statement <- glue::glue_sql(
        "SHOW INDEX FROM {`table_name`}",
        table_name = table_name,
        .con = db_no_pool
      )

      index_odbc_result <- DBI::dbSendQuery(
        conn = db_no_pool,
        statement = statement
      )

      index_odbc_result_fetched <- DBI::dbFetch(index_odbc_result)
      pool::poolReturn(db_no_pool)

      output <- tibble::as_tibble(index_odbc_result_fetched)
    } else {
      statement <- glue::glue_sql(
        "SHOW INDEX FROM {`table_name`}",
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )

      output <- result > 0
    }
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db
  )
  output
}


#' Add index to caching table for search queries for increased speed
#'
#' Tested only with SQLite and MySql. May work with other drivers.
#'
#' To ensure smooth functioning, the search column in the cache table is
#' transformed into a column of type `varchar` and length 255.
#'
#' @param table_name Name of the table in the database. If given, it takes
#'   precedence over other parameters.
#' @param check_first Logical, defaults to `TRUE`. If `TRUE`, then before
#'   executing anything on the database it checks if the given table has already
#'   been indexed. If it has, it does nothing and returns only an informative
#'   message.
#' @param show_details Logical, defaults to `FALSE`. If `FALSE`, return the
#'   function adds the index to the database, but does not return anything. If
#'   `TRUE`, returns a data frame with more details about the index.
#'
#' @inheritParams tw_get_cache_table_name
#' @inheritParams tw_get
#'
#' @return If `show_details` is set to `FALSE`, nothing, used only for its side
#'   effects (add index to caching table). If `TRUE`, a data frame, same as the
#'   output of `tw_check_cache_index(show_details = TRUE)`.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_enable_cache()
#'   tw_set_cache_folder(path = fs::path(
#'     fs::path_home_r(),
#'     "R",
#'     "tw_data"
#'   ))
#'
#'   tw_index_cache_search()
#' }
tw_index_cache_search <- function(
  table_name = NULL,
  check_first = TRUE,
  type = "item",
  show_details = FALSE,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (is.null(table_name)) {
    table_name <- tw_get_cache_table_name(
      type = stringr::str_c("search_", type),
      language = language,
      response_language = response_language
    )
    language <- stringr::str_c(language, "_", response_language)
  } else {
    language <- stringr::str_extract(
      string = table_name,
      "[a-z][a-z]_[a-z][a-z]$"
    )
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (!pool::dbExistsTable(conn = db, name = table_name)) {
    cli::cli_warn("Table `{table_name}` does not exist.")
    return(invisible())
  }

  if (check_first) {
    check <- tw_check_cache_index(
      table_name = table_name,
      type = type,
      show_details = FALSE,
      language = language,
      response_language = response_language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )

    if (check) {
      cli::cli_inform(c(
        "The table {.code {table_name}} is already indexed. No action taken.",
        i = "Use `check = FALSE` to ignore this check."
      ))
      return(invisible(NULL))
    }
  }

  db_no_pool <- pool::poolCheckout(db)
  driver <- db$fetch()

  if (inherits(x = driver, "SQLiteConnection") == FALSE) {
    statement <- glue::glue_sql(
      "ALTER TABLE {`table_name`} MODIFY search VARCHAR(255);",
      table_name = table_name,
      .con = db_no_pool
    )

    index_odbc_result <- DBI::dbSendQuery(
      conn = db_no_pool,
      statement = statement
    )

    statement <- glue::glue_sql(
      "ALTER TABLE {`table_name`} MODIFY id VARCHAR(50);",
      table_name = table_name,
      .con = db_no_pool
    )

    index_odbc_result <- DBI::dbSendQuery(
      conn = db_no_pool,
      statement = statement
    )
  }

  statement <- glue::glue_sql(
    "CREATE INDEX search on {`table_name`} (search);",
    table_name = table_name,
    .con = db_no_pool
  )

  index_odbc_result <- DBI::dbExecute(
    conn = db_no_pool,
    statement = statement
  )

  pool::poolReturn(db_no_pool)
  if (show_details) {
    output <- tw_check_cache_index(
      table_name = table_name,
      type = type,
      show_details = TRUE,
      language = language,
      response_language = response_language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db
    )

    return(output)
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db
  )
}


#' Add index to caching table for search queries for increased speed
#'
#' Tested only with SQLite and MySql. May work with other drivers.
#'
#' To ensure smooth functioning, the search column in the cache table is
#' transformed into a column of type `varchar` and length 255.
#'
#' @param table_name Name of the table in the database. If given, it takes
#'   precedence over other parameters.
#' @param check_first Logical, defaults to `TRUE`. If `TRUE`, then before
#'   executing anything on the database it checks if the given table has already
#'   been indexed. If it has, it does nothing and returns only an informative
#'   message.
#' @param show_details Logical, defaults to `FALSE`. If `FALSE`, return the
#'   function adds the index to the database, but does not return anything. If
#'   `TRUE`, returns a data frame with more details about the index.
#'
#' @inheritParams tw_get_cache_table_name
#' @inheritParams tw_get
#'
#' @return If `show_details` is set to `FALSE`, nothing, used only for its side
#'   effects (add index to caching table). If `TRUE`, a data frame, same as the
#'   output of `tw_check_cache_index(show_details = TRUE)`.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_enable_cache()
#'   tw_set_cache_folder(path = fs::path(
#'     fs::path_home_r(),
#'     "R",
#'     "tw_data"
#'   ))
#'
#'   tw_index_cache_search()
#' }
tw_index_cache_item <- function(
  table_name = NULL,
  check_first = TRUE,
  type = "item",
  show_details = FALSE,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (is.null(table_name)) {
    table_name <- tw_get_cache_table_name(
      type = type,
      language = language
    )
  } else {
    language <- stringr::str_extract(string = table_name, "[a-z][a-z]$")
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    cli::cli_warn("Table `{table_name}` does not exist.")
    return(invisible())
  }

  if (check_first) {
    check <- tw_check_cache_index(
      table_name = table_name,
      type = type,
      show_details = FALSE,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )

    if (check) {
      cli::cli_alert_info(
        "The table {.code {table_name}} is already indexed. No action taken. Set `check` to FALSE to ignore this check."
      )
      return(invisible(NULL))
    }
  }

  db_no_pool <- pool::poolCheckout(db)

  driver <- db$fetch()

  if (inherits(x = driver, "SQLiteConnection") == FALSE) {
    statement <- glue::glue_sql(
      "ALTER TABLE {`table_name`} MODIFY id VARCHAR(50);",
      table_name = table_name,
      .con = db_no_pool
    )

    index_odbc_result <- DBI::dbSendQuery(
      conn = db_no_pool,
      statement = statement
    )

    statement <- glue::glue_sql(
      "ALTER TABLE {`table_name`} MODIFY property VARCHAR(50);",
      table_name = table_name,
      .con = db_no_pool
    )

    index_odbc_result <- DBI::dbSendQuery(
      conn = db_no_pool,
      statement = statement
    )
  }

  statement <- glue::glue_sql(
    "CREATE INDEX qid on {`table_name`} (id, property);",
    table_name = table_name,
    .con = db_no_pool
  )

  index_odbc_result <- DBI::dbSendQuery(
    conn = db_no_pool,
    statement = statement
  )

  pool::poolReturn(db_no_pool)

  if (show_details) {
    output <- tw_check_cache_index(
      table_name = table_name,
      type = type,
      show_details = TRUE,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )

    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db
    )

    return(output)
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db
  )
}

#' Return a connection to be used for caching
#'
#' @param connection Defaults to `NULL`. If `NULL`, uses local SQLite database.
#'   If given, must be a connection object or a list with relevant connection
#'   settings (see example).
#' @param RSQLite Defaults to `NULL`, expected either `NULL` or logical. If set
#'   to `FALSE`, details on the database connection must be given either as a
#'   named list in the connection parameter, or with [tw_set_cache_db()] as
#'   environment variables.
#' @inheritParams tw_get
#'
#' @return A connection object.
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   cache_connection <- pool::dbPool(
#'     RSQLite::SQLite(), # or e.g. odbc::odbc(),
#'     Driver = ":memory:", # or e.g. "MariaDB",
#'     Host = "localhost",
#'     database = "example_db",
#'     UID = "example_user",
#'     PWD = "example_pwd"
#'   )
#'   tw_connect_to_cache(cache_connection)
#'
#'
#'   db_settings <- list(
#'     driver = "MySQL",
#'     host = "localhost",
#'     server = "localhost",
#'     port = 3306,
#'     database = "tidywikidatar",
#'     user = "secret_username",
#'     pwd = "secret_password"
#'   )
#'
#'   tw_connect_to_cache(db_settings)
#' }
#' }
#'
tw_connect_to_cache <- function(
  connection = NULL,
  RSQLite = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL
) {
  if (!tw_check_cache(cache)) {
    return(NULL)
  }

  if (!is.null(connection) & !is.list(connection)) {
    if (!DBI::dbIsValid(connection)) {
      connection <- NULL
    }
  }

  if (is.null(connection)) {
    if (is.null(language)) {
      language <- tw_get_language()
    }

    if (is.null(RSQLite)) {
      RSQLite <- as.logical(Sys.getenv(x = "tw_cache_SQLite", unset = TRUE))
    }

    if (isTRUE(RSQLite)) {
      tw_check_cache_folder()
      db_file <- tw_get_cache_file(
        language = language
      )

      if (fs::file_exists(db_file) == FALSE) {
        db <- DBI::dbConnect(
          drv = RSQLite::SQLite(),
          db_file
        )
      }

      db <- pool::dbPool(
        drv = RSQLite::SQLite(),
        dbname = db_file
      )
      return(db)
    } else {
      connection <- tw_get_cache_db()

      if (connection[["driver"]] == "SQLite") {
        rlang::check_installed("RSQLite", "to use SQLite databases.")
        drv <- RSQLite::SQLite()
      } else {
        rlang::check_installed("odbc", "to use custom databases.")
        drv <- odbc::odbc()
      }

      db <- pool::dbPool(
        drv = drv,
        driver = connection[["driver"]],
        host = connection[["host"]],
        server = connection[["server"]],
        port = as.integer(connection[["port"]]),
        database = connection[["database"]],
        user = connection[["user"]],
        pwd = connection[["pwd"]]
      )
      return(db)
    }
  } else {
    if (is.list(connection)) {
      if (connection[["driver"]] == "SQLite") {
        rlang::check_installed("RSQLite", "to use SQLite databases.")
        drv <- RSQLite::SQLite()
      } else {
        rlang::check_installed("odbc", "to use custom databases.")
        drv <- odbc::odbc()
      }

      db <- pool::dbPool(
        drv = drv,
        driver = connection[["driver"]],
        host = connection[["host"]],
        server = connection[["server"]],
        port = as.integer(connection[["port"]]),
        database = connection[["database"]],
        dbname = connection[["database"]],
        user = connection[["user"]],
        pwd = connection[["pwd"]]
      )
      return(db)
    } else {
      return(connection)
    }
  }
}

#' Retrieve cached search
#'
#' @inheritParams tw_search
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#' \dontrun{
#'
#' tw_set_cache_folder(path = tempdir())
#' tw_enable_cache()
#' tw_create_cache_folder(ask = FALSE)
#'
#' search_from_api <- tw_search("Sylvia Pankhurst")
#' search_from_api
#'
#' df_from_cache <- tw_get_cached_search("Sylvia Pankhurst")
#' df_from_cache
#'
#' }
tw_get_cached_search <- function(
  search,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  include_search = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(tidywikidatar::tw_empty_search)
  }

  language_combo <- stringr::str_c(language, "_", response_language)

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language_combo,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = stringr::str_c("search_", type),
    language = language,
    response_language = response_language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language_combo
    )

    search_df <- tidywikidatar::tw_empty_search

    if (include_search) {
      return(search_df)
    } else {
      return(
        search_df %>%
          dplyr::select(-"search")
      )
    }
  }

  search_string <- search
  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(.data$search %in% !!search_string),
    error = function(e) {
      logical(1L)
    }
  )
  if (isFALSE(db_result)) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language_combo
    )
    search_df <- tidywikidatar::tw_empty_search

    if (include_search) {
      return(search_df)
    } else {
      return(
        search_df %>%
          dplyr::select(-"search")
      )
    }
  }

  if (include_search) {
    cached_items_df <- db_result %>%
      dplyr::collect()
  } else {
    cached_items_df <- db_result %>%
      dplyr::collect() %>%
      dplyr::select(-"search")
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language_combo
  )

  cached_items_df
}

#' Creates the base cache folder where `tidywikidatar` caches data.
#'
#' @inheritParams tw_reset_item_cache
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_create_cache_folder()
#' }
#' }
tw_create_cache_folder <- function(ask = TRUE) {
  if (!fs::file_exists(tidywikidatar::tw_get_cache_folder())) {
    if (!ask) {
      fs::dir_create(
        path = tidywikidatar::tw_get_cache_folder(),
        recurse = TRUE
      )
    } else {
      cli::cli_inform(c(
        "The cache folder {.path {tw_get_cache_folder()}} does not exist.",
        "If you prefer to cache files elsewhere, reply no and set your preferred cache folder with {.fn tw_set_cache_folder}."
      ))
      msg_yes_no <- cli::format_inline(
        "Do you want to create {.path {tw_get_cache_folder()}} for caching data?"
      )
      check <- utils::menu(choices = c("Yes", "No"), title = msg_yes_no)
      if (check == 1) {
        fs::dir_create(
          path = tidywikidatar::tw_get_cache_folder(),
          recurse = TRUE
        )
      }
    }
    if (!fs::file_exists(tidywikidatar::tw_get_cache_folder())) {
      cli::cli_abort(c(
        "This function requires a valid cache folder.",
        i = "Set one with {.fn tw_set_cache_folder}."
      ))
    }
  }
}


#' Set folder for caching data
#'
#' Consider using a folder out of your current project directory, e.g.
#' `tw_set_cache_folder("~/R/tw_data/")`: you will be able to use the same cache
#' in different projects, and prevent cached files from being sync-ed if you use
#' services such as Nextcloud or Dropbox.
#'
#' @param path A path to a location used for caching data. If the folder does
#'   not exist, it will be created.
#'
#' @return The path to the caching folder, if previously set; the same path as
#'   given to the function; or the default, `tw_data` is none is given.
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_set_cache_folder(fs::path(fs::path_home_r(), "R", "tw_data"))
#' }
#' }
tw_set_cache_folder <- function(path = NULL) {
  if (is.null(path)) {
    path <- Sys.getenv("tw_cache_folder")
  } else {
    Sys.setenv(tw_cache_folder = path)
  }
  if (path == "") {
    path <- fs::path("tw_data")
  }
  invisible(path)
}

#' @rdname tw_set_cache_folder
#' @examples
#' tw_get_cache_folder()
#' @export
tw_get_cache_folder <- tw_set_cache_folder


#' Set database connection settings for the session
#'
#'
#' @param db_settings A list of database connection settings (see example)
#' @param driver A database driver. Common database drivers include `MySQL`,
#'   `PostgreSQL`, and `MariaDB`. See `unique(odbc::odbcListDrivers()[[1]])` for
#'   a list of locally available drivers.
#' @param host Host address, e.g. "localhost". Different drivers use server or
#'   host parameter, only one of them is likely needed.
#' @param server Server address, e.g. "localhost". Different drivers use server
#'   or host parameter, only one of them is likely needed.
#' @param port Port to use to connect to the database.
#' @param database Database name.
#' @param user Database user name.
#' @param pwd Password for the database user.
#'
#' @return A list with all given parameters (invisibly).
#' @export
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   # Settings can be provided either as a list
#'   db_settings <- list(
#'     driver = "MySQL",
#'     host = "localhost",
#'     server = "localhost",
#'     port = 3306,
#'     database = "tidywikidatar",
#'     user = "secret_username",
#'     pwd = "secret_password"
#'   )
#'
#'   tw_set_cache_db(db_settings)
#'
#'   # or as parameters
#'
#'   tw_set_cache_db(
#'     driver = "MySQL",
#'     host = "localhost",
#'     server = "localhost",
#'     port = 3306,
#'     database = "tidywikidatar",
#'     user = "secret_username",
#'     pwd = "secret_password"
#'   )
#'
#'   # or ignoring fields that can be left to default values, such as "localhost" and port 3306
#'
#'   tw_set_cache_db(
#'     driver = "MySQL",
#'     database = "tidywikidatar",
#'     user = "secret_username",
#'     pwd = "secret_password"
#'   )
#' }
#' }
tw_set_cache_db <- function(
  db_settings = NULL,
  driver = NULL,
  host = NULL,
  server = NULL,
  port = NULL,
  database = NULL,
  user = NULL,
  pwd = NULL
) {
  if (is.null(db_settings)) {
    if (is.null(driver) == FALSE) {
      Sys.setenv(tw_db_driver = driver)
    }
    if (is.null(host) == FALSE) {
      Sys.setenv(tw_db_host = host)
    }
    if (is.null(server) == FALSE) {
      Sys.setenv(tw_db_host = server)
    }
    if (is.null(port) == FALSE) {
      Sys.setenv(tw_db_port = port)
    }
    if (is.null(database) == FALSE) {
      Sys.setenv(tw_db_database = database)
    }
    if (is.null(user) == FALSE) {
      Sys.setenv(tw_db_user = user)
    }
    if (is.null(pwd) == FALSE) {
      Sys.setenv(tw_db_pwd = pwd)
    }
    return(invisible(
      list(
        driver = driver,
        host = host,
        server = server,
        port = port,
        database = database,
        user = user,
        pwd = pwd
      )
    ))
  } else {
    if (!is.null(db_settings$driver)) {
      Sys.setenv(tw_db_driver = db_settings$driver)
    }
    if (!is.null(db_settings$host)) {
      Sys.setenv(tw_db_host = db_settings$host)
    }
    if (!is.null(db_settings$server)) {
      Sys.setenv(tw_db_server = db_settings$server)
    }
    if (!is.null(db_settings$port)) {
      Sys.setenv(tw_db_port = db_settings$port)
    }
    if (!is.null(db_settings$database)) {
      Sys.setenv(tw_db_database = db_settings$database)
    }
    if (!is.null(db_settings$user)) {
      Sys.setenv(tw_db_user = db_settings$user)
    }
    if (!is.null(db_settings$pwd)) {
      Sys.setenv(tw_db_pwd = db_settings$pwd)
    }
    return(invisible(db_settings))
  }
}

#' Get database connection settings from the environment
#'
#' Typically set with [tw_set_cache_db()].
#'
#' @return A list with all database parameters as stored in environment
#'   variables.
#' @export
#'
#' @examples
#'
#' tw_get_cache_db()
tw_get_cache_db <- function() {
  list(
    driver = Sys.getenv("tw_db_driver"),
    host = Sys.getenv("tw_db_host"),
    server = Sys.getenv("tw_db_server"),
    port = Sys.getenv("tw_db_port"),
    database = Sys.getenv("tw_db_database"),
    user = Sys.getenv("tw_db_user"),
    pwd = Sys.getenv("tw_db_pwd")
  )
}


#' Enable caching for the current session
#'
#' @param SQLite Logical, defaults to `TRUE`. Set to `FALSE` to use custom
#'   database options. See [tw_set_cache_db()] for details.
#'
#' @return Nothing, used for its side effects.
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_enable_cache()
#' }
#' }
tw_enable_cache <- function(SQLite = TRUE) {
  Sys.setenv(tw_cache = TRUE)
  Sys.setenv(tw_cache_SQLite = SQLite)
}


#' Disable caching for the current session
#'
#' @return Nothing, used for its side effects.
#' @export

#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_disable_cache()
#' }
#' }
tw_disable_cache <- function() {
  Sys.setenv(tw_cache = FALSE)
}

#' Check caching status in the current session, and override it upon request
#'
#' Mostly used internally in functions, exported for reference.
#'
#' @param cache Defaults to `NULL`. If `NULL`, checks current cache settings. If
#'   given, returns given value, ignoring cache.
#'
#' @return Either `TRUE` or `FALSE`, depending on current cache settings.
#' @export
#' @examples
#' \donttest{
#' if (interactive()) {
#'   tw_check_cache()
#' }
#' }
tw_check_cache <- function(cache = NULL) {
  if (!is.null(cache)) {
    return(as.logical(cache))
  }
  current_cache <- Sys.getenv("tw_cache", unset = FALSE)
  as.logical(current_cache)
}

#' Checks if cache folder exists, if not returns an informative message
#'
#' @return If the cache folder exists, returns `TRUE`. Otherwise throws an
#'   error.
#' @export
#'
#' @examples
#'
#' # If cache folder does not exist, it throws an error
#' tryCatch(tw_check_cache_folder(),
#'   error = function(e) {
#'     return(e)
#'   }
#' )
#'
#' # Create cache folder
#' tw_set_cache_folder(path = fs::path(
#'   tempdir(),
#'   "tw_cache_folder"
#' ))
#' tw_create_cache_folder(ask = FALSE)
#'
#' tw_check_cache_folder()
tw_check_cache_folder <- function() {
  if (!fs::file_exists(tw_get_cache_folder())) {
    cli::cli_abort(c(
      "Cache folder does not exist.",
      "i" = "Set it with {.fn tw_set_cache_folder} and create it with {.fn tw_create_cache_folder}."
    ))
  }
  TRUE
}


#' Ensure that connection to cache is disconnected consistently
#'
#' @param cache Defaults to NULL. If given, it should be given either `TRUE` or
#'   FALSE. Typically set with [tw_enable_cache()] or [tw_disable_cache()].
#' @param cache_connection Defaults to `NULL`. If `NULL`, and caching is enabled,
#'   `tidywikidatar` will use a local sqlite database. A custom connection to
#'   other databases can be given (see vignette `caching` for details).
#' @param disconnect_db Defaults to `TRUE`. If `FALSE`, leaves the connection to
#'   cache open.
#' @param language Defaults to language set with [tw_set_language()]; if not
#'   set, "en". Use "all_available" to keep all languages. For available
#'   language values, see the
#'   \href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{relevant
#'   Wikimedia page}.
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_get(
#'     id = c("Q180099"),
#'     language = "en"
#'   )
#'   tw_disconnect_from_cache()
#' }
tw_disconnect_from_cache <- function(
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  language = tidywikidatar::tw_get_language()
) {
  if (isFALSE(disconnect_db)) {
    return(invisible(NULL))
  }

  if (isTRUE(tw_check_cache(cache))) {
    db <- tw_connect_to_cache(
      connection = cache_connection,
      language = language,
      cache = cache
    )

    if (pool::dbIsValid(dbObj = db)) {
      if (inherits(db, "Pool")) {
        pool::poolClose(db)
      } else {
        DBI::dbDisconnect(db)
      }
    }
  }
}

# run only in interactive sessions due to issues with API timeouts
if (interactive()) {
  library("dplyr", warn.conflicts = FALSE)
  library("tidywikidatar")

  tw_set_language(language = "en")

  tw_enable_cache()
  tw_set_cache_folder(
    path = fs::path(
      fs::path_temp(),
      "tw_data_testing"
    )
  )
  tw_create_cache_folder(ask = FALSE)

  testthat::test_that("tw_get_wikipedia_category_members() works with sub-categories", {
    testthat::expect_true({
      sub_category_members_df <- tw_get_wikipedia_category_members(
        url = "https://en.wikipedia.org/wiki/Category:American_women_anthropologists",
        category = NULL,
        language = "en",
        cache = TRUE,
        type = "subcat"
      )

      nrow(sub_category_members_df) > 1
    })
  })

  testthat::test_that("tw_get_wikipedia_category_members() works when category, not url is given", {
    testthat::expect_true({
      category_members_df <- tw_get_wikipedia_category_members(
        category = "Category:Puerto Rican women anthropologists",
        language = "en",
        cache = TRUE
      )

      nrow(category_members_df) > 0
    })
  })

  testthat::test_that("tw_get_wikipedia_category_members() works when url, not category is given", {
    testthat::expect_true({
      category_members_df <- tw_get_wikipedia_category_members(
        url = "https://en.wikipedia.org/wiki/Category:Puerto Rican women anthropologists",
        category = NULL,
        type = "page",
        language = "en",
        cache = TRUE
      )

      nrow(category_members_df) > 0
    })
  })
}

library("testthat")

test_that("cache respects param when given", {
  testthat::skip_if_offline()

  expect_equal(
    object = tw_check_cache(cache = TRUE),
    expected = TRUE
  )
  expect_equal(
    object = tw_check_cache(cache = FALSE),
    expected = FALSE
  )
})


test_that(desc = "Cache file location is returned correctly", code = {
  testthat::skip_if_offline()

  expect_equal(
    object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()
      tw_get_cache_file(
        language = "en"
      )
    },
    expected = {
      fs::path(
        tempdir(),
        "tw_cache_en.sqlite"
      )
    }
  )
})

library("testthat")

test_that("tw_get_p_wide works with labels", {
  testthat::skip_if_offline()
  testthat::skip_on_cran()

  tw_set_cache_folder(path = tempdir())
  tw_enable_cache()

  expect_equal(
    object = {
      df <- tw_get_p_wide(
        id = c("Q180099", "Q228822", "Q191095"),
        p = c("P27", "P19", "P20"),
        both_id_and_label = FALSE,
        label = TRUE,
        property_label_as_column_name = TRUE,
        only_first = TRUE
      )

      l <- list(
        ncol = ncol(df),
        nrow = nrow(df),
        colnames = colnames(df)
      )

      l
    },
    expected = {
      list(
        ncol = 5,
        nrow = 3,
        colnames = c(
          "id",
          "label",
          "country_of_citizenship",
          "place_of_birth",
          "place_of_death"
        )
      )
    }
  )

  expect_equal(
    object = {
      df <- tw_get_p_wide(
        id = c("Q180099", "Q228822", "Q191095"),
        p = c("P27", "P19", "P20"),
        both_id_and_label = FALSE,
        label = TRUE,
        property_label_as_column_name = FALSE,
        only_first = TRUE
      )

      l <- list(
        ncol = ncol(df),
        nrow = nrow(df),
        colnames = colnames(df)
      )

      l
    },
    expected = {
      list(
        ncol = 5,
        nrow = 3,
        colnames = c(
          "id",
          "label",
          "P27",
          "P19",
          "P20"
        )
      )
    }
  )

  expect_equal(
    object = {
      df <- tw_get_p_wide(
        id = c("Q180099", "Q228822", "Q191095"),
        p = c("P27", "P19", "P20"),
        both_id_and_label = FALSE,
        label = TRUE,
        property_label_as_column_name = FALSE,
        only_first = FALSE
      )

      l <- list(
        ncol = ncol(df),
        nrow = nrow(df),
        colnames = colnames(df)
      )

      l
    },
    expected = {
      list(
        ncol = 5,
        nrow = 3,
        colnames = c(
          "id",
          "label",
          "P27",
          "P19",
          "P20"
        )
      )
    }
  )

  expect_equal(
    object = {
      df <- tw_get_p_wide(
        id = c("Q180099", "Q228822", "Q191095"),
        p = c("P27", "P19", "P20"),
        both_id_and_label = TRUE,
        label = TRUE,
        property_label_as_column_name = FALSE,
        only_first = FALSE
      )

      l <- list(
        ncol = ncol(df),
        nrow = nrow(df),
        colnames = colnames(df),
        country_id = df$P27[[1]]
      )

      l
    },
    expected = {
      list(
        ncol = 8,
        nrow = 3,
        colnames = c(
          "id",
          "label",
          "P27",
          "P27_label",
          "P19",
          "P19_label",
          "P20",
          "P20_label"
        ),
        country_id = "Q30"
      )
    }
  )

  expect_equal(
    object = {
      df <- tw_get_p_wide(
        id = c("Q180099", "Q228822", "Q191095"),
        p = c("P27", "P19", "P20"),
        both_id_and_label = TRUE,
        label = TRUE,
        property_label_as_column_name = TRUE,
        only_first = FALSE
      )

      l <- list(
        ncol = ncol(df),
        nrow = nrow(df),
        colnames = colnames(df),
        country_id = df$country_of_citizenship[[1]],
        country_label = df$country_of_citizenship_label[[1]]
      )

      l
    },
    expected = {
      list(
        ncol = 8,
        nrow = 3,
        colnames = c(
          "id",
          "label",
          "country_of_citizenship",
          "country_of_citizenship_label",
          "place_of_birth",
          "place_of_birth_label",
          "place_of_death",
          "place_of_death_label"
        ),
        country_id = "Q30",
        country_label = "United States"
      )
    }
  )
})

library("testthat")

test_that("check if qualifiers are returned correctly when more than 1 id and one property", {
  testthat::skip_if_offline()

  expect_equal(
    object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()
      tw_create_cache_folder(ask = FALSE)

      q_df <- tw_get_qualifiers(
        id = c("Q180099", "Q314252"),
        p = "P26",
        language = "en",
        id_l = tw_test_items
      )

      list(
        total_id = length(unique(q_df$id)),
        total_p = length(unique(q_df$property)),
        nrow_check = nrow(q_df) > 7
      )
    },
    expected = list(
      total_id = 2,
      total_p = 1,
      nrow_check = TRUE
    )
  )
})


test_that("check if qualifiers are returned correctly when one of qualifiers has string value type (not id, not time)", {
  testthat::skip_if_offline()

  expect_equal(
    object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()
      tw_create_cache_folder(ask = FALSE)

      q_df <- tw_get_qualifiers(
        id = c("Q4644021"),
        p = "P31",
        language = "en",
        id_l = tw_test_items
      )

      q_df %>%
        dplyr::filter(.data$qualifier_property == "P1545") %>%
        dplyr::pull(.data$qualifier_value)
    },
    expected = "7"
  )
})

if (Sys.getenv("tidywikidatar_testing") == "manual") {
  test_that("check if tw_get cached correctly deleted Wikidata items", {
    testthat::skip_if_offline()

    expect_true(object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()

      item <- tw_get(
        id = "Q4007665",
        cache = TRUE
      )
      cached_item_df <- tw_get_cached_item(id = "Q4007665")

      cached_item_df$property == "error"
    })
  })
}

# run only in interactive sessions due to occasion change of responses from API
if (interactive()) {
  library("testthat")
  test_that("check if property are returned correctly when more than 1 id and one property", {
    testthat::skip_if_offline()

    expect_equal(
      object = {
        tw_disable_cache()
        tw_get_property(
          id = c(
            "Q180099",
            "Q228822"
          ),
          p = "P31"
        ) %>%
          dplyr::pull(value)
      },
      expected = c(
        c("Q5", "Q5")
      )
    )
  })

  test_that("check if property are returned correctly and in the same as order as given when 1 id and more than one property", {
    testthat::skip_if_offline()

    expect_equal(
      object = {
        tw_disable_cache()
        tw_get_property(
          id = "Q180099",
          p = c("P31", "P27")
        ) %>%
          dplyr::pull(value)
      },
      expected = c(
        c("Q5", "Q30")
      )
    )

    expect_equal(
      object = {
        tw_disable_cache()
        tw_get_property(
          id = "Q180099",
          p = c("P27", "P31")
        ) %>%
          dplyr::pull(value)
      },
      expected = c(
        c("Q30", "Q5")
      )
    )
  })

  test_that("check if tw_get_property with invalid identifiers", {
    testthat::skip_if_offline()

    expect_true(object = {
      tw_disable_cache()
      tw_get_property(
        id = "non_qid_string",
        p = "P27"
      ) %>%
        is.null()
    })

    expect_true(object = {
      tw_disable_cache()
      tw_get_property_same_length(
        id = "non_qid_string",
        p = "P27"
      ) %>%
        is.na()
    })
  })

  test_that("check if property labels are returned correctly", {
    testthat::skip_if_offline()

    expect_equal(
      object = {
        tw_get_property_label(property = "P31")
      },
      expected = c(
        "instance of"
      )
    )

    expect_equal(
      object = {
        tw_get_property_label(property = c("P31", "P361"))
      },
      expected = c(
        "instance of",
        "part of"
      )
    )
  })

  test_that("check if property labels are returned correctly with cache", {
    testthat::skip_if_offline()

    expect_equal(
      object = {
        tw_set_cache_folder(path = tempdir())
        tw_enable_cache()
        tw_get_property_label(property = "P31")
      },
      expected = c(
        "instance of"
      )
    )

    expect_equal(
      object = {
        tw_set_cache_folder(path = tempdir())
        tw_get_property_label(property = c("P31", "P361"))
      },
      expected = c(
        "instance of",
        "part of"
      )
    )

    expect_equal(
      object = {
        tw_set_cache_folder(path = tempdir())
        tw_get_property_label(property = c("P31", "P361", "non_id_string", NA))
      },
      expected = c(
        "instance of",
        "part of",
        NA,
        NA
      )
    )
  })

  test_that("check if property descriptions are returned correctly", {
    testthat::skip_if_offline()

    expect_true(object = {
      tw_disable_cache()
      result <- tw_get_property_description(property = "P31")

      stringr::str_starts(string = result, pattern = "that class of")
    })

    expect_equal(
      object = {
        tw_disable_cache()
        tw_get_property_description(property = c("P31", "P361")) %>%
          stringr::word(end = 3)
      },
      expected = c(
        "that class of",
        "object of which"
      )
    )
  })

  test_that("check if property descriptions are returned correctly with cache", {
    testthat::skip_if_offline()

    expect_true(object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()
      result <- tw_get_property_description(property = "P31")

      stringr::str_starts(string = result, pattern = "that class of which")
    })

    expect_equal(
      object = {
        tw_disable_cache()
        tw_get_property_description(property = c("P31", "P361")) %>%
          stringr::word(end = 3)
      },
      expected = c(
        "that class of",
        "object of which"
      )
    )
  })

  test_that("check if property are returned correctly when multiple NA included", {
    testthat::skip_if_offline()

    expect_equal(
      object = {
        tw_get_property_same_length(
          id = c("Q180099", NA, "Q228822", NA),
          p = "P31",
          only_first = FALSE
        ) %>%
          unlist() %>%
          is.na() %>%
          sum()
      },
      expected = 2
    )
  })
}

library("testthat")
test_that("tw_get_description works with more than one id without cache", {
  test_id_df <- tw_get(
    id = c(
      "Q180099",
      "Q228822"
    ),
    language = "en",
    id_l = tw_test_items
  )

  expect_equal(
    object = {
      tw_disable_cache()
      description <- tw_get_description(
        id = c(
          "Q180099",
          "Q228822"
        ),
        language = "en",
        id_df = test_id_df
      )

      stringr::str_detect(
        string = description,
        pattern = "American anthropologist"
      )
    },
    c(TRUE, TRUE)
  )
})


test_that("tw_get_description works with one id with cache", {
  test_id_df <- tw_get(
    id = c(
      "Q180099",
      "Q228822"
    ),
    language = "en",
    id_l = tw_test_items
  )

  expect_true(
    object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()
      description <- tw_get_description(
        id = c("Q180099"),
        language = "en",
        id_df = test_id_df
      )

      stringr::str_detect(
        string = description,
        pattern = "American anthropologist"
      )
    }
  )
})

test_that("tw_get_description works with more than one id with cache", {
  test_id_df <- tw_get(
    id = c(
      "Q180099",
      "Q228822"
    ),
    language = "en",
    id_l = tw_test_items
  )

  expect_equal(
    object = {
      tw_set_cache_folder(path = tempdir())
      tw_enable_cache()
      description <- tw_get_description(
        id = c(
          "Q180099",
          "Q228822"
        ),
        language = "en",
        id_df = test_id_df
      )

      stringr::str_detect(
        string = description,
        pattern = "American anthropologist"
      )
    },
    c(TRUE, TRUE)
  )
})

library("testthat")

test_that("Preferred result is chosen when requested", {
  testthat::skip_if_offline()
  tw_set_cache_folder(path = tempdir())
  tw_enable_cache()
  tw_create_cache_folder(ask = FALSE)

  q_df <- tw_get_qualifiers_single(
    id = c("Q220"),
    p = "P17",
    language = "en"
  )

  expect_equal(
    object = {
      preferred_v <- q_df %>%
        dplyr::filter(rank == "preferred") %>%
        dplyr::pull(qualifier_id)
    },
    expected = {
      tw_get_p(id = "Q220", p = "P17", preferred = TRUE, only_first = TRUE) %>%
        unlist()
    }
  )
})

# run only in interactive sessions due to issues with API timeouts
if (interactive()) {
  library("dplyr", warn.conflicts = FALSE)
  library("tidywikidatar")

  tw_set_language(language = "en")

  tw_enable_cache()
  tw_set_cache_folder(
    path = fs::path(
      fs::path_temp(),
      "tw_data_testing"
    )
  )
  tw_create_cache_folder(ask = FALSE)

  testthat::test_that("tw_get_wikipedia_page_sections() works with url as input", {
    testthat::expect_true({
      sections_df <- tw_get_wikipedia_page_sections(
        url = "https://en.wikipedia.org/wiki/List_of_2020_Summer_Olympics_medal_winners",
        language = "en",
        cache = TRUE
      )

      sum(
        nrow(sections_df) > 1,
        "Archery" %in% sections_df[["line"]]
      ) ==
        2
    })
  })

  testthat::test_that("tw_get_wikipedia_page_section_links() works with title as input", {
    testthat::expect_true({
      section_links_df <- tw_get_wikipedia_page_section_links(
        title = "List of 2020 Summer Olympics medal winners",
        section_title = "Archery",
        language = "en",
        cache = TRUE
      )

      nrow(section_links_df) > 1
    })
  })

  testthat::test_that("tw_get_wikipedia_page_section_links() works with url as input", {
    testthat::expect_true({
      section_links_df <- tw_get_wikipedia_page_section_links(
        url = "https://en.wikipedia.org/wiki/List_of_2020_Summer_Olympics_medal_winners",
        section_title = "Archery",
        language = "en",
        cache = TRUE
      )

      nrow(section_links_df) > 1
    })
  })
}

#' Get links from a specific section of a Wikipedia page
#'
#' @param section_title Defaults to `NULL`. If given, it should correspond to
#'   the human-readable title of a section of the relevant Wikipedia page. See
#'   also [tw_get_wikipedia_page_sections()]
#' @param section_index Defaults to `NULL`. If given, it should correspond to
#'   the ordinal of a section of the relevant Wikipedia page. See also
#'   [tw_get_wikipedia_page_sections()]
#'
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_wikipedia_page_links_single
#' @return A data frame (a tibble).
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_section_links(title = "Margaret Mead", language = "en", section_index = 1)
#' }
tw_get_wikipedia_page_section_links <- function(
  url = NULL,
  title = NULL,
  section_title = NULL,
  section_index = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10,
  wikipedia_page_qid_df = NULL
) {
  if (is.null(section_index) && is.null(section_title)) {
    cli::cli_abort(c(
      "Either {.arg section_index} or {.arg section_title} must be provided.",
      i = "See also {.help tidywikidatar::tw_get_wikipedia_page_sections}."
    ))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  sections_df <- tw_get_wikipedia_page_sections(
    url = url,
    title = title,
    language = language,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  if (is.null(section_index)) {
    section_index <- sections_df %>%
      dplyr::filter(.data$line == section_title) %>%
      dplyr::pull("index") %>%
      utils::head(1)

    if (length(section_index) == 0) {
      cli::cli_abort(c(
        "Section title does not exist.",
        i = "Consider running `tw_get_wikipedia_sections(overwrite_cache = TRUE)` if you believe this may be due to oudated cache."
      ))
    }
  }

  if (is.null(section_title)) {
    section_title <- sections_df %>%
      dplyr::filter(.data$index == as.character(section_index)) %>%
      dplyr::pull("fromtitle") %>%
      utils::head(1)

    if (length(section_index) == 0) {
      cli::cli_abort(c(
        "Section title does not exist.",
        i = "Consider running `tw_get_wikipedia_sections(overwrite_cache = TRUE)` if you believe this may be due to oudated cache."
      ))
    }
  }

  wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
    title = title,
    language = language,
    url = url,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  json_url <- tw_get_wikipedia_section_links_api_url(
    url = url,
    title = title,
    language = language,
    section_index = section_index
  )

  api_result <- FALSE

  attempt_n <- 1

  while (isFALSE(api_result) & attempt_n <= attempts) {
    attempt_n <- sum(attempt_n, 1)
    api_result <- tryCatch(
      jsonlite::read_json(path = json_url),
      error = function(e) {
        logical(1L)
      }
    )
    Sys.sleep(time = wait)
  }

  if (isFALSE(api_result)) {
    cli::cli_abort(c(
      "Could not reach the API with {attempts} attempts.",
      i = "Consider increasing the waiting time between calls with the {.arg wait} parameter or check your internet connection."
    ))
  } else if ("error" %in% names(api_result)) {
    cli::cli_abort(
      "{api_result[['error']][['code']]}: {api_result[['error']][['info']]} - {json_url}"
    )
    api_result[["error"]]
  } else {
    base_json <- api_result
  }

  links_df <- purrr::map_dfr(
    .x = base_json %>%
      purrr::pluck("parse", "links"),
    .f = tibble::as_tibble_row
  )

  if (nrow(links_df) < 1) {
    return(tidywikidatar::tw_empty_wikipedia_page)
  }

  output_df <- tw_get_wikipedia_page_qid(
    title = links_df[["*"]],
    language = language,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  output_df
}


#' Facilitates the creation of MediaWiki API base URLs to retrieve sections of a
#' page
#'
#' Mostly used internally
#'
#' @param url A character vector with the full URL to one or more Wikipedia
#'   pages. If given, title and language can be left empty.
#' @param title Title of a Wikipedia page or final parts of its url. If given,
#'   url can be left empty, but language must be provided.
#' @param section_index Required. It should correspond to the ordinal of a
#'   section of the relevant Wikipedia page. See also
#'   [tw_get_wikipedia_page_sections()].
#' @param language Two-letter language code used to define the Wikipedia version
#'   to use. Defaults to language set with [tw_set_language()]; if not set,
#'   "en". If URL given, this can be left empty.
#'
#' @return A character vector of base urls to be used with the MediaWiki API
#' @export
#'
#' @examples
#' tw_get_wikipedia_section_links_api_url(title = "Margaret Mead", section_index = 1, language = "en")
tw_get_wikipedia_section_links_api_url <- function(
  url = NULL,
  title = NULL,
  section_index,
  language = tidywikidatar::tw_get_language()
) {
  stringr::str_c(
    tw_get_wikipedia_base_api_url(
      url = url,
      title = title,
      language = language,
      action = "parse"
    ),
    "&section=",
    section_index
  )
}

#' Get sections of a Wikipedia page
#'
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_wikipedia_page_links_single
#'
#' @return A data frame (a tibble), with the same columns as
#'   `tw_empty_wikipedia_page_sections`.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_sections(title = "Margaret Mead", language = "en")
#' }
tw_get_wikipedia_page_sections <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
    title = title,
    language = language,
    url = url,
    cache = cache,
    overwrite_cache = overwrite_cache,
    cache_connection = db,
    disconnect_db = FALSE,
    wait = wait,
    attempts = attempts
  )

  source_df <- wikipedia_page_qid_df %>%
    dplyr::transmute(
      source_title_url = .data$title_url,
      source_wikipedia_title = .data$wikipedia_title,
      source_qid = .data$qid,
      language = .data$language
    ) %>%
    dplyr::distinct(.data$source_title_url, .keep_all = TRUE)

  wikipedia_page_sections_df <- purrr::map_dfr(
    .x = seq_along(source_df$source_wikipedia_title),
    .f = function(i) {
      current_slice_df <- source_df %>%
        dplyr::slice(i)

      tw_get_wikipedia_page_sections_single(
        title = current_slice_df$source_title_url,
        language = current_slice_df$language,
        url = NULL,
        cache = cache,
        overwrite_cache = overwrite_cache,
        cache_connection = db,
        disconnect_db = FALSE,
        wait = wait,
        attempts = attempts
      )
    }
  )

  if (disconnect_db) {
    tw_disconnect_from_cache(
      cache = cache,
      cache_connection = db,
      disconnect_db = disconnect_db,
      language = language
    )
  }
  wikipedia_page_sections_df
}


#' Get all Wikidata Q identifiers of all Wikipedia pages that appear in a given
#' page
#'
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_wikipedia_page_links_single
#' @inheritParams tw_get_wikipedia_page_links_single
#'
#' @return A data frame (a tibble) with four columns: `wikipedia_title`,
#'   `wikipedia_id`, `wikidata_id`, `wikidata_description`.
#'
#' @examples
#' if (interactive()) {
#'   tw_get_wikipedia_page_sections_single(title = "Margaret Mead", language = "en")
#' }
tw_get_wikipedia_page_sections_single <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10,
  wikipedia_page_qid_df = NULL
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  if (is.null(wikipedia_page_qid_df)) {
    wikipedia_page_qid_df <- tw_get_wikipedia_page_qid(
      title = title,
      language = language,
      url = url,
      cache = cache,
      overwrite_cache = overwrite_cache,
      cache_connection = db,
      disconnect_db = FALSE,
      wait = wait,
      attempts = attempts
    )
  }

  if (tw_check_cache(cache) & overwrite_cache == FALSE) {
    db_result <- tw_get_cached_wikipedia_page_sections(
      title = title,
      language = language,
      cache = cache,
      cache_connection = db,
      disconnect_db = FALSE
    )
    if (is.data.frame(db_result) & nrow(db_result) > 0) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )

      return(
        db_result %>%
          dplyr::collect()
      )
    }
  }

  json_url <- tw_get_wikipedia_sections_api_url(
    url = url,
    title = title,
    language = language
  )

  api_result <- FALSE

  attempt_n <- 1

  while (isFALSE(api_result) & attempt_n <= attempts) {
    attempt_n <- sum(attempt_n, 1)
    api_result <- tryCatch(
      jsonlite::read_json(path = json_url),
      error = function(e) {
        logical(1L)
      }
    )
    Sys.sleep(time = wait)
  }

  if (isFALSE(api_result)) {
    cli::cli_abort(c(
      "Could not reach the API with {attempts} attempts.",
      i = "Consider increasing the waiting time between calls with the {.arg wait} parameter or check your internet connection."
    ))
  } else if ("error" %in% names(api_result)) {
    cli::cli_abort(
      "{api_result[['error']][['code']]}: {api_result[['error']][['info']]} - {json_url}"
    )
    api_result[["error"]]
  } else {
    base_json <- api_result
  }

  sections_df <- purrr::map_dfr(
    .x = base_json %>%
      purrr::pluck("parse", "sections"),
    .f = tibble::as_tibble_row
  )

  if (nrow(sections_df) < 1) {
    return(tidywikidatar::tw_empty_wikipedia_page_sections)
  }

  if (tw_check_cache(cache)) {
    tw_write_wikipedia_page_sections_to_cache(
      df = sections_df,
      cache_connection = db,
      language = language,
      overwrite_cache = overwrite_cache,
      disconnect_db = disconnect_db
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  sections_df
}


#' Gets sections of Wikipedia pages from local cache
#'
#' Mostly used internally.
#'
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_wikipedia_page_links_single
#'
#' @return If data present in cache, returns a data frame with cached data.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_set_cache_folder(path = tempdir())
#'   tw_enable_cache()
#'   tw_create_cache_folder(ask = FALSE)
#'
#'   df_from_api <- tw_get_wikipedia_page_qid(title = "Margaret Mead", language = "en")
#'
#'   df_from_cache <- tw_get_cached_wikipedia_page_sections(
#'     title = "Margaret Mead",
#'     language = "en"
#'   )
#'
#'   df_from_cache
#' }
tw_get_cached_wikipedia_page_sections <- function(
  title,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page_sections",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_page_sections)
  }

  db_result <- tryCatch(
    dplyr::tbl(src = db, table_name) %>%
      dplyr::filter(
        .data$fromtitle %in% !!stringr::str_c(title)
      ),
    error = function(e) {
      logical(1L)
    }
  )

  if (isFALSE(db_result)) {
    if (disconnect_db) {
      tw_disconnect_from_cache(
        cache = cache,
        cache_connection = db,
        disconnect_db = disconnect_db,
        language = language
      )
    }
    return(tidywikidatar::tw_empty_wikipedia_page_sections)
  }

  cached_df <- db_result %>%
    dplyr::collect()

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  cached_df
}


#' Write Wikipedia page links to cache
#'
#' Mostly used internally by `tidywikidatar`, use with caution to keep caching
#' consistent.
#'
#' @param df A data frame typically generated with
#'   [tw_get_wikipedia_page_sections()].
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get_wikipedia_page_links
#'
#' @return Silently returns the same data frame provided as input. Mostly used
#'   internally for its side effects.
#'
#' @export
#'
#' @examples
#' if (interactive()) {
#'   df <- tw_get_wikipedia_page_sections(
#'     title = "Margaret Mead",
#'     language = "en",
#'     cache = FALSE
#'   )
#'
#'   tw_write_wikipedia_page_sections_to_cache(
#'     df = df,
#'     language = "en"
#'   )
#' }
tw_write_wikipedia_page_sections_to_cache <- function(
  df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
) {
  if (isFALSE(tw_check_cache(cache = cache))) {
    return(invisible(NULL))
  }

  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page_sections",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, previous data cannot be there
  } else {
    if (overwrite_cache) {
      statement <- glue::glue_sql(
        "DELETE FROM {`table_name`} WHERE fromtitle = {fromtitle*}",
        fromtitle = unique(df$fromtitle),
        table_name = table_name,
        .con = db
      )
      result <- pool::dbExecute(
        conn = db,
        statement = statement
      )
    }
  }

  pool::dbWriteTable(db, name = table_name, value = df, append = TRUE)

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )

  invisible(df)
}

#' Reset Wikipedia page link cache
#'
#' Removes from cache the table where data typically gathered with
#' [tw_get_wikipedia_page_sections()] are stored.
#'
#' @inheritParams tw_get
#' @inheritParams tw_reset_item_cache
#'
#' @return Nothing, used for its side effects.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   tw_reset_wikipedia_page_sections_cache()
#' }
tw_reset_wikipedia_page_sections_cache <- function(
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
) {
  db <- tw_connect_to_cache(
    connection = cache_connection,
    language = language,
    cache = cache
  )

  table_name <- tw_get_cache_table_name(
    type = "wikipedia_page_sections",
    language = language
  )

  if (pool::dbExistsTable(conn = db, name = table_name) == FALSE) {
    # do nothing: if table does not exist, nothing to delete
  } else if (isFALSE(ask)) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Wikipedia page sections cache reset for language {.val {language}} completed."
    )
  } else if (
    utils::menu(
      c("Yes", "No"),
      title = paste0(
        "Are you sure you want to remove from cache the Wikipedia page links cache for language: ",
        sQuote(language),
        "?"
      )
    ) ==
      1
  ) {
    pool::dbRemoveTable(conn = db, name = table_name)
    cli::cli_alert_info(
      "Wikipedia page sections cache reset for language {.val {language}} completed."
    )
  }

  tw_disconnect_from_cache(
    cache = cache,
    cache_connection = db,
    disconnect_db = disconnect_db,
    language = language
  )
}

#' Facilitates the creation of MediaWiki API base URLs to retrieve sections of a
#' page
#'
#' Mostly used internally
#'
#' @inheritParams tw_get_image
#' @inheritParams tw_get_image_metadata
#' @inheritParams tw_get_wikipedia_page_links
#' @inheritParams tw_get_wikipedia_page_links_single
#'
#' @return A character vector of base urls to be used with the MediaWiki API
#' @export
#'
#' @examples
#' tw_get_wikipedia_sections_api_url(title = "Margaret Mead", language = "en")
tw_get_wikipedia_sections_api_url <- function(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language()
) {
  stringr::str_c(
    tw_get_wikipedia_base_api_url(
      url = url,
      title = title,
      language = language,
      action = "parse"
    ),
    "&prop=sections"
  )
}

---
title: "How caching works in tidywikidatar"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How caching works in tidywikidatar}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(tidywikidatar)
```

In order to reduce load on Wikidata's server and to speed up the processing of data, `tidywikidatar` makes extensive use of local caching.

## What data are cached locally

There are a few types of data that are cached locally:

- searches run with `tw_search()`
- data about an item, typically retrieved with `tw_get()` or `tw_get_property()`
- labels or description of properties, typically retrieved with `tw_get_property_label()` and `tw_get_property_description()`
- qualifiers of properties, typically retrieved with `tw_get_qualifiers()`
- data retrieved from (or about) Wikipedia pages, with `tw_get_wikipedia()`, and `tw_get_wikipedia_page_links()`

To reduce space used for local caching and speed up processing time, it is possible to store only labels and information available in a given language when relevant. 


## Caching with SQLite

In `tidywikidatar`, it is possible to enable caching with:

```{r}
tw_enable_cache()
```


If you do not include further parameters, by default `tidywikidatar` will use a local SQLite database for caching.

You can choose in which folder the SQLite database will be stored with `tw_set_cache_folder()`; if not already existing, you can create that folder with `tw_create_cache_folder()`.

```{r eval=FALSE}
tw_set_cache_folder(path = fs::path(
  fs::path_home_r(),
  "R",
  "tw_data"
))
tw_create_cache_folder()
```


## Caching with other database backends

Support for other database backends is now available. They can be accessed most easily using the following approach, having ensured that the relevant driver (and `odbc` package) have previously been installed:

```{r eval = FALSE}
tw_enable_cache(SQLite = FALSE)
tw_set_cache_db(
  driver = "MySQL",
  host = "localhost",
  port = 3306,
  database = "tidywikidatar",
  user = "secret_username",
  pwd = "secret_password"
)


# for testing, consider running a local database e.g. with:
# docker run --name tidywikidatar_db -p 3306:3306 -e MYSQL_ROOT_PASSWORD=secret_root_password -e MYSQL_USER=secret_username -e MYSQL_PASSWORD=secret_password -e MYSQL_DATABASE=tidywikidatar mysql:latest
```

It is also technically possible to pass directly a connection generated with `DBI::dbConnect()` to each function. 


## Name of tables in cached databases

Each database has a table for each language and type of content. For example, item information retrieved with `tw_get(id = "Q180099", language = "en")` will be stored in a table called `tw_item_en`.

The name of the table is unique and is generated by `tw_get_cache_table_name()`. For example:

```{r}
tw_get_cache_table_name(type = "item", language = "en")
```


## Column types and indexing

Due to limited familiarity with different database backends and limited time for testing, the creation of database tables is left to the default values of `DBI::dbWriteTable()`. For occasional use, this should not be an issue. However, when the local cache reaches millions rather than only thousands of rows, response time from a MySql database can take a few seconds, rather than a fraction of a second as would be expected. To deal with this, new functions for adding indexing to cache tables have been introduced, `tw_check_cache_index()`, `tw_index_cache_item()`, and `tw_index_cache_search()`. It is possible to apply this to all existing functions of a given type as outlined below. This speeds up retrieval time dramatically on MySql databases; impact on other types of databases has not been thoroughly tested.

```{r eval=FALSE}
db <- tw_connect_to_cache()


tables_v <- DBI::dbListTables(conn = db)

# for search cache tables
purrr::walk(
  .x = tables_v[stringr::str_starts(string = tables_v, "tw_search_item")],
  .f = function(x) {
    tw_index_cache_search(table_name = x)
  }
)

# for item cache tables
purrr::walk(
  .x = tables_v[stringr::str_starts(string = tables_v, "tw_item")],
  .f = function(x) {
    tw_index_cache_item(table_name = x)
  }
)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_cached_item.R
\name{tw_get_cache_file}
\alias{tw_get_cache_file}
\title{Gets location of cache file}
\usage{
tw_get_cache_file(
  extension = "sqlite",
  language = tidywikidatar::tw_get_language()
)
}
\arguments{
\item{extension}{Defaults to \code{sqlite}. Extension of the database cache file.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}
}
\value{
A character vector of length one with location of item cache file.
}
\description{
Gets location of cache file
}
\examples{

tw_set_cache_folder(path = tempdir())
sqlite_cache_file_location <- tw_get_cache_file() # outputs location of cache file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_item.R
\name{tw_get_item}
\alias{tw_get_item}
\title{Retrieve item from the Wikidata API and returns it as a list}
\usage{
tw_get_item(id, user_agent = tidywikidatar::tw_get_user_agent(), retry = 10)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}
}
\value{
A list, with as many elements as the unique given id.
}
\description{
Retrieve item from the Wikidata API and returns it as a list
}
\examples{
\dontrun{
  item_l <- tw_get_item(id = "Q180099")

  tidywikidatar:::tw_extract_single(w = item_l)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_page_sections.R
\name{tw_get_wikipedia_sections_api_url}
\alias{tw_get_wikipedia_sections_api_url}
\title{Facilitates the creation of MediaWiki API base URLs to retrieve sections of a
page}
\usage{
tw_get_wikipedia_sections_api_url(
  url = NULL,
  title = NULL,
  language = tidywikidatar::tw_get_language()
)
}
\arguments{
\item{url}{Full url to a Wikipedia page. If given, title and language can be
left empty.}

\item{title}{Title of a Wikipedia page or final parts of its url. If given,
url can be left empty, but language must be provided.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}
}
\value{
A character vector of base urls to be used with the MediaWiki API
}
\description{
Mostly used internally
}
\examples{
tw_get_wikipedia_sections_api_url(title = "Margaret Mead", language = "en")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_cached_search.R
\name{tw_get_cached_search}
\alias{tw_get_cached_search}
\title{Retrieve cached search}
\usage{
tw_get_cached_search(
  search,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  include_search = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{search}{A string to be searched in Wikidata}

\item{type}{Defaults to "item". Either "item" or "property".}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE.} Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{include_search}{Logical, defaults to \code{FALSE}. If \code{TRUE}, the search is
returned as an additional column.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
If data present in cache, returns a data frame with cached data.
}
\description{
Retrieve cached search
}
\examples{
\dontrun{

tw_set_cache_folder(path = tempdir())
tw_enable_cache()
tw_create_cache_folder(ask = FALSE)

search_from_api <- tw_search("Sylvia Pankhurst")
search_from_api

df_from_cache <- tw_get_cached_search("Sylvia Pankhurst")
df_from_cache

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_create_cache_folder}
\alias{tw_create_cache_folder}
\title{Creates the base cache folder where \code{tidywikidatar} caches data.}
\usage{
tw_create_cache_folder(ask = TRUE)
}
\arguments{
\item{ask}{Logical, defaults to \code{TRUE}. If \code{FALSE}, and cache folder does not
exist, it just removes it without asking (useful for non-interactive
sessions).}
}
\value{
Nothing, used for its side effects.
}
\description{
Creates the base cache folder where \code{tidywikidatar} caches data.
}
\examples{
\donttest{
if (interactive()) {
  tw_create_cache_folder()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_page_links.R
\name{tw_write_wikipedia_page_links_to_cache}
\alias{tw_write_wikipedia_page_links_to_cache}
\title{Write Wikipedia page links to cache}
\usage{
tw_write_wikipedia_page_links_to_cache(
  df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{df}{A data frame typically generated with
\code{\link[=tw_get_wikipedia_page_links]{tw_get_wikipedia_page_links()}}.}

\item{language}{Two-letter language code used to define the Wikipedia version
to use. Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not set,
"en". If url given, this can be left empty.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
Silently returns the same data frame provided as input. Mostly used
internally for its side effects.
}
\description{
Mostly used internally by \code{tidywikidatar}, use with caution to keep caching
consistent.
}
\examples{
if (interactive()) {
  df <- tw_get_wikipedia_page_links(
    title = "Margaret Mead",
    language = "en",
    cache = FALSE
  )

  tw_write_wikipedia_page_links_to_cache(
    df = df,
    language = "en"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_property_with_details.R
\name{tw_get_property_with_details}
\alias{tw_get_property_with_details}
\title{Gets all details of a property for one or more Wikidata items.}
\usage{
tw_get_property_with_details(
  id,
  p,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{p}{A character vector, a property. Must always start with the capital
letter "P", e.g. "P31" for "instance of".}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}
}
\value{
A tibble, corresponding to the details for the given property. \code{NULL}
if no relevant property found.
}
\description{
Gets all details of a property for one or more Wikidata items.
}
\examples{
# Get "female form of label", including language
tw_get_property_with_details(id = "Q64733534", p = "P2521")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_property.R
\name{tw_get_property}
\alias{tw_get_property}
\title{Get Wikidata property of one or more items as a tidy data frame}
\usage{
tw_get_property(
  id,
  p,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{p}{A character vector, a property. Must always start with the capital
letter "P", e.g. "P31" for "instance of".}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{id_df}{Default to \code{NULL}. If given, it should be a dataframe typically
generated with \code{\link[=tw_get]{tw_get()}}, and is used instead of calling Wikidata or
using SQLite cache. Ignored when \code{id} is of length more than one.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}
}
\value{
A tibble, corresponding to the value for the given property. A tibble of zero rows if no relevant property found.
}
\description{
Get Wikidata property of one or more items as a tidy data frame
}
\examples{
\dontrun{
# Who were the doctoral advisors - P184 - of Margaret Mead - Q180099?
advisors <- tw_get_property(id = "Q180099", p = "P184")
advisors

# tw_get_label(advisors)

# It is also possible to get one property for many id

  tw_get_property(
    id = c(
      "Q180099",
      "Q228822"
    ),
    p = "P31"
  )

  # Or many properties for a single id

  tw_get_property(
    id = "Q180099",
    p = c("P21", "P31")
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_filter.R
\name{tw_filter_people}
\alias{tw_filter_people}
\title{Filter search result and keep only people}
\usage{
tw_filter_people(
  search,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  stop_at_first = TRUE,
  wait = 0,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{search}{A data frame generated by \code{tw_search()}, or a search query. If
a data frame is given, language and limits are ignore.}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}

\item{limit}{Maximum numbers of responses to be given.}

\item{include_search}{Logical, defaults to \code{FALSE}. If \code{TRUE}, the search is
returned as an additional column.}

\item{stop_at_first}{Logical, defaults to TRUE. If TRUE, returns only the
first match from the search that satisfies the criteria.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{overwrite_cache}{Defaults to \code{FALSE}. If \code{TRUE}, overwrites cache.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
A data frame with three columns, \code{id}, \code{label}, and \code{description};
all rows refer to a human being.
}
\description{
A wrapper of \code{\link[=tw_filter]{tw_filter()}} that defaults to keep only "instance of" (P31)
"human being" (Q5).
}
\examples{
\dontrun{
  tw_search("Ruth Benedict")

  tw_search("Ruth Benedict") \%>\%
    tw_filter_people()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{tw_empty_wikipedia_page}
\alias{tw_empty_wikipedia_page}
\title{A zero-rows tibble used internally when \code{tw_get_wikipedia_page_qid()} would not return any value.}
\format{
A data frame with 0 rows and 6 columns
}
\usage{
tw_empty_wikipedia_page
}
\description{
A zero-rows tibble used internally when \code{tw_get_wikipedia_page_qid()} would not return any value.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_page_section_links.R
\name{tw_get_wikipedia_page_section_links}
\alias{tw_get_wikipedia_page_section_links}
\title{Get links from a specific section of a Wikipedia page}
\usage{
tw_get_wikipedia_page_section_links(
  url = NULL,
  title = NULL,
  section_title = NULL,
  section_index = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10,
  wikipedia_page_qid_df = NULL
)
}
\arguments{
\item{url}{Full url to a Wikipedia page. If given, title and language can be
left empty.}

\item{title}{Title of a Wikipedia page or final parts of its url. If given,
url can be left empty, but language must be provided.}

\item{section_title}{Defaults to \code{NULL}. If given, it should correspond to
the human-readable title of a section of the relevant Wikipedia page. See
also \code{\link[=tw_get_wikipedia_page_sections]{tw_get_wikipedia_page_sections()}}}

\item{section_index}{Defaults to \code{NULL}. If given, it should correspond to
the ordinal of a section of the relevant Wikipedia page. See also
\code{\link[=tw_get_wikipedia_page_sections]{tw_get_wikipedia_page_sections()}}}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{attempts}{Defaults to 10. Number of times it re-attempts to reach the
API before failing.}

\item{wikipedia_page_qid_df}{Defaults to \code{NULL}. If given, used to reduce
calls to cache. Must be a data frame.}
}
\value{
A data frame (a tibble).
}
\description{
Get links from a specific section of a Wikipedia page
}
\examples{
if (interactive()) {
  tw_get_wikipedia_page_section_links(title = "Margaret Mead", language = "en", section_index = 1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_category_members.R
\name{tw_get_cached_wikipedia_category_members}
\alias{tw_get_cached_wikipedia_category_members}
\title{Gets members of Wikipedia categories from local cache}
\usage{
tw_get_cached_wikipedia_category_members(
  category,
  type = "page",
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{category}{Title of a Wikipedia category page or final parts of its url.
Must include "Category:", or equivalent in other languages. If given, url
can be left empty, but language must be provided.}

\item{type}{Defaults to "page", defines which kind of members of a category
are returned. Valid values include "page", "file", and "subcat" (for
sub-category). Corresponds to \code{cmtype}. For details, see
\href{https://www.mediawiki.org/wiki/API:Categorymembers}{the relevant page
of the official documentation}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
If data present in cache, returns a data frame with cached data.
}
\description{
Mostly used internally.
}
\examples{
if (interactive()) {
  tw_set_cache_folder(path = tempdir())
  tw_enable_cache()
  tw_create_cache_folder(ask = FALSE)

  df_from_api <- tw_get_wikipedia_page_qid(category = "Margaret Mead", language = "en")

  df_from_cache <- tw_get_cached_wikipedia_category_members(
    category = "Margaret Mead",
    language = "en"
  )

  df_from_cache
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_cached_item.R
\name{tw_get_cache_table_name}
\alias{tw_get_cache_table_name}
\title{Gets name of table inside the database}
\usage{
tw_get_cache_table_name(
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language()
)
}
\arguments{
\item{type}{Defaults to "item". Type of cache file to output. Values
typically used by \code{tidywikidatar} include "item", "search_item",
"search_property", and "qualifier".}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}
}
\value{
A character vector of length one with the name of the relevant table
in the cache file.
}
\description{
Gets name of table inside the database
}
\examples{
# outputs name of table used in the cache database
tw_get_cache_table_name(type = "item", language = "en")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_language_settings.R
\name{tw_set_language}
\alias{tw_set_language}
\alias{tw_get_language}
\title{Set language to be used by all functions}
\usage{
tw_set_language(language = NULL)

tw_get_language(language = NULL)
}
\arguments{
\item{language}{A character vector of length one, with a string of two letters such as "en". For a full list of available values, see: https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}
}
\value{
A two letter code for the language, if previously set; the same language as given to the function; or the default, \code{en} is none is given.
}
\description{
Defaults to "en".
}
\examples{
\donttest{
if (interactive()) {
  tw_set_language(language = "en")
}
}

tw_get_language()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{tw_qid_airports}
\alias{tw_qid_airports}
\title{The Wikidata Q identifier of European airports found in Eurostat's \code{avia_par_} dataset}
\format{
A data frame with 429 rows and 1 column:
\describe{
\item{id}{Q identifiers}
}
}
\source{
\url{https://www.wikidata.org/wiki/Wikidata:Main_Page}
}
\usage{
tw_qid_airports
}
\description{
The Wikidata Q identifier of European airports found in Eurostat's \code{avia_par_} dataset
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_extract.R
\name{tw_extract_single}
\alias{tw_extract_single}
\title{Extract item data from a list object retrieved from the Wikidata API}
\usage{
tw_extract_single(w, language = tidywikidatar::tw_get_language())
}
\arguments{
\item{w}{A list, typically retrieved with \code{\link[=tw_get_item]{tw_get_item()}}}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}
}
\value{
A data frame (a tibble) with four columns, such as the one created by
\code{\link[=tw_get]{tw_get()}}.
}
\description{
This function is mostly used internally and for testing.
}
\examples{
#' Retrieving from tests, but normally:
# w <- tw_get_item(id = "Q180099")

tidywikidatar:::tw_extract_single(w = list(tw_test_items[["Q180099"]]))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_set_cache_folder}
\alias{tw_set_cache_folder}
\alias{tw_get_cache_folder}
\title{Set folder for caching data}
\usage{
tw_set_cache_folder(path = NULL)

tw_get_cache_folder(path = NULL)
}
\arguments{
\item{path}{A path to a location used for caching data. If the folder does
not exist, it will be created.}
}
\value{
The path to the caching folder, if previously set; the same path as
given to the function; or the default, \code{tw_data} is none is given.
}
\description{
Consider using a folder out of your current project directory, e.g.
\code{tw_set_cache_folder("~/R/tw_data/")}: you will be able to use the same cache
in different projects, and prevent cached files from being sync-ed if you use
services such as Nextcloud or Dropbox.
}
\examples{
\donttest{
if (interactive()) {
  tw_set_cache_folder(fs::path(fs::path_home_r(), "R", "tw_data"))
}
}
tw_get_cache_folder()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_check_cache}
\alias{tw_check_cache}
\title{Check caching status in the current session, and override it upon request}
\usage{
tw_check_cache(cache = NULL)
}
\arguments{
\item{cache}{Defaults to \code{NULL}. If \code{NULL}, checks current cache settings. If
given, returns given value, ignoring cache.}
}
\value{
Either \code{TRUE} or \code{FALSE}, depending on current cache settings.
}
\description{
Mostly used internally in functions, exported for reference.
}
\examples{
\donttest{
if (interactive()) {
  tw_check_cache()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_property_label.R
\name{tw_get_property_label}
\alias{tw_get_property_label}
\title{Get label of a Wikidata property in a given language}
\usage{
tw_get_property_label(
  property,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
)
}
\arguments{
\item{property}{A character vector. Each element must start with P, e.g.
"P31".}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}
}
\value{
A character vector, with the Wikidata label in the requested
language.
}
\description{
Get label of a Wikidata property in a given language
}
\examples{
tw_get_property_label(property = "P31")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_query.R
\name{tw_query}
\alias{tw_query}
\title{Perform simple Wikidata queries}
\usage{
tw_query(
  query,
  fields = c("item", "itemLabel", "itemDescription"),
  language = tidywikidatar::tw_get_language(),
  return_as_tw_search = TRUE,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{query}{A list of named vectors, or a data frame (see example and
readme).}

\item{fields}{A character vector of Wikidata fields. Ignored if
\code{return_as_tw_search} is set to \code{TRUE} (as per default). Defaults to
\verb{("item", "itemLabel", "itemDescription")}}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". If more than one, can be set in order of preference, e.g.
\code{c("it", "fr", "en")}. Use "all_available" to keep all languages. For
available language values, see the
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{list
of available language codes in the official documentation}.}

\item{return_as_tw_search}{Logical, defaults to \code{TRUE}. If \code{TRUE}, returns a
data frame with three columns (\code{id}, \code{label}, and \code{description}) that can
be piped to other \code{tw_} functions. If \code{FALSE}, a data frame with as many
columns as fields.}

\item{user_agent}{Defaults to a combination of \code{tidywikidatar} and package
version number. Consider customising it for the current sessions with
\code{\link[=tw_set_user_agent]{tw_set_user_agent()}}, in particular if you are making many queries.}
}
\value{
A data frame
}
\description{
This function aims to facilitate only the most basic type of queries: return
which items have the following property pairs. For more details on Wikidata
queries, the
\href{https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service/queries/examples}{examples
in the official documentation}.
}
\details{
Consider \code{\link[=tw_get_all_with_p]{tw_get_all_with_p()}} if you want to get all items with a given
property, irrespective of the value.
}
\examples{

if (interactive()) {
  query <- list(
    c(p = "P106", q = "Q1397808"),
    c(p = "P21", q = "Q6581072")
  )
  tw_query(query)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_category_members.R
\name{tw_get_wikipedia_category_members_single}
\alias{tw_get_wikipedia_category_members_single}
\title{Get all Wikidata Q identifiers of all Wikipedia pages that appear in a given page}
\usage{
tw_get_wikipedia_category_members_single(
  url = NULL,
  category = NULL,
  type = "page",
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 1,
  attempts = 10
)
}
\arguments{
\item{url}{Full URL to a Wikipedia category page. If given, title and
language can be left empty.}

\item{category}{Title of a Wikipedia category page or final parts of its url.
Must include "Category:", or equivalent in other languages. If given, url
can be left empty, but language must be provided.}

\item{type}{Defaults to "page", defines which kind of members of a category
are returned. Valid values include "page", "file", and "subcat" (for
sub-category). Corresponds to \code{cmtype}. For details, see
\href{https://www.mediawiki.org/wiki/API:Categorymembers}{the relevant page
of the official documentation}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{attempts}{Defaults to 10. Number of times it re-attempts to reach the
API before failing.}
}
\value{
A data frame (a tibble) with four columns: \code{wikipedia_title}, \code{wikipedia_id}, \code{wikidata_id}, \code{wikidata_description}.
}
\description{
Get all Wikidata Q identifiers of all Wikipedia pages that appear in a given page
}
\examples{
if (interactive()) {
  tidywikidatar:::tw_get_wikipedia_category_members_single(
    category = "Category:American women anthropologists",
    type = "subcat"
  )

  tidywikidatar:::tw_get_wikipedia_category_members_single(
    category = "Category:Puerto Rican women anthropologists",
    type = "page"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_all_with_p.R
\name{tw_get_all_with_p}
\alias{tw_get_all_with_p}
\title{Get all items that have a given property (irrespective of the value)}
\usage{
tw_get_all_with_p(
  p,
  fields = c("item", "itemLabel", "itemDescription"),
  language = tidywikidatar::tw_get_language(),
  method = "SPARQL",
  wait = 0.1,
  limit = Inf,
  return_as_tw_search = TRUE,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{p}{A character vector, a property. Must always start with the capital
letter "P", e.g. "P31" for "instance of".}

\item{fields}{A character vector of Wikidata fields. Ignored if
\code{return_as_tw_search} is set to \code{TRUE} (as per default). Defaults to
\verb{("item", "itemLabel", "itemDescription")}}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". If more than one, can be set in order of preference, e.g.
\code{c("it", "fr", "en")}. Use "all_available" to keep all languages. For
available language values, see the
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{list
of available language codes in the official documentation}.}

\item{method}{Defaults to "SPARQL". The only accepted alternative value is
"JSON", to use instead json-based API.}

\item{wait}{Defaults to 0.1. Used only in method is set to "JSON".}

\item{limit}{Defaults to \code{Inf}. Set to smaller values for testing and cache
locally when possible to reduce load on servers.}

\item{return_as_tw_search}{Logical, defaults to \code{TRUE}. If \code{TRUE}, returns a
data frame with three columns (\code{id}, \code{label}, and \code{description}) that can
be piped to other \code{tw_} functions. If \code{FALSE}, a data frame with as many
columns as fields.}

\item{user_agent}{Defaults to a combination of \code{tidywikidatar} and package
version number. Consider customising it for the current sessions with
\code{\link[=tw_set_user_agent]{tw_set_user_agent()}}, in particular if you are making many queries.}
}
\value{
A data frame with three columns is method is set to "SPARQL", or as
many columns as fields if more are given and \code{return_as_tw_search} is set
to \code{FALSE}. A single column with Wikidata identifier if method is set to
"JSON".
}
\description{
This function does not cache results.
}
\examples{
if (interactive()) {
  # get all Wikidata items with an ICAO airport code ("P239")
  tw_get_all_with_p(p = "P239", limit = 10)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_cached_item.R
\name{tw_get_cached_item}
\alias{tw_get_cached_item}
\title{Retrieve cached item}
\usage{
tw_get_cached_item(
  id,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
If data present in cache, returns a data frame with cached data.
}
\description{
Retrieve cached item
}
\examples{

tw_set_cache_folder(path = tempdir())
tw_enable_cache()
tw_create_cache_folder(ask = FALSE)

df_from_api <- tw_get(id = "Q180099", language = "en")

df_from_cache <- tw_get_cached_item(
  id = "Q180099",
  language = "en"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_cached_item.R
\name{tw_check_cached_items}
\alias{tw_check_cached_items}
\title{Check if given items are present in cache}
\usage{
tw_check_cached_items(
  id,
  language = tidywikidatar::tw_get_language(),
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
A character vector with IDs of items present in cache. If no item
found in cache, returns \code{NULL}.
}
\description{
Check if given items are present in cache
}
\examples{
if (interactive()) {
  tw_set_cache_folder(path = tempdir())
  tw_enable_cache()
  tw_create_cache_folder(ask = FALSE)

  # add three items to local cache
  invisible(tw_get(id = "Q180099", language = "en"))
  invisible(tw_get(id = "Q228822", language = "en"))
  invisible(tw_get(id = "Q184992", language = "en"))

  # check if these other items are in cache
  items_in_cache <- tw_check_cached_items(
    id = c(
      "Q180099",
      "Q228822",
      "Q76857"
    ),
    language = "en"
  )
  # it should return only the two items from the current list of id
  # but not other item already in cache
  items_in_cache
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_page_qid.R
\name{tw_reset_wikipedia_page_cache}
\alias{tw_reset_wikipedia_page_cache}
\title{Reset Wikipedia page cache}
\usage{
tw_reset_wikipedia_page_cache(
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
)
}
\arguments{
\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{ask}{Logical, defaults to \code{TRUE}. If \code{FALSE}, and cache folder does not
exist, it just removes it without asking (useful for non-interactive
sessions).}
}
\value{
Nothing, used for its side effects.
}
\description{
Removes the table where data typically gathered with
\code{\link[=tw_get_wikipedia_page_qid]{tw_get_wikipedia_page_qid()}} from cache.
}
\examples{
if (interactive()) {
  tw_reset_wikipedia_page_cache()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_set_cache_db}
\alias{tw_set_cache_db}
\title{Set database connection settings for the session}
\usage{
tw_set_cache_db(
  db_settings = NULL,
  driver = NULL,
  host = NULL,
  server = NULL,
  port = NULL,
  database = NULL,
  user = NULL,
  pwd = NULL
)
}
\arguments{
\item{db_settings}{A list of database connection settings (see example)}

\item{driver}{A database driver. Common database drivers include \code{MySQL},
\code{PostgreSQL}, and \code{MariaDB}. See \code{unique(odbc::odbcListDrivers()[[1]])} for
a list of locally available drivers.}

\item{host}{Host address, e.g. "localhost". Different drivers use server or
host parameter, only one of them is likely needed.}

\item{server}{Server address, e.g. "localhost". Different drivers use server
or host parameter, only one of them is likely needed.}

\item{port}{Port to use to connect to the database.}

\item{database}{Database name.}

\item{user}{Database user name.}

\item{pwd}{Password for the database user.}
}
\value{
A list with all given parameters (invisibly).
}
\description{
Set database connection settings for the session
}
\examples{
\donttest{
if (interactive()) {
  # Settings can be provided either as a list
  db_settings <- list(
    driver = "MySQL",
    host = "localhost",
    server = "localhost",
    port = 3306,
    database = "tidywikidatar",
    user = "secret_username",
    pwd = "secret_password"
  )

  tw_set_cache_db(db_settings)

  # or as parameters

  tw_set_cache_db(
    driver = "MySQL",
    host = "localhost",
    server = "localhost",
    port = 3306,
    database = "tidywikidatar",
    user = "secret_username",
    pwd = "secret_password"
  )

  # or ignoring fields that can be left to default values, such as "localhost" and port 3306

  tw_set_cache_db(
    driver = "MySQL",
    database = "tidywikidatar",
    user = "secret_username",
    pwd = "secret_password"
  )
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_page_section_links.R
\name{tw_get_wikipedia_section_links_api_url}
\alias{tw_get_wikipedia_section_links_api_url}
\title{Facilitates the creation of MediaWiki API base URLs to retrieve sections of a
page}
\usage{
tw_get_wikipedia_section_links_api_url(
  url = NULL,
  title = NULL,
  section_index,
  language = tidywikidatar::tw_get_language()
)
}
\arguments{
\item{url}{A character vector with the full URL to one or more Wikipedia
pages. If given, title and language can be left empty.}

\item{title}{Title of a Wikipedia page or final parts of its url. If given,
url can be left empty, but language must be provided.}

\item{section_index}{Required. It should correspond to the ordinal of a
section of the relevant Wikipedia page. See also
\code{\link[=tw_get_wikipedia_page_sections]{tw_get_wikipedia_page_sections()}}.}

\item{language}{Two-letter language code used to define the Wikipedia version
to use. Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not set,
"en". If URL given, this can be left empty.}
}
\value{
A character vector of base urls to be used with the MediaWiki API
}
\description{
Mostly used internally
}
\examples{
tw_get_wikipedia_section_links_api_url(title = "Margaret Mead", section_index = 1, language = "en")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_category_members.R
\name{tw_reset_wikipedia_category_members_cache}
\alias{tw_reset_wikipedia_category_members_cache}
\title{Reset Wikipedia category members cache}
\usage{
tw_reset_wikipedia_category_members_cache(
  language = tidywikidatar::tw_get_language(),
  type = "page",
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  ask = TRUE
)
}
\arguments{
\item{language}{Two-letter language code used to define the Wikipedia version
to use. Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not set,
"en". If url given, this can be left empty.}

\item{type}{Defaults to "page", defines which kind of members of a category
are returned. Valid values include "page", "file", and "subcat" (for
sub-category). Corresponds to \code{cmtype}. For details, see
\href{https://www.mediawiki.org/wiki/API:Categorymembers}{the relevant page
of the official documentation}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{ask}{Logical, defaults to \code{TRUE}. If \code{FALSE}, and cache folder does not
exist, it just removes it without asking (useful for non-interactive
sessions).}
}
\value{
Nothing, used for its side effects.
}
\description{
Removes from cache the table where data typically gathered with
\code{\link[=tw_get_wikipedia_category_members]{tw_get_wikipedia_category_members()}} are stored.
}
\examples{
if (interactive()) {
  tw_reset_wikipedia_category_members_cache()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_disable_cache}
\alias{tw_disable_cache}
\title{Disable caching for the current session}
\usage{
tw_disable_cache()
}
\value{
Nothing, used for its side effects.
}
\description{
Disable caching for the current session
}
\examples{
\donttest{
if (interactive()) {
  tw_disable_cache()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_property.R
\name{tw_get_property_same_length}
\alias{tw_get_property_same_length}
\alias{tw_get_p}
\title{Get Wikidata property of an item as a vector or list of the same length as
input}
\usage{
tw_get_property_same_length(
  id,
  p,
  only_first = FALSE,
  preferred = FALSE,
  latest_start_time = FALSE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
)

tw_get_p(
  id,
  p,
  only_first = FALSE,
  preferred = FALSE,
  latest_start_time = FALSE,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{p}{A character vector, a property. Must always start with the capital
letter "P", e.g. "P31" for "instance of".}

\item{only_first}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it just keeps the
first relevant property value for each id (or \code{NA} if none is available),
and returns a character vector. Warning: this likely discards valid values,
so make sure this is really what you want. If \code{FALSE}, returns a list of
the same length as input, with all values for each id stored in a list if
more than one is found.}

\item{preferred}{Logical, defaults to \code{FALSE}. If \code{TRUE}, returns properties
that have rank "preferred" if available; if no "preferred" property is
found, then it is ignored.}

\item{latest_start_time}{Logical, defaults to \code{FALSE}. If \code{TRUE}, returns the
property that has the most recent start time ("P580") as qualifier if
\code{only_first} is set to \code{TRUE}, or returns a list ordered by start time if
\code{only_first} is set to \code{FALSE}. If no such qualifier is found, then it is
ignored.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{id_df}{Default to \code{NULL}. If given, it should be a dataframe typically
generated with \code{\link[=tw_get]{tw_get()}}, and is used instead of calling Wikidata or
using SQLite cache. Ignored when \code{id} is of length more than one.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}
}
\value{
A list of the same length as input (or a character vector is
only_first is set to \code{TRUE})
}
\description{
Get Wikidata property of an item as a vector or list of the same length as
input
}
\examples{


# By default, it returns a list of the same length as input,
# no matter how many values for each id/property


\dontrun{
  tw_get_property_same_length(
    id = c(
      "Q180099",
      "Q228822",
      "Q76857"
    ),
    p = "P26"
  )
  # Notice that if no relevant match is found, it returns a NA
  # This is useful for piped operations

  tibble::tibble(id = c(
    "Q180099",
    "Q228822",
    "Q76857"
  )) \%>\%
    dplyr::mutate(spouse = tw_get_property_same_length(id, "P26"))

  # Consider unnesting for further analysis

  tibble::tibble(id = c(
    "Q180099",
    "Q228822",
    "Q76857"
  )) \%>\%
    dplyr::mutate(spouse = tw_get_property_same_length(id, "P26")) \%>\%
    tidyr::unnest(cols = spouse)

  # If you are sure that you are interested only in the first return value,
  # consider setting only_first=TRUE to get a character vector rather than a list
  # Be mindful: you may well be discarding valid values.
  tibble::tibble(id = c(
    "Q180099",
    "Q228822",
    "Q76857"
  )) \%>\%
    dplyr::mutate(spouse = tw_get_property_same_length(id, "P26",
      only_first = TRUE
    ))
}
tw_get_p(id = "Q180099", "P26")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_enable_cache}
\alias{tw_enable_cache}
\title{Enable caching for the current session}
\usage{
tw_enable_cache(SQLite = TRUE)
}
\arguments{
\item{SQLite}{Logical, defaults to \code{TRUE}. Set to \code{FALSE} to use custom
database options. See \code{\link[=tw_set_cache_db]{tw_set_cache_db()}} for details.}
}
\value{
Nothing, used for its side effects.
}
\description{
Enable caching for the current session
}
\examples{
\donttest{
if (interactive()) {
  tw_enable_cache()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_field.R
\name{tw_get_field}
\alias{tw_get_field}
\title{Gets a field such a label or description from a data frame typically generated
with \code{\link[=tw_get]{tw_get()}}}
\usage{
tw_get_field(df, field, id, language = tidywikidatar::tw_get_language())
}
\arguments{
\item{df}{A data frame typically generated with \code{\link[=tw_get]{tw_get()}}. It should include
data for the \code{id} included in the dedicated parameter.}

\item{field}{A character vector of length one. Typically, either "label" or
"description".}

\item{id}{A character vector, typically of Wikidata identifiers. The output
will be of the same length and in the same order as the identifiers
provided with this parameter.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}
}
\value{
A character vector of the same length, and with data in the same
order, as \code{id}.
}
\description{
Gets a field such a label or description from a data frame typically generated
with \code{\link[=tw_get]{tw_get()}}
}
\examples{
\dontrun{
tw_get("Q180099") \%>\%
  tw_get_field(field = "label", id = "Q180099")
}

# using test item for the sake of this example
tw_get("Q180099", id_l = tw_test_items) \%>\%
 tw_get_field(field = "label", id = "Q180099")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_p_wide.R
\name{tw_get_p_wide}
\alias{tw_get_p_wide}
\title{Efficiently get a wide table with various properties of a given set of
Wikidata identifiers}
\usage{
tw_get_p_wide(
  id,
  p,
  label = FALSE,
  property_label_as_column_name = FALSE,
  both_id_and_label = FALSE,
  only_first = FALSE,
  preferred = FALSE,
  unlist = FALSE,
  collapse = ";",
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  id_df_label = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{p}{A character vector, a property. Must always start with the capital
letter "P", e.g. "P31" for "instance of".}

\item{label}{Logical, defaults to \code{FALSE}. If \code{TRUE} labels of Wikidata Q
identifiers are reported instead of the identifiers themselves (or labels
are presented along of them, if \code{both_id_and_label} is set to \code{TRUE})}

\item{property_label_as_column_name}{Logical, defaults to \code{FALSE}. If \code{FALSE},
names of columns with properties are the "P" identifiers of the property.
If \code{TRUE}, the label of the correspondent property is assigned as column
name.}

\item{both_id_and_label}{Logical, defaults to \code{FALSE}. Relevant only if \code{label}
is set to \code{TRUE}, otherwise ignored. If \code{TRUE}, the label is added as a
separate column along the original one. Column name is the same as the
property column, followed by "_label".}

\item{only_first}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it just keeps the
first relevant property value for each id (or \code{NA} if none is available),
and returns a character vector. Warning: this likely discards valid values,
so make sure this is really what you want. If \code{FALSE}, returns a list of
the same length as input, with all values for each id stored in a list if
more than one is found.}

\item{preferred}{Logical, defaults to \code{FALSE}. If \code{TRUE}, returns properties
that have rank "preferred" if available; if no "preferred" property is
found, then it is ignored.}

\item{unlist}{Logical, defaults to \code{FALSE}. Typically used sharing or exporting
data as csv files. Collapses all properties in a single string. The
separator is defined by the \code{collapse} parameter. Relevant only when
\code{only_first} is set to \code{FALSE}.}

\item{collapse}{Defaults to ";". Character used to separate results when
\code{unlist} is set to \code{TRUE}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{id_df}{Default to \code{NULL}. If given, it should be a dataframe typically
generated with \code{\link[=tw_get]{tw_get()}}, and is used instead of calling Wikidata or
using SQLite cache. Ignored when \code{id} is of length more than one.}

\item{id_df_label}{Defaults to NULL. If given, it should be a dataframe
typically generated with \code{\link[=tw_get]{tw_get()}} with \emph{all} items for which labels will
be requested. It is used instead of calling Wikidata or relying on cache.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}
}
\value{
A data frame, with a column for each given property.
}
\description{
Efficiently get a wide table with various properties of a given set of
Wikidata identifiers
}
\examples{

if (interactive()) {
  tw_get_p_wide(
    id = c("Q180099", "Q228822", "Q191095"),
    p = c("P27", "P19", "P20"),
    label = TRUE,
    only_first = TRUE
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_search.R
\name{tw_search_item}
\alias{tw_search_item}
\title{Search for Wikidata properties in Wikidata and return Wikidata id, label, and
description.}
\usage{
tw_search_item(
  search,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  include_search = FALSE,
  wait = 0,
  user_agent = tidywikidatar::tw_get_user_agent(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{search}{A string to be searched in Wikidata}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}

\item{limit}{Maximum numbers of responses to be given.}

\item{include_search}{Logical, defaults to \code{FALSE}. If \code{TRUE}, the search is
returned as an additional column.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE.} Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Defaults to \code{FALSE}. If \code{TRUE}, overwrites cache.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
A data frame (a tibble) with three columns (\code{id}, \code{label}, and
\code{description}), and as many rows as there are results (by default, limited
to 10).
}
\description{
This search returns only items, use \code{\link[=tw_search_property]{tw_search_property()}} for properties.
}
\examples{
\dontrun{
  tw_search_item(search = "Sylvia Pankhurst")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_check_pid.R
\name{tw_check_pid}
\alias{tw_check_pid}
\title{Ensures that input appears to be a valid Wikidata property id (i.e. it starts
with P and is followed only by digits)}
\usage{
tw_check_pid(property, logical_vector = FALSE, non_pid_as_NA = FALSE)
}
\arguments{
\item{property}{A character vector of one or more Wikidata property
identifiers.}

\item{logical_vector}{Logical, defaults to \code{FALSE}. If \code{TRUE}, returns a
logical vector of the same length as input, where \code{TRUE} corresponds to
seemingly meaningful property identifiers.}

\item{non_pid_as_NA}{Logical, defaults to \code{FALSE}. If \code{TRUE} (and if
\code{logical_vector} is set to \code{FALSE}), a vector of the same length is
returned, with \code{NA} replacing items that are seemingly not meaningful
property identifiers.}
}
\value{
A character vector with only strings appearing to be Wikidata
identifiers; possibly shorter than input.
}
\description{
Mostly used internally by other functions.
}
\examples{

tw_check_pid(property = c("P19", "p20", "Not an property id", "20", NA, "Q5", ""))

tw_check_pid(
  property = c("P19", "p20", "Not an property id", "20", NA, "Q5", ""),
  logical_vector = TRUE
)

tw_check_pid(
  property = c("P19", "p20", "Not an property id", "20", NA, "Q5", ""),
  non_pid_as_NA = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{tw_test_items}
\alias{tw_test_items}
\title{A list mostly used for testing with some Wikidata items in the format resulting from \code{tw_get_item()}}
\format{
A list, an object such as the one resulting from \code{tw_get_item()}
}
\usage{
tw_test_items
}
\description{
A list mostly used for testing with some Wikidata items in the format resulting from \code{tw_get_item()}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_user_agent_settings.R
\name{tw_set_user_agent}
\alias{tw_set_user_agent}
\alias{tw_get_user_agent}
\title{Set language to be used by all functions}
\usage{
tw_set_user_agent(user_agent = NULL)

tw_get_user_agent(user_agent = NULL)
}
\arguments{
\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}
}
\value{
The user agent set for the session, implicitly.
}
\description{
Defaults to current package name (\code{tidywikidatar}) and version.
}
\examples{
# Default user agent
default_user_agent <- tw_get_user_agent()
default_user_agent
# Custom user agent
tw_set_user_agent(user_agent = "custom_project_name/email")
new_user_agent <- tw_get_user_agent()
new_user_agent
# Restore
tw_set_user_agent(user_agent = default_user_agent)
tw_get_user_agent()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get.R
\name{tw_get_single}
\alias{tw_get_single}
\title{Return (most) information from a Wikidata item in a tidy format from a single
Wikidata identifier}
\usage{
tw_get_single(
  id,
  language = tidywikidatar::tw_get_language(),
  retry = 10,
  cache = NULL,
  overwrite_cache = FALSE,
  read_cache = TRUE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  id_l = NULL,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{read_cache}{Logical, defaults to TRUE. Mostly used internally to
prevent checking if an item is in cache if it is already known that it is
not in cache.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{id_l}{Defaults to \code{NULL}. If given, must be an object or list such as
the one generated with \code{\link[=tw_get_item]{tw_get_item()}}. If given, and the requested
id is actually present in \code{id_l}, then no query to Wikidata servers is
made.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}
}
\value{
A data.frame (a tibble) with four columns (id, property, value, and
rank). If item not found or trouble connecting with the server, a data
frame with four columns and zero rows is returned, with the warning as an
attribute, which can be retrieved with \verb{attr(output, "warning"))}
}
\description{
Return (most) information from a Wikidata item in a tidy format from a single
Wikidata identifier
}
\examples{
if (interactive()) {
  tidywikidatar:::tw_get_single(
    id = "Q180099",
    language = "en"
  )
}

## using `tw_test_items` in examples in order to show output without calling
## on Wikidata servers

tidywikidatar:::tw_get_single(
  id = "Q180099",
  language = "en",
  id_l = tw_test_items
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_wikipedia_category_members.R
\name{tw_write_wikipedia_category_members_to_cache}
\alias{tw_write_wikipedia_category_members_to_cache}
\title{Write Wikipedia page links to cache}
\usage{
tw_write_wikipedia_category_members_to_cache(
  df,
  language = tidywikidatar::tw_get_language(),
  type = "page",
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{df}{A data frame typically generated with \code{tw_get_wikipedia_category_members()}.}

\item{language}{Defaults to language set with \code{tw_set_language()}; if not set, "en". Use "all_available" to keep all languages. For available language values, see https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}

\item{type}{Defaults to "page", defines which kind of members of a category are returned. Valid values include "page", "file", and "subcat" (for sub-category). Corresponds to \code{cmtype}. For details, see \url{https://www.mediawiki.org/wiki/API:Categorymembers}}

\item{cache}{Defaults to NULL. If given, it should be given either TRUE or FALSE. Typically set with \code{tw_enable_cache()} or \code{tw_disable_cache()}.}

\item{overwrite_cache}{Logical, defaults to FALSE. If TRUE, it overwrites the table in the local sqlite database. Useful if the original Wikidata object has been updated.}

\item{cache_connection}{Defaults to NULL. If NULL, and caching is enabled, \code{tidywikidatar} will use a local sqlite database. A custom connection to other databases can be given (see vignette \code{caching} for details).}

\item{disconnect_db}{Defaults to TRUE. If FALSE, leaves the connection to cache open.}
}
\value{
Silently returns the same data frame provided as input. Mostly used internally for its side effects.
}
\description{
Mostly used internally by \code{tidywikidatar}, use with caution to keep caching consistent.
}
\examples{
if (interactive()) {
  df <- tw_get_wikipedia_category_members(
    category = "American women anthropologists",
    language = "en",
    cache = FALSE
  )

  tw_write_wikipedia_category_members_to_cache(
    df = df,
    language = "en"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_write_item_to_cache.R
\name{tw_write_item_to_cache}
\alias{tw_write_item_to_cache}
\title{Writes item to cache}
\usage{
tw_write_item_to_cache(
  item_df,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{item_df}{A data frame with three columns typically generated with
\code{\link[=tw_get]{tw_get()}}.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
Nothing, used for its side effects.
}
\description{
Writes item to cache. Typically used internally, but exported to enable
custom caching solutions.
}
\examples{

tw_set_cache_folder(path = fs::path(tempdir(), paste(sample(letters, 24), collapse = "")))
tw_create_cache_folder(ask = FALSE)
tw_disable_cache()

df_from_api <- tw_get(id = "Q180099", language = "en")

df_from_cache <- tw_get_cached_item(
  id = "Q180099",
  language = "en"
)

is.null(df_from_cache) # expect TRUE, as nothing has yet been stored in cache

tw_write_item_to_cache(
  item_df = df_from_api,
  language = "en",
  cache = TRUE
)

df_from_cache <- tw_get_cached_item(
  id = "Q180099",
  language = "en",
  cache = TRUE
)

is.null(df_from_cache) # expect a data frame, same as df_from_api
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_get_cache_db}
\alias{tw_get_cache_db}
\title{Get database connection settings from the environment}
\usage{
tw_get_cache_db()
}
\value{
A list with all database parameters as stored in environment
variables.
}
\description{
Typically set with \code{\link[=tw_set_cache_db]{tw_set_cache_db()}}.
}
\examples{

tw_get_cache_db()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_check_qid.R
\name{tw_check_qid}
\alias{tw_check_qid}
\title{Ensures that input appears to be a valid Wikidata id}
\usage{
tw_check_qid(id, logical_vector = FALSE, non_id_as_NA = FALSE)
}
\arguments{
\item{id}{A character vector of one or more Wikidata id.}

\item{logical_vector}{Logical, defaults to \code{FALSE}. If \code{TRUE}, returns a
logical vector of the same length as input, where \code{TRUE} corresponds to
seemingly meaningful Q identifiers.}

\item{non_id_as_NA}{Logical, defaults to \code{FALSE}. If \code{TRUE} (and if
\code{logical_vector} is set to \code{FALSE}), a vector of the same length is
returned, with \code{NA} replacing items that are seemingly not meaningful Q
identifiers.}
}
\value{
A character vector with only strings appearing to be Wikidata
identifiers; possibly shorter than input.
}
\description{
Mostly used internally by other functions.
}
\examples{

tw_check_qid(id = c("Q180099", "q228822", "Not an id", "00180099", NA, "Q5"))

tw_check_qid(
  id = c("Q180099", "q228822", "Not an id", "00180099", NA, "Q5"),
  logical_vector = TRUE
)

tw_check_qid(
  id = c("Q180099", "q228822", "Not an id", "00180099", NA, "Q5"),
  non_id_as_NA = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_cache_indexing.R
\name{tw_check_cache_index}
\alias{tw_check_cache_index}
\title{Check if cache table is indexed}
\usage{
tw_check_cache_index(
  table_name = NULL,
  type = "item",
  show_details = FALSE,
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{table_name}{Name of the table in the database. If given, it takes
precedence over other parameters.}

\item{type}{Defaults to "item". Type of cache file to output. Values
typically used by \code{tidywikidatar} include "item", "search_item",
"search_property", and "qualifier".}

\item{show_details}{Logical, defaults to \code{FALSE}. If \code{FALSE}, return a
logical vector of length one (\code{TRUE} if the table was indexed, \code{FALSE} if
it was not). If \code{TRUE}, returns a data frame with more details about the
index.}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
If \code{show_details} is set to \code{FALSE}, return a logical vector of
length one (\code{TRUE} if the table was indexed, \code{FALSE} if it was not). If
\code{show_details} is set to \code{TRUE}, returns a data frame with more details
about the index.
}
\description{
Tested only with SQLite and MySql. May work with other drivers. Used to check
if given cache table is indexed (if created with any version of
\code{tidywikidatar} before 0.6, they are probably not indexed and less
efficient).
}
\examples{
if (interactive()) {
  tw_enable_cache()
  tw_set_cache_folder(path = fs::path(
    fs::path_home_r(),
    "R",
    "tw_data"
  ))

  tw_set_language(language = "en")

  tw_check_cache_index()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{tw_empty_wikipedia_category_members}
\alias{tw_empty_wikipedia_category_members}
\title{A zero-rows tibble used internally when \code{tw_empty_wikipedia_category_members()} would not return any value.}
\format{
A data frame with 0 rows and 3 columns
}
\usage{
tw_empty_wikipedia_category_members
}
\description{
A zero-rows tibble used internally when \code{tw_empty_wikipedia_category_members()} would not return any value.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_search.R
\name{tw_search_single}
\alias{tw_search_single}
\title{Search for Wikidata items or properties and return Wikidata id, label, and
description.}
\usage{
tw_search_single(
  search,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  limit = 10,
  retry = 10,
  include_search = FALSE,
  user_agent = tidywikidatar::tw_get_user_agent(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0
)
}
\arguments{
\item{search}{A string to be searched in Wikidata}

\item{type}{Defaults to "item". Either "item" or "property".}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}

\item{limit}{Maximum numbers of responses to be given.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{include_search}{Logical, defaults to \code{FALSE}. If \code{TRUE}, the search is
returned as an additional column.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE.} Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Defaults to \code{FALSE}. If \code{TRUE}, overwrites cache.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}
}
\value{
A data frame (a tibble) with three columns (\code{id}, \code{label}, and
\code{description}), and as many rows as there are results (by default, limited
to 10). Four columns when \code{include_search} is set to \code{TRUE}.
}
\description{
This search returns only items, use \code{\link[=tw_search_property]{tw_search_property()}} for properties.
}
\examples{
\dontrun{
  tidywikidatar:::tw_search_single(search = "Sylvia Pankhurst")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{tw_qid_meps}
\alias{tw_qid_meps}
\title{The Wikidata Q identifier of all members of the European Parliament since its establishment}
\format{
A data frame with 4581 rows and 1 column:
\describe{
\item{id}{Q identifiers}
}
}
\source{
\url{https://www.wikidata.org/wiki/Wikidata:Main_Page}
}
\usage{
tw_qid_meps
}
\description{
A dataset with all the Wikidata items that have "Q27169" (member of the European Parliament) for the property "P39" (position held).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-tidy-eval.R
\name{tidyeval}
\alias{tidyeval}
\alias{expr}
\alias{enquo}
\alias{enquos}
\alias{sym}
\alias{syms}
\alias{.data}
\alias{:=}
\alias{as_name}
\alias{as_label}
\title{Tidy eval helpers}
\description{
\itemize{
\item \code{\link[rlang]{sym}()} creates a symbol from a string and
\code{\link[rlang:sym]{syms}()} creates a list of symbols from a
character vector.
\item \code{\link[rlang:nse-defuse]{enquo}()} and
\code{\link[rlang:nse-defuse]{enquos}()} delay the execution of one or
several function arguments. \code{enquo()} returns a single quoted
expression, which is like a blueprint for the delayed computation.
\code{enquos()} returns a list of such quoted expressions.
\item \code{\link[rlang:nse-defuse]{expr}()} quotes a new expression \emph{locally}. It
is mostly useful to build new expressions around arguments
captured with \code{\link[=enquo]{enquo()}} or \code{\link[=enquos]{enquos()}}:
\code{expr(mean(!!enquo(arg), na.rm = TRUE))}.
\item \code{\link[rlang]{as_name}()} transforms a quoted variable name
into a string. Supplying something else than a quoted variable
name is an error.

That's unlike \code{\link[rlang]{as_label}()} which also returns
a single string but supports any kind of R object as input,
including quoted function calls and vectors. Its purpose is to
summarise that object into a single label. That label is often
suitable as a default name.

If you don't know what a quoted expression contains (for instance
expressions captured with \code{enquo()} could be a variable
name, a call to a function, or an unquoted constant), then use
\code{as_label()}. If you know you have quoted a simple variable
name, or would like to enforce this, use \code{as_name()}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_qualifiers.R
\name{tw_get_qualifiers_single}
\alias{tw_get_qualifiers_single}
\title{Get Wikidata qualifiers for a given property of a given item}
\usage{
tw_get_qualifiers_single(
  id,
  p,
  language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent(),
  id_l = NULL
)
}
\arguments{
\item{id}{A character vector, must start with Q, e.g. "Q180099" for the
anthropologist Margaret Mead. Can also be a data frame of one row,
typically generated with \code{\link[=tw_search]{tw_search()}} or a combination of \code{\link[=tw_search]{tw_search()}}
and \code{\link[=tw_filter_first]{tw_filter_first()}}.}

\item{p}{A character vector, a property. Must always start with the capital
letter "P", e.g. "P31" for "instance of".}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}

\item{id_l}{Defaults to \code{NULL}. If given, must be an object or list such as
the one generated with \code{\link[=tw_get_item]{tw_get_item()}}. If given, and the requested
id is actually present in \code{id_l}, then no query to Wikidata servers is
made.}
}
\value{
A data frame (a tibble) with eight columns: \code{id} for the input id,
\code{property},  \code{qualifier_id}, \code{qualifier_property}, \code{qualifier_value},
\code{rank}, \code{qualifier_value_type}, and \code{set} (to distinguish sets of data when
a property is present more than once)
}
\description{
N.B. In order to provide for consistently structured output, this function
outputs either id or value for each  qualifier. The user should keep in mind
that some of these come with additional detail (e.g. the unit, precision, or
reference calendar).
}
\examples{
if (interactive()) {
  tidywikidatar:::tw_get_qualifiers_single(id = "Q180099", p = "P26", language = "en")
}

#' ## using `tw_test_items` in examples in order to show output without calling
## on Wikidata servers

tidywikidatar:::tw_get_qualifiers_single(
  id = "Q180099",
  p = "P26",
  language = "en",
  id_l = tw_test_items
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_get_image.R
\name{tw_get_image}
\alias{tw_get_image}
\title{Get image from Wikimedia Commons}
\usage{
tw_get_image(
  id,
  format = "filename",
  width = NULL,
  language = tidywikidatar::tw_get_language(),
  id_df = NULL,
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE,
  wait = 0,
  retry = 10,
  user_agent = tidywikidatar::tw_get_user_agent()
)
}
\arguments{
\item{id}{A character vector of length 1, must start with Q, e.g. "Q254" for
Wolfgang Amadeus Mozart.}

\item{format}{A character vector, defaults to \code{filename}. If set to
\code{commons}, outputs the link to the Wikimedia Commons page. If set to
\code{embed}, outputs a link that can be used to embed.}

\item{width}{A numeric value, defaults to \code{NULL}, relevant only if format is
set to 'embed'. If not given, defaults to full resolution image.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{id_df}{Default to \code{NULL}. If given, it should be a dataframe typically
generated with \code{\link[=tw_get]{tw_get()}}, and is used instead of calling Wikidata or
using SQLite cache. Ignored when \code{id} is of length more than one.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Logical, defaults to \code{FALSE}. If \code{TRUE}, it overwrites
the table in the local sqlite database. Useful if the original Wikidata
object has been updated.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{wait}{In seconds, defaults to 0. Time to wait between queries to
Wikidata. If data are cached locally, wait time is not applied. If you are
running many queries systematically you may want to add some waiting time
between queries.}

\item{retry}{Defaults to 10. Maximum number of times to retry if the API
throws an error, such as "too many requests". Each time, it will wait as
much time as requested by the API. Notice that this can be a long time,
e.g. 30 minutes. Set to \code{FALSE} if you prefer the API to throw an error
immediately. Consider adjusting the \code{wait} parameter, or customising the
\code{user_agent} if relevant.}

\item{user_agent}{Defaults to \code{NULL}. If not given, implicitly defaults to
current package name (\code{tidywikidatar}) and version.}
}
\value{
A data frame of two columns, \code{id} and \code{image}, corresponding to
reference to the image in the requested format.
}
\description{
Please consult the
\href{https://commons.wikimedia.org/wiki/Commons:Reusing_content_outside_Wikimedia/technical}{relevant
documentation for reusing content outside Wikimedia}.
}
\examples{
\dontrun{
tw_get_image("Q180099",
  format = "filename"
)

  tw_get_image("Q180099",
    format = "commons"
  )

  tw_get_image("Q180099",
    format = "embed",
    width = 300
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_write_search_to_cache.R
\name{tw_write_search_to_cache}
\alias{tw_write_search_to_cache}
\title{Writes search to cache}
\usage{
tw_write_search_to_cache(
  search_df,
  type = "item",
  language = tidywikidatar::tw_get_language(),
  response_language = tidywikidatar::tw_get_language(),
  cache = NULL,
  overwrite_cache = FALSE,
  cache_connection = NULL,
  disconnect_db = TRUE
)
}
\arguments{
\item{search_df}{A data frame with four columns typically generated with
\code{tw_search(include_search = TRUE)}.}

\item{type}{Defaults to "item". Either "item" or "property".}

\item{language}{Language to be used for the search. Can be set once per
session with \code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full
list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{response_language}{Language to be used for the returned labels and
descriptions. Corresponds to the \code{uselang} parameter of the MediaWiki API,
as described
\href{https://www.wikidata.org/w/api.php?action=help&modules=wbsearchentities}{in
the official documentation}. Can be set once per session with
\code{\link[=tw_set_language]{tw_set_language()}}. If not set, defaults to "en". For a full list, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{all
available language codes}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE.} Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{overwrite_cache}{Defaults to \code{FALSE}. If \code{TRUE}, overwrites cache.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is
enabled, \code{tidywikidatar} will use a local sqlite database. A custom
connection to other databases can be given (see vignette \code{caching} for
details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}
}
\value{
Nothing, used for its side effects.
}
\description{
Writes search to cache. Typically used internally, but exported to enable
custom caching solutions.
}
\examples{
\dontrun{
tw_set_cache_folder(path = fs::path(tempdir(), paste(sample(letters, 24), collapse = "")))
tw_create_cache_folder(ask = FALSE)
tw_disable_cache()

search_from_api <- tw_search(search = "Sylvia Pankhurst", include_search = TRUE)

search_from_cache <- tw_get_cached_search("Sylvia Pankhurst")

nrow(search_from_cache) == 0 # expect TRUE, as nothing has yet been stored in cache

tw_write_search_to_cache(search_df = search_from_api)

search_from_cache <- tw_get_cached_search("Sylvia Pankhurst")

search_from_cache
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_connect_to_cache.R
\name{tw_connect_to_cache}
\alias{tw_connect_to_cache}
\title{Return a connection to be used for caching}
\usage{
tw_connect_to_cache(
  connection = NULL,
  RSQLite = NULL,
  language = tidywikidatar::tw_get_language(),
  cache = NULL
)
}
\arguments{
\item{connection}{Defaults to \code{NULL}. If \code{NULL}, uses local SQLite database.
If given, must be a connection object or a list with relevant connection
settings (see example).}

\item{RSQLite}{Defaults to \code{NULL}, expected either \code{NULL} or logical. If set
to \code{FALSE}, details on the database connection must be given either as a
named list in the connection parameter, or with \code{\link[=tw_set_cache_db]{tw_set_cache_db()}} as
environment variables.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{the
dedicated Wikimedia page}.}

\item{cache}{Defaults to \code{NULL}. If given, it should be given either \code{TRUE}
or \code{FALSE}. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}
}
\value{
A connection object.
}
\description{
Return a connection to be used for caching
}
\examples{
\donttest{
if (interactive()) {
  cache_connection <- pool::dbPool(
    RSQLite::SQLite(), # or e.g. odbc::odbc(),
    Driver = ":memory:", # or e.g. "MariaDB",
    Host = "localhost",
    database = "example_db",
    UID = "example_user",
    PWD = "example_pwd"
  )
  tw_connect_to_cache(cache_connection)


  db_settings <- list(
    driver = "MySQL",
    host = "localhost",
    server = "localhost",
    port = 3306,
    database = "tidywikidatar",
    user = "secret_username",
    pwd = "secret_password"
  )

  tw_connect_to_cache(db_settings)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_check_cache_folder}
\alias{tw_check_cache_folder}
\title{Checks if cache folder exists, if not returns an informative message}
\usage{
tw_check_cache_folder()
}
\value{
If the cache folder exists, returns \code{TRUE}. Otherwise throws an
error.
}
\description{
Checks if cache folder exists, if not returns an informative message
}
\examples{

# If cache folder does not exist, it throws an error
tryCatch(tw_check_cache_folder(),
  error = function(e) {
    return(e)
  }
)

# Create cache folder
tw_set_cache_folder(path = fs::path(
  tempdir(),
  "tw_cache_folder"
))
tw_create_cache_folder(ask = FALSE)

tw_check_cache_folder()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tw_caching.R
\name{tw_disconnect_from_cache}
\alias{tw_disconnect_from_cache}
\title{Ensure that connection to cache is disconnected consistently}
\usage{
tw_disconnect_from_cache(
  cache = NULL,
  cache_connection = NULL,
  disconnect_db = TRUE,
  language = tidywikidatar::tw_get_language()
)
}
\arguments{
\item{cache}{Defaults to NULL. If given, it should be given either \code{TRUE} or
FALSE. Typically set with \code{\link[=tw_enable_cache]{tw_enable_cache()}} or \code{\link[=tw_disable_cache]{tw_disable_cache()}}.}

\item{cache_connection}{Defaults to \code{NULL}. If \code{NULL}, and caching is enabled,
\code{tidywikidatar} will use a local sqlite database. A custom connection to
other databases can be given (see vignette \code{caching} for details).}

\item{disconnect_db}{Defaults to \code{TRUE}. If \code{FALSE}, leaves the connection to
cache open.}

\item{language}{Defaults to language set with \code{\link[=tw_set_language]{tw_set_language()}}; if not
set, "en". Use "all_available" to keep all languages. For available
language values, see the
\href{https://www.wikidata.org/wiki/Help:Wikimedia_language_codes/lists/all}{relevant
Wikimedia page}.}
}
\value{
Nothing, used for its side effects.
}
\description{
Ensure that connection to cache is disconnected consistently
}
\examples{
if (interactive()) {
  tw_get(
    id = c("Q180099"),
    language = "en"
  )
  tw_disconnect_from_cache()
}
}
